!> \file
!! Contains subroutine conduction_step()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CONDUCTION_STEP
!
!> Performs the conduction evolution at each timestep.
!<
subroutine conduction_step(t_conduc,t_comm)

  use parameters, only : nx
  use divers    , only : nsx,nsy,nsz
  use var
  use varold
  use conduction

  implicit none

#if PARA==1
  include 'mpif.h'
#else
  real     :: MPI_Wtime
  external :: MPI_Wtime
#endif

  real    :: t_conduc,t_comm
  integer :: i,j,k,ierr
  integer :: ixm,ixp,iym,iyp,izm,izp
  real    :: tr1,tr2
  
  tr1 = MPI_Wtime(ierr)
  
  rhoold  = rho
  Eold    = E
  rhouold = rhou
#if SCALP==1
  fxold = fx
#endif

  if (implicit_con) then
     call conduction_imp_gc
  else
     do i = 1,nsx
        do j = 1,nsy
           do k = 1,nsz
              ixm = 1 + (i-1)*nx(1)/nsx
              iym = 1 + (j-1)*nx(2)/nsy
              izm = 1 + (k-1)*nx(3)/nsz
              ixp = i*nx(1)/nsx
              iyp = j*nx(2)/nsy
              izp = k*nx(3)/nsz           
              call conduction_ex(ixm,ixp,iym,iyp,izm,izp)           
           enddo
        enddo
     enddo
  endif
  
  tr2 = MPI_Wtime(ierr)     
  t_conduc = t_conduc + tr2-tr1
  
  call communications(t_comm)
  call limites

  return
  
end subroutine conduction_step
