!> \file
!! Contains subroutine pasdt_con()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PASDT_CON
!
!> Computes the maximum timestep dictated by thermal
!! conduction.
!<
subroutine pasdt_con

  use parameters
  use var    
  use unites    , only : an,seconde,kb,uma,kg,metre,degre
  use geom      , only : x,geom_dir
  use para      , only : mype
  use conduction
  use divers    , only : verbose
  
  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  ! local variables
  integer                 :: i,j,k,ixmin,ixmax,iymin,iymax,izmin,izmax
  real, dimension(1:ndim) :: dx
  real                    :: dt,dt_glob
  real                    :: cv,TT,RR,kappa,cal_kappa_con

  if (verbose) write(*,*) 'Start of pasdt_con'

  dt = 1.0E+50*an
  
  ixmin =  0 ; ixmax = nx(1)+1
  iymin =  1 ; iymax = 1
  izmin =  1 ; izmax = 1
  if(ndim.gt.1) then
     iymin =  0 ; iymax = nx(2)+1
  endif
  if(ndim.gt.2) then
     izmin =  0 ; izmax = nx(3)+1
  endif

  ! compute Ein
  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)

  do k = izmin,izmax
  do j = iymin,iymax
  do i = ixmin,ixmax

                      dx(1) = x(i+1,1)-x(i,1)
     if (ndim .gt. 1) dx(2) = x(j+1,2)-x(j,2)
     if (ndim .gt. 2) dx(3) = x(k+1,3)-x(k,3)
     
     if (geom_dir(2) == 'phi_c') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
     if (geom_dir(2) == 'the_s') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
     if (geom_dir(3) == 'phi_s') dx(3) = 0.5*(x(i,1)+x(i+1,1))*sin(0.5*(x(j,2)+x(j+1,2))) * dx(3)
     
     RR = rho(i,j,k)
     call temperature  (RR,E(i,j,k),TT,1,1,1,1,1,1)
     call heat_capacity(RR,TT,cv,1,1,1,1,1,1)
     kappa = cal_kappa_con(RR,TT)
     
     dt = min(dt, 0.5 * dx(1)**2 *rho(i,j,k)*Cv/kappa)

  enddo
  enddo
  enddo

  ! restore E
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  ! Timestep reduction over cpu grid
#if PARA==1
  info = 0
  call MPI_ALLREDUCE(dt, dt_glob,1,MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,info)
#else
  dt_glob = dt
#endif
  
  dt_con_exp = dt_glob

  if (implicit_con) then
     dt_con = dt_con_imp     
  else
     dt_con = dt_con_exp
  endif

  if (verbose) write(*,*) 'End of pasdt_con'

  return

end subroutine pasdt_con
