!> \file
!! Contains subroutine grav_predictor()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GRAV_PREDICTOR
!
!> Performs prediction step in the gravity calculation.
!<
subroutine grav_predictor(u,phi,dt,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use parameters, only : ndim,N_vit
  use geom      , only : Cylindrique,Spherique,xcloc,geom_dir
  use const
  
  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  integer :: i,j,k,ilo,ihi,jlo,jhi,klo,khi
  
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:N_vit) :: u
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax        ) :: phi
  real :: dt,dphi,dtdx,dtdy,dtdz,xc,sinxc

  ilo = min(1,nxmin+1) ; ihi = max(1,nxmax-1)
  jlo = min(1,nymin+1) ; jhi = max(1,nymax-1)
  klo = min(1,nzmin+1) ; khi = max(1,nzmax-1)

  ! Gravitational acceleration along first direction
  do k = klo,khi
     do j = jlo,jhi
        do i = ilo,ihi
           
           dtdx = dt/(xcloc(i+1,1)-xcloc(i-1,1))
           
           ! slope of phi along x direction
           dphi = half*(phi(i+1,j,k)-phi(i-1,j,k))
           
           u(i,j,k,1) = u(i,j,k,1) - dphi*dtdx
           
        end do
     end do
  end do

  ! Gravitational acceleration along second direction
  if (ndim >= 2) then
     do k = klo,khi
        do j = jlo,jhi
           do i = ilo,ihi
              
              dtdy = dt/(xcloc(j+1,2)-xcloc(j-1,2))
              xc = 1.
              if ((geom_dir(2).eq.'phi_c').or.(Spherique)) xc = xcloc(i,1)
              
              ! Slope of phi along y direction
              dphi = half*(phi(i,j+1,k)-phi(i,j-1,k))
              
              u(i,j,k,2) = u(i,j,k,2) - dphi*dtdy/xc
              
           end do
        end do
     end do
  endif

  ! Gravitational acceleration along third direction
  if (ndim == 3) then
     do k = klo,khi
        do j = jlo,jhi
           do i = ilo,ihi
              
              dtdz = dt/(xcloc(k+1,3)-xcloc(k-1,3))
              xc = 1. ; sinxc = 1.
              if (Spherique) then
                 xc    = xcloc(i,1)
                 sinxc = sin(xcloc(j,2))
              endif
              
              ! Slope of phi along z direction
              dphi = half*(phi(i,j,k+1)-phi(i,j,k-1))
              
              u(i,j,k,3) = u(i,j,k,3) - dphi*dtdz/xc/sinxc
              
           end do
        end do
     end do
  endif

  return
  
end subroutine grav_predictor
