!> \file
!! Contains subroutine pasdt_hydro()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PASDT_HYDRO
!
!> Computes the maximum CFL limited timestep in the grid.
!<
subroutine pasdt_hydro

  use parameters
  use divers     , only : verbose,dt_cfl
  use var    
  use geom       , only : x,geom_dir
  use para       , only : mype
  use unites     , only : an

  implicit None

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

#if EOS==1
  real :: Sound_speed, dens
#endif

  ! Local variables
  real                    :: xx,dt_cfl_loc
  integer                 :: i,j,k,idim,ixmin,ixmax,iymin,iymax,izmin,izmax
  integer,dimension(3)    :: iloc
  real, dimension(1:ndim) :: dx
  real, dimension(:,:,:), allocatable :: cs

  if(verbose) write(*,*) 'Start of pasdt_hydro'

  allocate(cs(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))

  dt_cfl_loc = 1.0E+50*an

  ! compute Ein
  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)

  call sound_speed(rho,E,cs,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
 
  ! restore E
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  ixmin =  0 ; ixmax = nx(1)+1
  iymin =  1 ; iymax = 1
  izmin =  1 ; izmax = 1
  if(ndim.gt.1) then
     iymin =  0 ; iymax = nx(2)+1
  endif
  if(ndim.gt.2) then
     izmin =  0 ; izmax = nx(3)+1
  endif

  do k = izmin,izmax
     iloc(3) = k
     do j = iymin,iymax
        iloc(2) = j
        do i = ixmin,ixmax
           iloc(1) = i
           dx = 1.
           do idim = 1,ndim
              dx(idim) = ( x(iloc(idim)+1,idim)-x(iloc(idim),idim) )
           enddo
     
           if (geom_dir(2) == 'phi_c') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
           if (geom_dir(2) == 'the_s') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
           if (geom_dir(3) == 'phi_s') dx(3) = 0.5*(x(i,1)+x(i+1,1))*sin(0.5*(x(j,2)+x(j+1,2))) * dx(3)
           
           xx = 0.
           do idim = 1,ndim
              xx = xx + (cs(i,j,k) + abs(rhou(i,j,k,idim)/rho(i,j,k))) /dx(idim)
           enddo
           
           dt_cfl_loc = Min(1./xx, dt_cfl_loc)
        enddo
     enddo
  enddo

  ! Timestep reduction
#if PARA==1
  info = 0
  call MPI_ALLREDUCE(dt_cfl_loc, dt_cfl,1,MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,info)
#else
  dt_cfl = dt_cfl_loc
#endif

  dt_cfl = 0.8*dt_cfl

  if(verbose) write(*,*) 'End of pasdt_hydro'

  return
  
end subroutine pasdt_hydro
