!> \file
!! Contains module mfilm and subroutines film_() and
!! init_film()
!!
!! This file contains the routines and variables used to
!! create movies.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module MFILM
!
!> Contains variables used to create movies.
!<
module mfilm
  integer :: ifilm    !< Current number of film output
  real    :: dt_film  !< Timestep between film outputs
  real    :: t_film   !< t_film
  real    :: t_start  !< t_start
  real    :: t_end    !< t_end
  logical :: FILM     !< Creates movie outputs if .true.
  integer :: nl1      !< nl1
  integer :: nl2      !< nl2
  integer :: nl3      !< nl3
end module mfilm

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FILM_
!
!> Writes many density of velocity slices to create a movie.
!<
subroutine film_

  use parameters , only : nx, ndim,nx_glob,nx_cpu
  use var        , only : rho, E, rhou, fx
  use geom       , only : x_glob,nshift_gr
  use gammas
  use divers     , only : mu, verbose, temps
  use para       , only : mype,ncpu,x_start,x_start_cpu
  use unites     , only : gramme, cm3, erg
  use mfilm      , only : t_film, dt_film, ifilm

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: ierr,iproc,emetteur,nnx,nny,nnz
  integer, dimension(MPI_STATUS_SIZE) :: statut
  real, dimension(:,:,:), allocatable :: Image_rec
#endif

  integer :: i,j,k,tag1,count
  integer :: destination
  integer :: mil,cent,diz,unit,ii,jj,kk

  character(len=1 ) :: uu,dd,cc,mm
  character(len=13) :: fichf

!  real    :: idim,U2
!  real, dimension(3) :: U

  real , dimension(:,:,:), allocatable :: Image_glob, Image

  if (temps < t_film) return

  count = nx(1)*nx(2)*nx(3)

  allocate(Image(1:nx(1),1:nx(2),1:nx(3)))
  if (mype == 0) then
     allocate(Image_glob(1:nx_glob(1),1:nx_glob(2),1:nx_glob(3)))
     Image = 0.
  endif
  
  do k=1,nx(3)
  do j=1,nx(2)
  do i=1,nx(1)

!!$     U2 = 0. 
!!$     do idim=1,ndim
!!$        U(idim) = rhou(i,j,k,idim)/ rho(i,j,k)
!!$        U2 = U2 + U(idim)**2
!!$     enddo
!!$     Image(i,j,k) =  (gamma-1.)*(E(i,j,k) - 0.5*rho(i,j,k)*U2)

     Image(i,j,k) =  Rho(i,j,k)/(gramme/cm3)
     
  enddo
  enddo
  enddo

  if (mype > 0 ) then
     destination = 0
     tag1        = mype
     
#if PARA==1
     call MPI_send(Image,count,MPI_DOUBLE_PRECISION,destination,tag1 ,MPI_Comm_world,ierr)     
#endif

  endif


  if (mype == 0) then

#if PARA==1
     do iproc=1,ncpu-1
        emetteur = iproc
        tag1     = iproc
        nnx = nx_cpu(1,iproc) ; nny = nx_cpu(2,iproc); nnz = nx_cpu(3,iproc)
        count = nnx*nny*nnz
        allocate(Image_rec(nnx,nny,nnz))

        call MPI_recv(Image_rec ,count, MPI_DOUBLE_PRECISION,emetteur,tag1,MPI_COMM_WORLD,statut,ierr)

          do k=1,nnz
          do j=1,nny
          do i=1,nnx
             ii = i-1+x_start_cpu(1,iproc)
             jj = j-1+x_start_cpu(2,iproc)
             kk = k-1+x_start_cpu(3,iproc)
             Image_glob(ii,jj,kk) = Image_rec(i,j,k)

          enddo
          enddo
          enddo
          
          deallocate(Image_rec)
       enddo
#endif

       do k=1,nx(3)
       do j=1,nx(2)
       do i=1,nx(1)
          ii = i-1+x_start_cpu(1,mype)
          jj = j-1+x_start_cpu(2,mype)
          kk = k-1+x_start_cpu(3,mype)
          Image_glob(ii,jj,kk) = Image(i,j,k)

       enddo
       enddo
       enddo
     
        
       fichf = 'film_0000.bin'
       ii   = ifilm
       if (ii >= 1000) then
          mil  = ii/1000
          ii   = ii-1000*mil
       else
          mil = 0
       endif
       
       if (ii >= 100) then
          cent = ii/100
          ii   = ii-100*cent
       else
          cent = 0
       endif
       
       if (ii >= 10) then
          diz = ii/10
          ii  = ii-10*diz
       else
          diz = 0
       endif
       
       unit = ii
       
       uu=char(unit+48)
       dd=char(diz +48)
       cc=char(cent+48)
       mm=char(mil +48)
       
       fichf(6:6)=mm
       fichf(7:7)=cc
       fichf(8:8)=dd
       fichf(9:9)=uu

       write(*,*) 'Writing ',fichf
       open(45,file=fichf,form='unformatted')
       write(45) nx_glob(1), nx_glob(2)
       write(45) temps
       write(45) Image_glob
       close(45)
       
    endif


 ifilm   = ifilm   + 1
 t_film  = max(temps+dt_film, t_film+dt_film)


 deallocate(Image)
 if (mype == 0)   deallocate(Image_glob)

 if (verbose) write(*,*) 'End of Film'
 
 return
 
end subroutine film_

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_FILM
!
!> Initialisation of variables for creation of movies.
!<
subroutine init_film

  use parameters, only : nx, ndim, nx_glob
  use para      , only : ncpu,mype
  use unites    , only : an, seconde
  use divers    , only : temps,verbose
  use mfilm

  implicit none

  if(verbose) write(*,*) 'Start of init_film'

  t_start = temps
  t_end   = temps + 5.0E6*an

   t_film = t_start
  dt_film = 1000. * an  

   if( ndim == 1) then
     nl1 = 1
     nl2 = 0
     nl3 = 0
  else if( ndim == 2) then 
     nl1 = 1
     nl2 = 1
     nl3 = 0
  else if( ndim == 3) then 
     nl1 = 1
     nl2 = 1
     nl3 = 1 
  endif

  if(verbose) write(*,*) 'End of init_film'

  return
  
end subroutine init_film
