!> \file
!! Contains modules communication and para and subroutines
!! init_para(), allocate_array_com(), communications(),
!! communications_all(), com_1D(), com_2D() and com_3D()
!!
!! This file contains all the variables and routines
!! necessary to the communications between cpus.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module COMMUNICATION
!
!> Contains variables arrays used to communicate ghost
!! cell data across cpus.
!<
module communication
  implicit none
  real, dimension(:,:,:,:), allocatable :: face1_e        !< Face 1 e
  real, dimension(:,:,:,:), allocatable :: face2_e        !< Face 2 e
  real, dimension(:,:,:,:), allocatable :: face3_e        !< Face 3 e
  real, dimension(:,:,:,:), allocatable :: face4_e        !< Face 4 e
  real, dimension(:,:,:,:), allocatable :: face5_e        !< Face 5 e
  real, dimension(:,:,:,:), allocatable :: face6_e        !< Face 6 e
  real, dimension(:,:,:,:), allocatable :: face1_r        !< Face 1 r
  real, dimension(:,:,:,:), allocatable :: face2_r        !< Face 2 r
  real, dimension(:,:,:,:), allocatable :: face3_r        !< Face 3 r
  real, dimension(:,:,:,:), allocatable :: face4_r        !< Face 4 r
  real, dimension(:,:,:,:), allocatable :: face5_r        !< Face 5 r
  real, dimension(:,:,:,:), allocatable :: face6_r        !< Face 6 r
  real, dimension(:,:,:,:), allocatable :: ar13_e         !< Edge 1 e
  real, dimension(:,:,:,:), allocatable :: ar14_e         !< Edge 2 e
  real, dimension(:,:,:,:), allocatable :: ar23_e         !< Edge 3 e
  real, dimension(:,:,:,:), allocatable :: ar24_e         !< Edge 4 e
  real, dimension(:,:,:,:), allocatable :: ar15_e         !< Edge 5 e
  real, dimension(:,:,:,:), allocatable :: ar25_e         !< Edge 6 e
  real, dimension(:,:,:,:), allocatable :: ar35_e         !< Edge 7 e
  real, dimension(:,:,:,:), allocatable :: ar45_e         !< Edge 8 e
  real, dimension(:,:,:,:), allocatable :: ar16_e         !< Edge 9 e
  real, dimension(:,:,:,:), allocatable :: ar26_e         !< Edge 10 e
  real, dimension(:,:,:,:), allocatable :: ar36_e         !< Edge 11 e
  real, dimension(:,:,:,:), allocatable :: ar46_e         !< Edge 12 e
  real, dimension(:,:,:,:), allocatable :: ar13_r         !< Edge 1 r
  real, dimension(:,:,:,:), allocatable :: ar14_r         !< Edge 2 r
  real, dimension(:,:,:,:), allocatable :: ar23_r         !< Edge 3 r
  real, dimension(:,:,:,:), allocatable :: ar24_r         !< Edge 4 r
  real, dimension(:,:,:,:), allocatable :: ar15_r         !< Edge 5 r
  real, dimension(:,:,:,:), allocatable :: ar25_r         !< Edge 6 r
  real, dimension(:,:,:,:), allocatable :: ar35_r         !< Edge 7 r
  real, dimension(:,:,:,:), allocatable :: ar45_r         !< Edge 8 r
  real, dimension(:,:,:,:), allocatable :: ar16_r         !< Edge 9 r
  real, dimension(:,:,:,:), allocatable :: ar26_r         !< Edge 10 r
  real, dimension(:,:,:,:), allocatable :: ar36_r         !< Edge 11 r
  real, dimension(:,:,:,:), allocatable :: ar46_r         !< Edge 12 r
  real, dimension(:,:,:,:), allocatable :: coin13_e       !< Corner 2D 1 e
  real, dimension(:,:,:,:), allocatable :: coin14_e       !< Corner 2D 2 e
  real, dimension(:,:,:,:), allocatable :: coin23_e       !< Corner 2D 3 e
  real, dimension(:,:,:,:), allocatable :: coin24_e       !< Corner 2D 4 e
  real, dimension(:,:,:,:), allocatable :: coin13_r       !< Corner 2D 1 r
  real, dimension(:,:,:,:), allocatable :: coin14_r       !< Corner 2D 2 r
  real, dimension(:,:,:,:), allocatable :: coin23_r       !< Corner 2D 3 r
  real, dimension(:,:,:,:), allocatable :: coin24_r       !< Corner 2D 4 r
  real, dimension(:,:,:,:), allocatable :: coin135_e      !< Corner 3D 1 e
  real, dimension(:,:,:,:), allocatable :: coin136_e      !< Corner 3D 2 e
  real, dimension(:,:,:,:), allocatable :: coin145_e      !< Corner 3D 3 e
  real, dimension(:,:,:,:), allocatable :: coin146_e      !< Corner 3D 4 e
  real, dimension(:,:,:,:), allocatable :: coin235_e      !< Corner 3D 5 e
  real, dimension(:,:,:,:), allocatable :: coin236_e      !< Corner 3D 6 e
  real, dimension(:,:,:,:), allocatable :: coin245_e      !< Corner 3D 7 e
  real, dimension(:,:,:,:), allocatable :: coin246_e      !< Corner 3D 8 e
  real, dimension(:,:,:,:), allocatable :: coin135_r      !< Corner 3D 1 r
  real, dimension(:,:,:,:), allocatable :: coin136_r      !< Corner 3D 2 r
  real, dimension(:,:,:,:), allocatable :: coin145_r      !< Corner 3D 3 r
  real, dimension(:,:,:,:), allocatable :: coin146_r      !< Corner 3D 4 r
  real, dimension(:,:,:,:), allocatable :: coin235_r      !< Corner 3D 5 r
  real, dimension(:,:,:,:), allocatable :: coin236_r      !< Corner 3D 6 r
  real, dimension(:,:,:,:), allocatable :: coin245_r      !< Corner 3D 7 r
  real, dimension(:,:,:,:), allocatable :: coin246_r      !< Corner 3D 8 r
end module communication

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module PARA
!
!> Contains variables defining the MPI parameters of the
!! simulation.
!<
module para
  implicit none
  integer, dimension(:,:,:), allocatable :: grid_cpu      !< Number of cpus in the x,y,z, directions
  integer, dimension(:,:  ), allocatable :: x_start_cpu   !< Starting x,y,z coordinates which cpu holds
  integer, dimension(:,:  ), allocatable :: x_end_cpu     !< Ending x,y,z coordinates which cpu holds
  integer, dimension(:    ), allocatable :: lim_x         !< lim_x
  integer, dimension(:    ), allocatable :: lim_y         !< lim_y
  integer, dimension(:    ), allocatable :: lim_z         !< lim_z
  integer, dimension(-1:1,-1:1,-1:1)     :: voisin        !< Cpu neighbours
  integer, dimension(3)                  :: grid_pos      !< grid_pos
  integer, dimension(3)                  :: x_start       !< x_start
  integer, dimension(3)                  :: x_end         !< x_end
  integer                                :: ncpu          !< Total number of cpus
  integer                                :: mype          !< Cpu rank
  integer                                :: ncpu_x        !< Number of cpus in the x direction
  integer                                :: ncpu_y        !< Number of cpus in the y direction
  integer                                :: ncpu_z        !< Number of cpus in the z direction
end module para

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_PARA
!
!> This subroutine initialises the parallelisation of the
!! simualtion. It decomposes the compuational domain by
!! distributing it evenly among the cpus.
!<
subroutine init_para

  use parameters 
  use para 
  use cl
  use cl_ray
  use cl_con
  use cl_dif

  implicit none

#if PARA==1
  include 'mpif.h'
#endif

  integer, parameter :: etiquette1=100, etiquette2=101
  integer            :: i,j,k
  integer            :: ix,iy,iz
  integer            :: ic

!  if (mype == 0) write(*,*) 'Para : N_vit =',N_vit

#if MHD==1
     nvar     = 2*N_vit + 2  
     nvar_ray = ndim+2
#else
     nvar     = ndim+2+nfx
     nvar_ray = ndim+2
#endif

  allocate(lim_x(1:ncpu_x+1))
  allocate(lim_y(1:ncpu_y+1))
  allocate(lim_z(1:ncpu_z+1))
  allocate(grid_cpu(ncpu_x,ncpu_y,ncpu_z))
  allocate(nx_cpu(3,0:ncpu-1))
  allocate(x_start_cpu(3,0:ncpu-1), x_end_cpu(3,0:ncpu-1))

  ! splitting in the x direction
  lim_x(1) = 1
  do ix=1,ncpu_x
     ic = nx_glob(1)/ncpu_x
     if (ix <= mod(nx_glob(1),ncpu_x)) then
        ic = ic+1
     endif
     lim_x(ix+1) = lim_x(ix) + ic     
  enddo

  ! splitting in the y direction
  lim_y(1) = 1
  do iy=1,ncpu_y
     ic = nx_glob(2)/ncpu_y
     if (iy <= mod(nx_glob(2),ncpu_y)) then
        ic = ic+1
     endif
     lim_y(iy+1) = lim_y(iy) + ic     
  enddo

  ! splitting in the z direction
  lim_z(1) = 1
  do iz=1,ncpu_z
     ic = nx_glob(3)/ncpu_z
     if (iz <= mod(nx_glob(3),ncpu_z)) then
        ic = ic+1
     endif
     lim_z(iz+1) = lim_z(iz) + ic     
  enddo


  ic = 0 ; icl = 0 ; icl_ray = 0 ; icl_dif = 0 ; icl_con = 0
  do iz = 1,ncpu_z 
  do iy = 1,ncpu_y 
  do ix = 1,ncpu_x 

     grid_cpu(ix,iy,iz) = ic 
     x_start_cpu(1,ic)  = lim_x(ix) ; x_end_cpu(1,ic) = lim_x(ix+1)-1
     x_start_cpu(2,ic)  = lim_y(iy) ; x_end_cpu(2,ic) = lim_y(iy+1)-1
     x_start_cpu(3,ic)  = lim_z(iz) ; x_end_cpu(3,ic) = lim_z(iz+1)-1
     
     nx_cpu(1,ic) = x_end_cpu(1,ic)-x_start_cpu(1,ic)+1
     nx_cpu(2,ic) = x_end_cpu(2,ic)-x_start_cpu(2,ic)+1
     nx_cpu(3,ic) = x_end_cpu(3,ic)-x_start_cpu(3,ic)+1
     
     if (mype == ic) then
        grid_pos(1) = ix
        grid_pos(2) = iy
        grid_pos(3) = iz
        x_start(1)  = lim_x(ix) ; x_end(1) = lim_x(ix+1)-1
        x_start(2)  = lim_y(iy) ; x_end(2) = lim_y(iy+1)-1
        x_start(3)  = lim_z(iz) ; x_end(3) = lim_z(iz+1)-1

        nx(1) = x_end(1)-x_start(1)+1
        nx(2) = x_end(2)-x_start(2)+1
        nx(3) = x_end(3)-x_start(3)+1

        if (ix == 1      ) then
           icl    (1) = icl_glob    (1)
           icl_ray(1) = icl_ray_glob(1)
           icl_con(1) = icl_con_glob(1)
           icl_dif(1) = icl_dif_glob(1)
        endif
        if (ix == ncpu_x ) then
           icl    (2) = icl_glob    (2) 
           icl_ray(2) = icl_ray_glob(2)
           icl_con(2) = icl_con_glob(2)
           icl_dif(2) = icl_dif_glob(2)
        endif

        if (iy == 1      ) then
           icl    (3) = icl_glob    (3) 
           icl_ray(3) = icl_ray_glob(3)
           icl_con(3) = icl_con_glob(3)
           icl_dif(3) = icl_dif_glob(3)
        endif
        if (iy == ncpu_y ) then
           icl    (4) = icl_glob    (4)
           icl_ray(4) = icl_ray_glob(4)
           icl_con(4) = icl_con_glob(4)
           icl_dif(4) = icl_dif_glob(4)
        endif

        if (iz == 1      ) then
           icl    (5) = icl_glob    (5) 
           icl_ray(5) = icl_ray_glob(5)
           icl_con(5) = icl_con_glob(5)
           icl_dif(5) = icl_dif_glob(5)
        endif

        if (iz == ncpu_z ) then
           icl    (6) = icl_glob    (6) 
           icl_ray(6) = icl_ray_glob(6)
           icl_con(6) = icl_con_glob(6)
           icl_dif(6) = icl_dif_glob(6)
        endif
     endif

     ic = ic + 1

  enddo
  enddo
  enddo

  icl0 = icl
  
  ! determine the neighbouring cpus

  do k=-1,1
     iz = grid_pos(3) + k
     if (iz == 0       ) iz = ncpu_z
     if (iz == ncpu_z+1) iz = 1

     do j=-1,1
        iy = grid_pos(2) + j
        if (iy == 0       ) iy = ncpu_y
        if (iy == ncpu_y+1) iy = 1

        do i=-1,1
           ix = grid_pos(1) +i
           if (ix == 0       ) ix = ncpu_x
           if (ix == ncpu_x+1) ix = 1

           voisin(i,j,k) =  grid_cpu(ix,iy,iz) 

        enddo
     enddo
  enddo

  return
  
end subroutine init_para

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COMMUNICATIONS_ALL
!
!> Performs communications for all physical modules.
!<
subroutine communications_all

  use cputime  , only : temps_comm
  use diffusion, only : Er_dif
  use gravity  , only : Phi
  use divers   , only : RAY,DIFFU,GRAV

  implicit none

  ! Communications and boundary conditions
  call communications(temps_comm)
  call limites

#if RAYO==1
  if (RAY) then
     call communications_ray(temps_comm)
     call limites_ray
  endif
#endif

#if DIF==1
  if (DIFFU) then
     call communications_dif(temps_comm)
     call limites_dif
  endif
#endif

#if GRA==1
  if (GRAV) then
     call communications_gen(Phi,1,temps_comm)
     call limites_gra(Phi)
  endif
#endif

  return

end subroutine communications_all

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine ALLOCATE_ARRAY_COM
!
!> Allocates the arrays used in the communication process.
!<
subroutine allocate_array_com

  use communication
  use parameters
  use para

  implicit none

  integer :: nnx,nny,nnz
  
  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  if (ndim == 1) then
     allocate(face1_e(Nbuf,1,1,nvar), face2_e(Nbuf,1,1,nvar))
     allocate(face1_r(Nbuf,1,1,nvar), face2_r(Nbuf,1,1,nvar))
  endif

  if (ndim == 2) then
     allocate(face1_e(Nbuf,1:nny,1,nvar), face2_e(Nbuf,1:nny,1,nvar))
     allocate(face1_r(Nbuf,1:nny,1,nvar), face2_r(Nbuf,1:nny,1,nvar))

     allocate(face3_e(1:nnx,Nbuf,1,nvar), face4_e(1:nnx,Nbuf,1,nvar))
     allocate(face3_r(1:nnx,Nbuf,1,nvar), face4_r(1:nnx,Nbuf,1,nvar))

     allocate(coin13_e(Nbuf,Nbuf,1,nvar),coin14_e(Nbuf,Nbuf,1,nvar),coin23_e(Nbuf,Nbuf,1,nvar),coin24_e(Nbuf,Nbuf,1,nvar))
     allocate(coin13_r(Nbuf,Nbuf,1,nvar),coin14_r(Nbuf,Nbuf,1,nvar),coin23_r(Nbuf,Nbuf,1,nvar),coin24_r(Nbuf,Nbuf,1,nvar))
  endif

  if (ndim == 3) then
     allocate(face1_e(Nbuf,1:nny,1:nnz,nvar), face2_e(Nbuf,1:nny,1:nnz,nvar))
     allocate(face1_r(Nbuf,1:nny,1:nnz,nvar), face2_r(Nbuf,1:nny,1:nnz,nvar))

     allocate(face3_e(1:nnx,Nbuf,1:nnz,nvar), face4_e(1:nnx,Nbuf,1:nnz,nvar))
     allocate(face3_r(1:nnx,Nbuf,1:nnz,nvar), face4_r(1:nnx,Nbuf,1:nnz,nvar))

     allocate(face5_e(1:nnx,1:nny,Nbuf,nvar), face6_e(1:nnx,1:nny,Nbuf,nvar))
     allocate(face5_r(1:nnx,1:nny,Nbuf,nvar), face6_r(1:nnx,1:nny,Nbuf,nvar))


     allocate(ar13_e(Nbuf,Nbuf,1:nnz,nvar), ar14_e(Nbuf,Nbuf,1:nnz,nvar), ar23_e(Nbuf,Nbuf,1:nnz,nvar), ar24_e(Nbuf,Nbuf,1:nnz,nvar))
     allocate(ar13_r(Nbuf,Nbuf,1:nnz,nvar), ar14_r(Nbuf,Nbuf,1:nnz,nvar), ar23_r(Nbuf,Nbuf,1:nnz,nvar), ar24_r(Nbuf,Nbuf,1:nnz,nvar))

     allocate(ar15_e(Nbuf,1:nny,Nbuf,nvar), ar25_e(Nbuf,1:nny,Nbuf,nvar), ar16_e(Nbuf,1:nny,Nbuf,nvar), ar26_e(Nbuf,1:nny,Nbuf,nvar))
     allocate(ar15_r(Nbuf,1:nny,Nbuf,nvar), ar25_r(Nbuf,1:nny,Nbuf,nvar), ar16_r(Nbuf,1:nny,Nbuf,nvar), ar26_r(Nbuf,1:nny,Nbuf,nvar))

     allocate(ar35_e(1:nnx,Nbuf,Nbuf,nvar), ar45_e(1:nnx,Nbuf,Nbuf,nvar), ar36_e(1:nnx,Nbuf,Nbuf,nvar), ar46_e(1:nnx,Nbuf,Nbuf,nvar))
     allocate(ar35_r(1:nnx,Nbuf,Nbuf,nvar), ar45_r(1:nnx,Nbuf,Nbuf,nvar), ar36_r(1:nnx,Nbuf,Nbuf,nvar), ar46_r(1:nnx,Nbuf,Nbuf,nvar))

     allocate(coin135_e(Nbuf,Nbuf,Nbuf,nvar),coin136_e(Nbuf,Nbuf,Nbuf,nvar),coin145_e(Nbuf,Nbuf,Nbuf,nvar),coin146_e(Nbuf,Nbuf,Nbuf,nvar))
     allocate(coin135_r(Nbuf,Nbuf,Nbuf,nvar),coin136_r(Nbuf,Nbuf,Nbuf,nvar),coin145_r(Nbuf,Nbuf,Nbuf,nvar),coin146_r(Nbuf,Nbuf,Nbuf,nvar))
     allocate(coin235_e(Nbuf,Nbuf,Nbuf,nvar),coin236_e(Nbuf,Nbuf,Nbuf,nvar),coin245_e(Nbuf,Nbuf,Nbuf,nvar),coin246_e(Nbuf,Nbuf,Nbuf,nvar))
     allocate(coin235_r(Nbuf,Nbuf,Nbuf,nvar),coin236_r(Nbuf,Nbuf,Nbuf,nvar),coin245_r(Nbuf,Nbuf,Nbuf,nvar),coin246_r(Nbuf,Nbuf,Nbuf,nvar))
  endif

  return

end subroutine allocate_array_com

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COMMUNICATIONS
!
!> Perform communications and return cputime used.
!<
subroutine communications(tcomm)

  use parameters, only : ndim

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: ierr
  real :: tr1,tr2
#endif

  real :: tcomm

#if PARA==1
  tr1 = MPI_Wtime(ierr)
  if (ndim == 1) then
     call com_1D
  else if (ndim == 2) then
     call com_2D
  else if (ndim == 3) then
     call com_3D
  endif
  tr2 = MPI_Wtime(ierr)     
  tcomm = tcomm + tr2-tr1
#endif

  return

end subroutine communications

!###########################################################
!###########################################################
!###########################################################
!###########################################################

#if PARA==1

!  Subroutine COM_1D
!
!> Perform 1D communications.
!<
subroutine com_1D

  use communication
  use parameters
  use divers       , only : verbose
  use para
  use var

  implicit none

  include 'mpif.h'

  integer, parameter                    :: LEFT=1,RIGHT=2
  integer                               :: icount,ierr,nnx
  integer                               :: destination, emetteur
  integer, dimension(4)                 :: Req
  integer, dimension(MPI_STATUS_SIZE,4) :: Stat

  if(verbose) write(*,*) 'Start of com_1d'

  nnx = nx(1)

  icount   = nvar*Nbuf
  Req(1:4) = MPI_REQUEST_NULL

  face1_e(1:Nbuf,1,1,1                ) = rho (1:Nbuf,1,1        ) 
  face1_e(1:Nbuf,1,1,2:N_vit+1        ) = rhou(1:Nbuf,1,1,1:N_vit) 
  face1_e(1:Nbuf,1,1,  N_vit+2        ) = E   (1:Nbuf,1,1        ) 
#if MHD==1
  face1_e(1:Nbuf,1,1,N_vit+3:2*N_vit+2) = B   (1:Nbuf,1,1,1:N_vit) 
#endif
#if SCALP==1
   face1_e(1:Nbuf,1,1,nvar-nfx+1:nvar) = fx  (1:Nbuf,1,1,1:nfx) 
#endif  
  
  face2_e(1:Nbuf,1,1,1                ) = rho (nnx+1-Nbuf:nnx,1,1        ) 
  face2_e(1:Nbuf,1,1,2:N_vit+1        ) = rhou(nnx+1-Nbuf:nnx,1,1,1:N_vit) 
  face2_e(1:Nbuf,1,1,N_vit+2          ) = E   (nnx+1-Nbuf:nnx,1,1        ) 
#if MHD==1
  face2_e(1:Nbuf,1,1,N_vit+3:2*N_vit+2) = B   (nnx+1-Nbuf:nnx,1,1,1:N_vit) 
#endif
#if SCALP==1
  face2_e(1:Nbuf,1,1,nvar-nfx+1:nvar) = fx  (nnx+1-Nbuf:nnx,1,1,1:nfx) 
#endif  

  ! send data
  destination = voisin(-1,0,0)
  call MPI_Isend(face1_e,icount,MPI_DOUBLE_PRECISION,destination,LEFT ,MPI_Comm_world,Req(1),ierr)

  destination = voisin(+1,0,0)
  call MPI_Isend(face2_e,icount,MPI_DOUBLE_PRECISION,destination,RIGHT,MPI_COMM_WORLD,Req(2),ierr)

  ! receive data
  emetteur = voisin(+1,0,0)
  call MPI_Irecv(face2_r,icount,MPI_DOUBLE_PRECISION,emetteur   ,LEFT ,MPI_COMM_WORLD,Req(3),ierr)
  
  emetteur = voisin(-1,0,0)
  call MPI_Irecv(face1_r,icount,MPI_DOUBLE_PRECISION,emetteur   ,RIGHT,MPI_COMM_WORLD,Req(4),ierr)
  
  call MPI_Waitall(4,Req,Stat,ierr)

  rho (1-Nbuf:0,1,1        ) = face1_r(1:Nbuf,1,1,1                )
  Rhou(1-Nbuf:0,1,1,1:N_vit) = face1_r(1:Nbuf,1,1,2:N_vit+1        )
  E   (1-Nbuf:0,1,1        ) = face1_r(1:Nbuf,1,1,  N_vit+2        )
#if MHD==1
  B   (1-Nbuf:0,1,1,1:N_vit) = face1_r(1:Nbuf,1,1,N_vit+3:2*N_vit+2)
#endif
#if SCALP==1
  fx  (1-Nbuf:0,1,1,1:nfx  ) = face1_r(1:Nbuf,1,1,nvar-Nfx+1:nvar  )
#endif  

  rho (nnx+1:nnx+Nbuf,1,1        ) = face2_r(1:Nbuf,1,1,1                )
  rhou(nnx+1:nnx+Nbuf,1,1,1:N_vit) = face2_r(1:Nbuf,1,1,2:N_vit+1        )
  E   (nnx+1:nnx+Nbuf,1,1        ) = face2_r(1:Nbuf,1,1,  N_vit+2        )
#if MHD==1
  B   (nnx+1:nnx+Nbuf,1,1,1:N_vit) = face2_r(1:Nbuf,1,1,N_vit+3:2*N_vit+2)
#endif
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf,1,1,1:nfx) = face2_r(1:Nbuf,1,1,nvar-Nfx+1:nvar)
#endif

  if(verbose) write(*,*) 'End of com_1d'
  
  return
  
end subroutine com_1D

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COM_2D
!
!> Perform 2D communications.
!<
subroutine com_2D

  use communication
  use parameters
  use para
  use var
  use divers       , only : verbose

  implicit none

  include 'mpif.h'

  integer, parameter                     :: LEFT=1,RIGHT=2,DOWN=3, UP=4, RD=13, RU=14, LD=23, LU = 24
  integer                                :: ierr
  integer                                :: ic_f12,ic_f34,ic_coin
  integer                                :: destination, emetteur
  integer, dimension(16  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,16) :: Stat

  integer :: nnx, nny

  if(verbose) write(*,*) 'Start of com_2d'

  nnx = nx(1) ; nny = nx(2)

  ic_f12  = nvar*Nbuf*nny
  ic_f34  = nvar*Nbuf*nnx
  ic_coin = nvar*Nbuf*Nbuf

  Req(1:16) = MPI_REQUEST_NULL

  ! 4 faces
  face1_e(1:Nbuf,1:nny,1,1                ) = rho (1:Nbuf,1:nny,1        ) 
  face1_e(1:Nbuf,1:nny,1,2:N_vit+1        ) = rhou(1:Nbuf,1:nny,1,1:N_vit) 
  face1_e(1:Nbuf,1:nny,1,  N_vit+2        ) = E   (1:Nbuf,1:nny,1        ) 
#if MHD==1
  face1_e(1:Nbuf,1:nny,1,N_vit+3:2*N_vit+2) = B   (1:Nbuf,1:nny,1,1:N_vit) 
#endif  
#if SCALP==1
  face1_e(1:Nbuf,1:nny,1,nvar-Nfx+1:nvar) = fx  (1:Nbuf,1:nny,1,1:nfx) 
#endif  

  face2_e(1:Nbuf,1:nny,1,1                ) = rho (nnx-(Nbuf-1):nnx,1:nny,1        ) 
  face2_e(1:Nbuf,1:nny,1,2:N_vit+1        ) = rhou(nnx-(Nbuf-1):nnx,1:nny,1,1:N_vit) 
  face2_e(1:Nbuf,1:nny,1,  N_vit+2        ) = E   (nnx-(Nbuf-1):nnx,1:nny,1        ) 
#if MHD==1
  face2_e(1:Nbuf,1:nny,1,N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx,1:nny,1,1:N_vit) 
#endif  
#if SCALP==1
  face2_e(1:Nbuf,1:nny,1,nvar-Nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx,1:nny,1,1:nfx) 
#endif  

  face3_e(1:nnx,1:Nbuf,1,1                ) = rho (1:nnx,1:Nbuf,1        ) 
  face3_e(1:nnx,1:Nbuf,1,2:N_vit+1        ) = rhou(1:nnx,1:Nbuf,1,1:N_vit) 
  face3_e(1:nnx,1:Nbuf,1,  N_vit+2        ) = E   (1:nnx,1:Nbuf,1        ) 
#if MHD==1
  face3_e(1:nnx,1:Nbuf,1,N_vit+3:2*N_vit+2) = B   (1:nnx,1:Nbuf,1,1:N_vit) 
#endif  
#if SCALP==1
  face3_e(1:nnx,1:Nbuf,1,nvar-Nfx+1:nvar) = fx  (1:nnx,1:Nbuf,1,1:nfx) 
#endif  

  face4_e(1:nnx,1:Nbuf,1,1                ) = rho (1:nnx,nny-(Nbuf-1):nny,1        ) 
  face4_e(1:nnx,1:Nbuf,1,2:N_vit+1        ) = rhou(1:nnx,nny-(Nbuf-1):nny,1,1:N_vit) 
  face4_e(1:nnx,1:Nbuf,1,  N_vit+2        ) = E   (1:nnx,nny-(Nbuf-1):nny,1        ) 
#if MHD==1
  face4_e(1:nnx,1:Nbuf,1,N_vit+3:2*N_vit+2) = B   (1:nnx,nny-(Nbuf-1):nny,1,1:N_vit) 
#endif  
#if SCALP==1
  face4_e(1:nnx,1:Nbuf,1,nvar-Nfx+1:nvar) = fx  (1:nnx,nny-(Nbuf-1):nny,1,1:nfx) 
#endif  

  ! 4 corners
  coin13_e(1:Nbuf,1:Nbuf,1,1                ) = rho (1:Nbuf,1:Nbuf,1        ) 
  coin13_e(1:Nbuf,1:Nbuf,1,2:N_vit+1        ) = rhou(1:Nbuf,1:Nbuf,1,1:N_vit) 
  coin13_e(1:Nbuf,1:Nbuf,1,  N_vit+2        ) = E   (1:Nbuf,1:Nbuf,1        ) 
#if MHD==1
  coin13_e(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2) = B  (1:Nbuf,1:Nbuf,1,1:N_vit) 
#endif  
#if SCALP==1
  coin13_e(1:Nbuf,1:Nbuf,1,nvar-Nfx+1:nvar) = fx  (1:Nbuf,1:Nbuf,1,1:nfx) 
#endif  

  coin23_e(1:Nbuf,1:Nbuf,1,1                ) = rho (nnx-(Nbuf-1):nnx,1:Nbuf,1        ) 
  coin23_e(1:Nbuf,1:Nbuf,1,2:N_vit+1        ) = rhou(nnx-(Nbuf-1):nnx,1:Nbuf,1,1:N_vit) 
  coin23_e(1:Nbuf,1:Nbuf,1,  N_vit+2        ) = E   (nnx-(Nbuf-1):nnx,1:Nbuf,1        ) 
#if MHD==1
  coin23_e(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx,1:Nbuf,1,1:N_vit) 
#endif  
#if SCALP==1
  coin23_e(1:Nbuf,1:Nbuf,1,nvar-Nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx,1:Nbuf,1,1:nfx) 
#endif  
  
  coin14_e(1:Nbuf,1:Nbuf,1,1                ) = rho (1:Nbuf,nny-(Nbuf-1):nny,1        ) 
  coin14_e(1:Nbuf,1:Nbuf,1,2:N_vit+1        ) = rhou(1:Nbuf,nny-(Nbuf-1):nny,1,1:N_vit) 
  coin14_e(1:Nbuf,1:Nbuf,1,  N_vit+2        ) = E   (1:Nbuf,nny-(Nbuf-1):nny,1        ) 
#if MHD==1
  coin14_e(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2) = B   (1:Nbuf,nny-(Nbuf-1):nny,1,1:N_vit) 
#endif  
#if SCALP==1
  coin14_e(1:Nbuf,1:Nbuf,1,nvar-Nfx+1:nvar) = fx  (1:Nbuf,nny-(Nbuf-1):nny,1,1:nfx) 
#endif  

  coin24_e(1:Nbuf,1:Nbuf,1,1                ) = rho (nnx-(Nbuf-1):nnx,nny-(Nbuf-1):nny,1        ) 
  coin24_e(1:Nbuf,1:Nbuf,1,2:N_vit+1        ) = rhou(nnx-(Nbuf-1):nnx,nny-(Nbuf-1):nny,1,1:N_vit) 
  coin24_e(1:Nbuf,1:Nbuf,1,  N_vit+2        ) = E   (nnx-(Nbuf-1):nnx,nny-(Nbuf-1):nny,1        ) 
#if MHD==1
  coin24_e(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx,nny-(Nbuf-1):nny,1,1:N_vit) 
#endif  
#if SCALP==1
  coin24_e(1:Nbuf,1:Nbuf,1,nvar-Nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx,nny-(Nbuf-1):nny,1,1:nfx) 
#endif  


  ! send data
  destination = voisin(-1, 0,0)
  call MPI_Isend(face1_e ,ic_f12 ,MPI_DOUBLE_PRECISION,destination,LEFT ,MPI_Comm_world,Req( 1),ierr)

  destination = voisin(+1, 0,0)
  call MPI_Isend(face2_e ,ic_f12 ,MPI_DOUBLE_PRECISION,destination,RIGHT,MPI_COMM_WORLD,Req( 2),ierr)

  destination = voisin( 0,-1,0)
  call MPI_Isend(face3_e ,ic_f34 ,MPI_DOUBLE_PRECISION,destination,DOWN ,MPI_COMM_WORLD,Req( 3),ierr)

  destination = voisin( 0,+1,0)
  call MPI_Isend(face4_e ,ic_f34 ,MPI_DOUBLE_PRECISION,destination,UP   ,MPI_COMM_WORLD,Req( 4),ierr)

  destination = voisin(-1,-1,0)
  call MPI_Isend(coin13_e,ic_coin,MPI_DOUBLE_PRECISION,destination,LD   ,MPI_COMM_WORLD,Req( 5),ierr)

  destination = voisin(-1,+1,0)
  call MPI_Isend(coin14_e,ic_coin,MPI_DOUBLE_PRECISION,destination,LU   ,MPI_COMM_WORLD,Req( 6),ierr)

  destination = voisin(+1,-1,0)
  call MPI_Isend(coin23_e,ic_coin,MPI_DOUBLE_PRECISION,destination,RD   ,MPI_COMM_WORLD,Req( 7),ierr)

  destination = voisin(+1,+1,0)
  call MPI_Isend(coin24_e,ic_coin,MPI_DOUBLE_PRECISION,destination,RU   ,MPI_COMM_WORLD,Req( 8),ierr)

  ! receive data
  emetteur = voisin(+1, 0,0)
  call MPI_Irecv(face2_r ,ic_f12 ,MPI_DOUBLE_PRECISION,emetteur   ,LEFT ,MPI_COMM_WORLD,Req( 9),ierr)
  
  emetteur = voisin(-1, 0,0)
  call MPI_Irecv(face1_r ,ic_f12 ,MPI_DOUBLE_PRECISION,emetteur   ,RIGHT,MPI_COMM_WORLD,Req(10),ierr)
  
  emetteur = voisin( 0,+1,0)
  call MPI_Irecv(face4_r ,ic_f34 ,MPI_DOUBLE_PRECISION,emetteur   ,DOWN ,MPI_COMM_WORLD,Req(11),ierr)
  
  emetteur = voisin( 0,-1,0)
  call MPI_Irecv(face3_r ,ic_f34 ,MPI_DOUBLE_PRECISION,emetteur   ,UP   ,MPI_COMM_WORLD,Req(12),ierr)
  

  emetteur = voisin(+1,+1,0)
  call MPI_Irecv(coin24_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,LD   ,MPI_COMM_WORLD,Req(13),ierr)
  
  emetteur = voisin(+1,-1,0)
  call MPI_Irecv(coin23_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,LU   ,MPI_COMM_WORLD,Req(14),ierr)

  emetteur = voisin(-1,+1,0)
  call MPI_Irecv(coin14_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,RD   ,MPI_COMM_WORLD,Req(15),ierr)

  emetteur = voisin(-1,-1,0)
  call MPI_Irecv(coin13_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,RU   ,MPI_COMM_WORLD,Req(16),ierr)


  call MPI_Waitall(16,Req,Stat,ierr)

  ! 4 faces
  rho (1-Nbuf:0,1:nny,1        ) = face1_r(1:Nbuf,1:nny,1,1                )
  rhou(1-Nbuf:0,1:nny,1,1:N_vit) = face1_r(1:Nbuf,1:nny,1,2:N_vit+1        )
  E   (1-Nbuf:0,1:nny,1        ) = face1_r(1:Nbuf,1:nny,1,  N_vit+2        )
#if MHD==1
  B   (1-Nbuf:0,1:nny,1,1:N_vit) = face1_r(1:Nbuf,1:nny,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1-Nbuf:0,1:nny,1,1:nfx) = face1_r(1:Nbuf,1:nny,1,nvar-nfx+1:nvar)
#endif  

  rho (nnx+1:nnx+Nbuf,1:nny,1        ) = face2_r(1:Nbuf,1:nny,1,1                )
  rhou(nnx+1:nnx+Nbuf,1:nny,1,1:N_vit) = face2_r(1:Nbuf,1:nny,1,2:N_vit+1        )
  E   (nnx+1:nnx+Nbuf,1:nny,1        ) = face2_r(1:Nbuf,1:nny,1,  N_vit+2        )
#if MHD==1
  B   (nnx+1:nnx+Nbuf,1:nny,1,1:N_vit) = face2_r(1:Nbuf,1:nny,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf,1:nny,1,1:nfx) = face2_r(1:Nbuf,1:nny,1,nvar-nfx+1:nvar)
#endif  

  rho (1:nnx,1-Nbuf:0,1        ) = face3_r(1:nnx,1:Nbuf,1,1                )
  rhou(1:nnx,1-Nbuf:0,1,1:N_vit) = face3_r(1:nnx,1:Nbuf,1,2:N_vit+1        )
  E   (1:nnx,1-Nbuf:0,1        ) = face3_r(1:nnx,1:Nbuf,1,  N_vit+2        )
#if MHD==1
  B   (1:nnx,1-Nbuf:0,1,1:N_vit) = face3_r(1:nnx,1:Nbuf,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1:nnx,1-Nbuf:0,1,1:nfx) = face3_r(1:nnx,1:Nbuf,1,nvar-nfx+1:nvar)
#endif  

  rho (1:nnx,nny+1:nny+Nbuf,1        ) = face4_r(1:nnx,1:Nbuf,1,1                )
  rhou(1:nnx,nny+1:nny+Nbuf,1,1:N_vit) = face4_r(1:nnx,1:Nbuf,1,2:N_vit+1        )
  E   (1:nnx,nny+1:nny+Nbuf,1        ) = face4_r(1:nnx,1:Nbuf,1,  N_vit+2        )
#if MHD==1
  B   (1:nnx,nny+1:nny+Nbuf,1,1:N_vit) = face4_r(1:nnx,1:Nbuf,1,N_vit+3:2*N_vit+2)  
#endif  
#if SCALP==1
  fx  (1:nnx,nny+1:nny+Nbuf,1,1:nfx) = face4_r(1:nnx,1:Nbuf,1,nvar-nfx+1:nvar)  
#endif  

  ! 4 corners
  rho (1-Nbuf:0,1-Nbuf:0,1        ) = coin13_r(1:Nbuf,1:Nbuf,1,1                )
  rhou(1-Nbuf:0,1-Nbuf:0,1,1:N_vit) = coin13_r(1:Nbuf,1:Nbuf,1,2:N_vit+1        )
  E   (1-Nbuf:0,1-Nbuf:0,1        ) = coin13_r(1:Nbuf,1:Nbuf,1,  N_vit+2        )     
#if MHD==1
  B   (1-Nbuf:0,1-Nbuf:0,1,1:N_vit) = coin13_r(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1-Nbuf:0,1-Nbuf:0,1,1:nfx) = coin13_r(1:Nbuf,1:Nbuf,1,nvar-nfx+1:nvar)
#endif  

  rho (1-Nbuf:0,nny+1:nny+Nbuf,1        ) = coin14_r(1:Nbuf,1:Nbuf,1,1                )
  rhou(1-Nbuf:0,nny+1:nny+Nbuf,1,1:N_vit) = coin14_r(1:Nbuf,1:Nbuf,1,2:N_vit+1        )
  E   (1-Nbuf:0,nny+1:nny+Nbuf,1        ) = coin14_r(1:Nbuf,1:Nbuf,1,  N_vit+2        )
#if MHD==1
  B   (1-Nbuf:0,nny+1:nny+Nbuf,1,1:N_vit) = coin14_r(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1-Nbuf:0,nny+1:nny+Nbuf,1,1:nfx) = coin14_r(1:Nbuf,1:Nbuf,1,nvar-nfx+1:nvar)
#endif  

  rho (nnx+1:nnx+Nbuf,1-Nbuf:0,1        ) = coin23_r(1:Nbuf,1:Nbuf,1,1                )
  rhou(nnx+1:nnx+Nbuf,1-Nbuf:0,1,1:N_vit) = coin23_r(1:Nbuf,1:Nbuf,1,2:N_vit+1        )
  E   (nnx+1:nnx+Nbuf,1-Nbuf:0,1        ) = coin23_r(1:Nbuf,1:Nbuf,1,  N_vit+2        )
#if MHD==1
  B   (nnx+1:nnx+Nbuf,1-Nbuf:0,1,1:N_vit) = coin23_r(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf,1-Nbuf:0,1,1:nfx) = coin23_r(1:Nbuf,1:Nbuf,1,nvar-nfx+1:nvar)
#endif  

  rho (nnx+1:nnx+Nbuf,nny+1:nny+Nbuf,1        ) = coin24_r(1:Nbuf,1:Nbuf,1,1                )
  rhou(nnx+1:nnx+Nbuf,nny+1:nny+Nbuf,1,1:N_vit) = coin24_r(1:Nbuf,1:Nbuf,1,2:N_vit+1        )
  E   (nnx+1:nnx+Nbuf,nny+1:nny+Nbuf,1        ) = coin24_r(1:Nbuf,1:Nbuf,1,  N_vit+2        )
#if MHD==1
  B   (nnx+1:nnx+Nbuf,nny+1:nny+Nbuf,1,1:N_vit) = coin24_r(1:Nbuf,1:Nbuf,1,N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf,nny+1:nny+Nbuf,1,1:nfx) = coin24_r(1:Nbuf,1:Nbuf,1,nvar-nfx+1:nvar)
#endif  

  if(verbose) write(*,*) 'End of com_2d'
  
  return

end subroutine com_2D

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COM_3D
!
!> Perform 3D communications.
!<
subroutine com_3D

  use communication
  use parameters
  use para
  use var
  use divers       , only : verbose

  implicit none

  include 'mpif.h'

  integer, parameter                     :: F1=1, F2=2, F3=3, F4=4, F5=5, F6=6
  integer, parameter                     :: AR13=13, AR14=14, AR15=15, AR16=16
  integer, parameter                     :: AR23=23, AR24=24, AR25=25, AR26=26
  integer, parameter                     :: AR35=35, AR36=36
  integer, parameter                     :: AR45=45, AR46=46
  integer, parameter                     :: C135=135, C136=136, C145=145, C146=146
  integer, parameter                     :: C235=235, C236=236, C245=245, C246=246
  integer                                :: ierr
  integer                                :: ic_f12,ic_f34,ic_f56, ic_arx, ic_ary, ic_arz, ic_coin
  integer                                :: destination, emetteur
  integer, dimension(52  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,52) :: Stat

  integer :: nnx, nny, nnz

  if(verbose) write(*,*) 'Start of com_3d'

  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  ic_f12  = nvar*Nbuf*nny*nnz
  ic_f34  = nvar*Nbuf*nnx*nnz
  ic_f56  = nvar*Nbuf*nnx*nny

  ic_arx  = nvar*Nbuf*Nbuf*nnx
  ic_ary  = nvar*Nbuf*Nbuf*nny
  ic_arz  = nvar*Nbuf*Nbuf*nnz

  ic_coin = nvar*Nbuf*Nbuf*Nbuf

  Req(1:52) = MPI_REQUEST_NULL


  ! 6 faces
  face1_e(1:Nbuf,1:nny,1:nnz,1                ) = rho (1:Nbuf,1:nny,1:nnz        ) 
  face1_e(1:Nbuf,1:nny,1:nnz,2:N_vit+1        ) = rhou(1:Nbuf,1:nny,1:nnz,1:N_vit) 
  face1_e(1:Nbuf,1:nny,1:nnz,  N_vit+2        ) = E   (1:Nbuf,1:nny,1:nnz        ) 
#if MHD==1
  face1_e(1:Nbuf,1:nny,1:nnz,N_vit+3:2*N_vit+2) = B   (1:Nbuf,1:nny,1:nnz,1:N_vit) 
#endif  
#if SCALP==1
  face1_e(1:Nbuf,1:nny,1:nnz,nvar-nfx+1:nvar) = fx  (1:Nbuf,1:nny,1:nnz,1:nfx) 
#endif  

  face2_e(1:Nbuf,1:nny,1:nnz,1                ) = rho (nnx-(Nbuf-1):nnx,1:nny,1:nnz        ) 
  face2_e(1:Nbuf,1:nny,1:nnz,2:N_vit+1        ) = rhou(nnx-(Nbuf-1):nnx,1:nny,1:nnz,1:N_vit) 
  face2_e(1:Nbuf,1:nny,1:nnz,  N_vit+2        ) = E   (nnx-(Nbuf-1):nnx,1:nny,1:nnz        ) 
#if MHD==1
  face2_e(1:Nbuf,1:nny,1:nnz,N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx,1:nny,1:nnz,1:N_vit) 
#endif  
#if SCALP==1
  face2_e(1:Nbuf,1:nny,1:nnz,nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx,1:nny,1:nnz,1:nfx) 
#endif  

  face3_e(1:nnx,1:Nbuf,1:nnz,1                ) = rho (1:nnx,1:Nbuf,1:nnz        ) 
  face3_e(1:nnx,1:Nbuf,1:nnz,2:N_vit+1        ) = rhou(1:nnx,1:Nbuf,1:nnz,1:N_vit) 
  face3_e(1:nnx,1:Nbuf,1:nnz,  N_vit+2        ) = E   (1:nnx,1:Nbuf,1:nnz        ) 
#if MHD==1
  face3_e(1:nnx,1:Nbuf,1:nnz,N_vit+3:2*N_vit+2) = B   (1:nnx,1:Nbuf,1:nnz,1:N_vit) 
#endif  
#if SCALP==1
  face3_e(1:nnx,1:Nbuf,1:nnz,nvar-nfx+1:nvar) = fx  (1:nnx,1:Nbuf,1:nnz,1:nfx) 
#endif  

  face4_e(1:nnx,1:Nbuf,1:nnz,1                ) = rho (1:nnx,nny-(Nbuf-1):nny,1:nnz        ) 
  face4_e(1:nnx,1:Nbuf,1:nnz,2:N_vit+1        ) = rhou(1:nnx,nny-(Nbuf-1):nny,1:nnz,1:N_vit) 
  face4_e(1:nnx,1:Nbuf,1:nnz,  N_vit+2        ) = E   (1:nnx,nny-(Nbuf-1):nny,1:nnz        ) 
#if MHD==1
  face4_e(1:nnx,1:Nbuf,1:nnz,N_vit+3:2*N_vit+2) = B   (1:nnx,nny-(Nbuf-1):nny,1:nnz,1:N_vit) 
#endif  
#if SCALP==1
  face4_e(1:nnx,1:Nbuf,1:nnz,nvar-nfx+1:nvar) = fx  (1:nnx,nny-(Nbuf-1):nny,1:nnz,1:nfx) 
#endif  

  face5_e(1:nnx,1:nny,1:Nbuf,1                ) = rho (1:nnx,1:nny,1:Nbuf        ) 
  face5_e(1:nnx,1:nny,1:Nbuf,2:N_vit+1        ) = rhou(1:nnx,1:nny,1:Nbuf,1:N_vit) 
  face5_e(1:nnx,1:nny,1:Nbuf,  N_vit+2        ) = E   (1:nnx,1:nny,1:Nbuf        ) 
#if MHD==1
  face5_e(1:nnx,1:nny,1:Nbuf,N_vit+3:2*N_vit+2) = B   (1:nnx,1:nny,1:Nbuf,1:N_vit) 
#endif  
#if SCALP==1
  face5_e(1:nnx,1:nny,1:Nbuf,nvar-nfx+1:nvar) = fx  (1:nnx,1:nny,1:Nbuf,1:nfx) 
#endif  

  face6_e(1:nnx,1:nny,1:Nbuf,1                ) = rho (1:nnx,1:nny,nnz-(Nbuf-1):nnz        ) 
  face6_e(1:nnx,1:nny,1:Nbuf,2:N_vit+1        ) = rhou(1:nnx,1:nny,nnz-(Nbuf-1):nnz,1:N_vit) 
  face6_e(1:nnx,1:nny,1:Nbuf,  N_vit+2        ) = E   (1:nnx,1:nny,nnz-(Nbuf-1):nnz        ) 
#if MHD==1
  face6_e(1:nnx,1:nny,1:Nbuf,N_vit+3:2*N_vit+2) = B   (1:nnx,1:nny,nnz-(Nbuf-1):nnz,1:N_vit) 
#endif  
#if SCALP==1
  face6_e(1:nnx,1:nny,1:Nbuf,nvar-nfx+1:nvar) = fx  (1:nnx,1:nny,nnz-(Nbuf-1):nnz,1:nfx) 
#endif  

  ! 12 edges
  ar13_e(1:Nbuf, 1:Nbuf, 1:nnz, 1                ) = rho (1:Nbuf, 1:Nbuf, 1:nnz        )
  ar13_e(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         ) = rhou(1:Nbuf, 1:Nbuf, 1:nnz,1:N_vit)
  ar13_e(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         ) = E   (1:Nbuf, 1:Nbuf, 1:nnz        )
#if MHD==1
  ar13_e(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2) = B   (1:Nbuf, 1:Nbuf, 1:nnz,1:N_vit)
#endif  
#if SCALP==1
  ar13_e(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar) = fx  (1:Nbuf, 1:Nbuf, 1:nnz,1:nfx)
#endif  


  ar14_e(1:Nbuf, 1:Nbuf, 1:nnz, 1                ) = rho (1:Nbuf, nny-(Nbuf-1):nny, 1:nnz        )
  ar14_e(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         ) = rhou(1:Nbuf, nny-(Nbuf-1):nny, 1:nnz,1:N_vit)
  ar14_e(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         ) = E   (1:Nbuf, nny-(Nbuf-1):nny, 1:nnz        )
#if MHD==1
  ar14_e(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2) = B   (1:Nbuf, nny-(Nbuf-1):nny, 1:nnz,1:N_vit)
#endif  
#if SCALP==1
  ar14_e(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar) = fx  (1:Nbuf, nny-(Nbuf-1):nny, 1:nnz,1:nfx)
#endif  

  ar23_e(1:Nbuf, 1:Nbuf, 1:nnz, 1                ) = rho (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:nnz        )
  ar23_e(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, 1:Nbuf, 1:nnz,1:N_vit)
  ar23_e(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:nnz        )
#if MHD==1
  ar23_e(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:nnz,1:N_vit)
#endif  
#if SCALP==1
  ar23_e(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:nnz,1:nfx)
#endif  

  ar24_e(1:Nbuf, 1:Nbuf, 1:nnz, 1                ) = rho (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:nnz        )
  ar24_e(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:nnz,1:N_vit)
  ar24_e(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:nnz        )
#if MHD==1
  ar24_e(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:nnz,1:N_vit)
#endif  
#if SCALP==1
  ar24_e(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:nnz,1:nfx)
#endif  

  ar15_e(1:Nbuf, 1:nny, 1:Nbuf, 1                ) = rho (1:Nbuf, 1:nny, 1:Nbuf        )
  ar15_e(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         ) = rhou(1:Nbuf, 1:nny, 1:Nbuf,1:N_vit)
  ar15_e(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         ) = E   (1:Nbuf, 1:nny, 1:Nbuf        )
#if MHD==1
  ar15_e(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:Nbuf, 1:nny, 1:Nbuf,1:N_vit)  
#endif  
#if SCALP==1
  ar15_e(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:Nbuf, 1:nny, 1:Nbuf,1:nfx)  
#endif  

  ar25_e(1:Nbuf, 1:nny, 1:Nbuf, 1                ) = rho (nnx-(Nbuf-1):nnx, 1:nny, 1:Nbuf        )
  ar25_e(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, 1:nny, 1:Nbuf,1:N_vit)
  ar25_e(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, 1:nny, 1:Nbuf        )
#if MHD==1
  ar25_e(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, 1:nny, 1:Nbuf,1:N_vit)
#endif  
#if SCALP==1
  ar25_e(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, 1:nny, 1:Nbuf,1:nfx)
#endif  

  ar16_e(1:Nbuf, 1:nny, 1:Nbuf, 1                ) = rho (1:Nbuf, 1:nny, nnz-(Nbuf-1):nnz        )
  ar16_e(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         ) = rhou(1:Nbuf, 1:nny, nnz-(Nbuf-1):nnz,1:N_vit)
  ar16_e(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         ) = E   (1:Nbuf, 1:nny, nnz-(Nbuf-1):nnz        )
#if MHD==1
  ar16_e(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:Nbuf, 1:nny, nnz-(Nbuf-1):nnz,1:N_vit)
#endif  
#if SCALP==1
  ar16_e(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:Nbuf, 1:nny, nnz-(Nbuf-1):nnz,1:nfx)
#endif  

  ar26_e(1:Nbuf, 1:nny, 1:Nbuf, 1                ) = rho (nnx-(Nbuf-1):nnx, 1:nny, nnz-(Nbuf-1):nnz        )
  ar26_e(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, 1:nny, nnz-(Nbuf-1):nnz,1:N_vit)
  ar26_e(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, 1:nny, nnz-(Nbuf-1):nnz        )
#if MHD==1
  ar26_e(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, 1:nny, nnz-(Nbuf-1):nnz,1:N_vit)
#endif  
#if SCALP==1
  ar26_e(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, 1:nny, nnz-(Nbuf-1):nnz,1:nfx)
#endif  

  ar35_e(1:nnx, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:nnx, 1:Nbuf, 1:Nbuf        )
  ar35_e(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:nnx, 1:Nbuf, 1:Nbuf,1:N_vit)
  ar35_e(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:nnx, 1:Nbuf, 1:Nbuf        )
#if MHD==1
  ar35_e(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:nnx, 1:Nbuf, 1:Nbuf,1:N_vit)
#endif  
#if SCALP==1
  ar35_e(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:nnx, 1:Nbuf, 1:Nbuf,1:nfx)
#endif  

  ar45_e(1:nnx, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:nnx, nny-(Nbuf-1):nny, 1:Nbuf        )
  ar45_e(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:nnx, nny-(Nbuf-1):nny, 1:Nbuf,1:N_vit)
  ar45_e(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:nnx, nny-(Nbuf-1):nny, 1:Nbuf        )
#if MHD==1
  ar45_e(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:nnx, nny-(Nbuf-1):nny, 1:Nbuf,1:N_vit)
#endif  
#if SCALP==1
  ar45_e(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:nnx, nny-(Nbuf-1):nny, 1:Nbuf,1:nfx)
#endif  

  ar36_e(1:nnx, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:nnx, 1:Nbuf, nnz-(Nbuf-1):nnz        )
  ar36_e(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:nnx, 1:Nbuf, nnz-(Nbuf-1):nnz,1:N_vit)
  ar36_e(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:nnx, 1:Nbuf, nnz-(Nbuf-1):nnz        )
#if MHD==1
  ar36_e(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:nnx, 1:Nbuf, nnz-(Nbuf-1):nnz,1:N_vit)
#endif  
#if SCALP==1
  ar36_e(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:nnx, 1:Nbuf, nnz-(Nbuf-1):nnz,1:nfx)
#endif  

  ar46_e(1:nnx, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz        )
  ar46_e(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:N_vit)
  ar46_e(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz        )
#if MHD==1
  ar46_e(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:N_vit)
#endif  
#if SCALP==1
  ar46_e(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:nfx)
#endif  

 

  ! 8 corners
  coin135_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:Nbuf, 1:Nbuf, 1:Nbuf        ) 
  coin135_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:Nbuf, 1:Nbuf, 1:Nbuf,1:N_vit) 
  coin135_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:Nbuf, 1:Nbuf, 1:Nbuf        ) 
#if MHD==1
  coin135_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:Nbuf, 1:Nbuf, 1:Nbuf,1:N_vit) 
#endif  
#if SCALP==1
  coin135_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:Nbuf, 1:Nbuf, 1:Nbuf,1:nfx) 
#endif  

  coin136_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:Nbuf, 1:Nbuf, nnz-(Nbuf-1):nnz        ) 
  coin136_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:Nbuf, 1:Nbuf, nnz-(Nbuf-1):nnz,1:N_vit) 
  coin136_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:Nbuf, 1:Nbuf, nnz-(Nbuf-1):nnz        ) 
#if MHD==1
  coin136_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:Nbuf, 1:Nbuf, nnz-(Nbuf-1):nnz,1:N_vit) 
#endif  
#if SCALP==1
  coin136_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:Nbuf, 1:Nbuf, nnz-(Nbuf-1):nnz,1:nfx) 
#endif  

  coin145_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:Nbuf, nny-(Nbuf-1):nny, 1:Nbuf        ) 
  coin145_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:Nbuf, nny-(Nbuf-1):nny, 1:Nbuf,1:N_vit) 
  coin145_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:Nbuf, nny-(Nbuf-1):nny, 1:Nbuf        ) 
#if MHD==1
  coin145_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:Nbuf, nny-(Nbuf-1):nny, 1:Nbuf,1:N_vit) 
#endif  
#if SCALP==1
  coin145_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:Nbuf, nny-(Nbuf-1):nny, 1:Nbuf,1:nfx) 
#endif  

  coin146_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (1:Nbuf, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz        ) 
  coin146_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(1:Nbuf, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:N_vit) 
  coin146_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (1:Nbuf, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz        ) 
#if MHD==1
  coin146_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (1:Nbuf, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:N_vit) 
#endif  
#if SCALP==1
  coin146_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (1:Nbuf, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:nfx) 
#endif  

  coin235_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:Nbuf        ) 
  coin235_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, 1:Nbuf, 1:Nbuf,1:N_vit) 
  coin235_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:Nbuf        ) 
#if MHD==1
  coin235_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:Nbuf,1:N_vit) 
#endif  
#if SCALP==1
  coin235_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, 1:Nbuf, 1:Nbuf,1:nfx) 
#endif  

  coin236_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (nnx-(Nbuf-1):nnx, 1:Nbuf, nnz-(Nbuf-1):nnz        ) 
  coin236_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, 1:Nbuf, nnz-(Nbuf-1):nnz,1:N_vit) 
  coin236_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, 1:Nbuf, nnz-(Nbuf-1):nnz        ) 
#if MHD==1
  coin236_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, 1:Nbuf, nnz-(Nbuf-1):nnz,1:N_vit) 
#endif  
#if SCALP==1
  coin236_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, 1:Nbuf, nnz-(Nbuf-1):nnz,1:nfx) 
#endif  

  coin245_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:Nbuf        ) 
  coin245_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:Nbuf,1:N_vit) 
  coin245_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:Nbuf        ) 
#if MHD==1
  coin245_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:Nbuf,1:N_vit) 
#endif  
#if SCALP==1
  coin245_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, 1:Nbuf,1:nfx) 
#endif  

  coin246_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) = rho (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz        ) 
  coin246_e(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) = rhou(nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:N_vit) 
  coin246_e(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) = E   (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz        ) 
#if MHD==1
  coin246_e(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) = B   (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:N_vit) 
#endif  
#if SCALP==1
  coin246_e(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) = fx  (nnx-(Nbuf-1):nnx, nny-(Nbuf-1):nny, nnz-(Nbuf-1):nnz,1:nfx) 
#endif  

  ! send data

  ! Faces
  destination = voisin(-1, 0, 0)
  call MPI_Isend(face1_e ,ic_f12, MPI_DOUBLE_PRECISION,destination,F1 ,MPI_Comm_world,Req( 1),ierr)

  destination = voisin(+1, 0, 0)
  call MPI_Isend(face2_e ,ic_f12, MPI_DOUBLE_PRECISION,destination,F2 ,MPI_COMM_WORLD,Req( 2),ierr)

  destination = voisin( 0,-1, 0)
  call MPI_Isend(face3_e ,ic_f34, MPI_DOUBLE_PRECISION,destination,F3 ,MPI_COMM_WORLD,Req( 3),ierr)

  destination = voisin( 0,+1, 0)
  call MPI_Isend(face4_e ,ic_f34, MPI_DOUBLE_PRECISION,destination,F4 ,MPI_COMM_WORLD,Req( 4),ierr)

  destination = voisin( 0, 0,-1)
  call MPI_Isend(face5_e ,ic_f56, MPI_DOUBLE_PRECISION,destination,F5 ,MPI_COMM_WORLD,Req( 5),ierr)

  destination = voisin( 0, 0,+1)
  call MPI_Isend(face6_e ,ic_f56, MPI_DOUBLE_PRECISION,destination,F6 ,MPI_COMM_WORLD,Req( 6),ierr)



  ! Edges
  destination = voisin(-1,-1,0)
  call MPI_Isend(ar13_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR13   ,MPI_COMM_WORLD,Req( 7),ierr)

  destination = voisin(-1,+1,0)
  call MPI_Isend(ar14_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR14   ,MPI_COMM_WORLD,Req( 8),ierr)

  destination = voisin(+1,-1,0)
  call MPI_Isend(ar23_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR23   ,MPI_COMM_WORLD,Req( 9),ierr)

  destination = voisin(+1,+1,0)
  call MPI_Isend(ar24_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR24   ,MPI_COMM_WORLD,Req(10),ierr)


  destination = voisin(-1, 0,-1)
  call MPI_Isend(ar15_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR15   ,MPI_COMM_WORLD,Req(11),ierr)

  destination = voisin(-1, 0,+1)
  call MPI_Isend(ar16_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR16   ,MPI_COMM_WORLD,Req(12),ierr)

  destination = voisin(+1, 0,-1)
  call MPI_Isend(ar25_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR25   ,MPI_COMM_WORLD,Req(13),ierr)

  destination = voisin(+1, 0,+1)
  call MPI_Isend(ar26_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR26   ,MPI_COMM_WORLD,Req(14),ierr)


  destination = voisin( 0,-1,-1)
  call MPI_Isend(ar35_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR35   ,MPI_COMM_WORLD,Req(15),ierr)

  destination = voisin( 0,-1,+1)
  call MPI_Isend(ar36_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR36   ,MPI_COMM_WORLD,Req(16),ierr)

  destination = voisin( 0,+1,-1)
  call MPI_Isend(ar45_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR45   ,MPI_COMM_WORLD,Req(17),ierr)

  destination = voisin( 0,+1,+1)
  call MPI_Isend(ar46_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR46   ,MPI_COMM_WORLD,Req(18),ierr)


  ! Corners
  destination = voisin( -1,-1,-1)
  call MPI_Isend(coin135_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C135   ,MPI_COMM_WORLD,Req(19),ierr)

  destination = voisin( -1,-1,+1)
  call MPI_Isend(coin136_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C136   ,MPI_COMM_WORLD,Req(20),ierr)

  destination = voisin( -1,+1,-1)
  call MPI_Isend(coin145_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C145   ,MPI_COMM_WORLD,Req(21),ierr)

  destination = voisin( -1,+1,+1)
  call MPI_Isend(coin146_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C146   ,MPI_COMM_WORLD,Req(22),ierr)

  destination = voisin( +1,-1,-1)
  call MPI_Isend(coin235_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C235   ,MPI_COMM_WORLD,Req(23),ierr)

  destination = voisin( +1,-1,+1)
  call MPI_Isend(coin236_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C236   ,MPI_COMM_WORLD,Req(24),ierr)

  destination = voisin( +1,+1,-1)
  call MPI_Isend(coin245_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C245   ,MPI_COMM_WORLD,Req(25),ierr)

  destination = voisin( +1,+1,+1)
  call MPI_Isend(coin246_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C246   ,MPI_COMM_WORLD,Req(26),ierr)

 

  ! receive data
  

  ! Faces
  emetteur = voisin(+1, 0, 0)
  call MPI_Irecv(face2_r ,ic_f12, MPI_DOUBLE_PRECISION,emetteur,F1 ,MPI_Comm_world,Req(27),ierr)

  emetteur = voisin(-1, 0, 0)
  call MPI_Irecv(face1_r ,ic_f12, MPI_DOUBLE_PRECISION,emetteur,F2 ,MPI_COMM_WORLD,Req(28),ierr)

  emetteur = voisin( 0,+1, 0)
  call MPI_Irecv(face4_r ,ic_f34, MPI_DOUBLE_PRECISION,emetteur,F3 ,MPI_COMM_WORLD,Req(29),ierr)

  emetteur = voisin( 0,-1, 0)
  call MPI_Irecv(face3_r ,ic_f34, MPI_DOUBLE_PRECISION,emetteur,F4 ,MPI_COMM_WORLD,Req(30),ierr)

  emetteur = voisin( 0, 0,+1)
  call MPI_Irecv(face6_r ,ic_f56, MPI_DOUBLE_PRECISION,emetteur,F5 ,MPI_COMM_WORLD,Req(31),ierr)

  emetteur = voisin( 0, 0,-1)
  call MPI_Irecv(face5_r ,ic_f56, MPI_DOUBLE_PRECISION,emetteur,F6 ,MPI_COMM_WORLD,Req(32),ierr)



  ! Edges
  emetteur = voisin(+1,+1,0)
  call MPI_Irecv(ar24_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR13   ,MPI_COMM_WORLD,Req(33),ierr)

  emetteur = voisin(+1,-1,0)
  call MPI_Irecv(ar23_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR14   ,MPI_COMM_WORLD,Req(34),ierr)

  emetteur = voisin(-1,+1,0)
  call MPI_Irecv(ar14_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR23   ,MPI_COMM_WORLD,Req(35),ierr)

  emetteur = voisin(-1,-1,0)
  call MPI_Irecv(ar13_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR24   ,MPI_COMM_WORLD,Req(36),ierr)


  emetteur = voisin(+1, 0,+1)
  call MPI_Irecv(ar26_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR15   ,MPI_COMM_WORLD,Req(37),ierr)

  emetteur = voisin(+1, 0,-1)
  call MPI_Irecv(ar25_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR16   ,MPI_COMM_WORLD,Req(38),ierr)

  emetteur = voisin(-1, 0,+1)
  call MPI_Irecv(ar16_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR25   ,MPI_COMM_WORLD,Req(39),ierr)

  emetteur = voisin(-1, 0,-1)
  call MPI_Irecv(ar15_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR26   ,MPI_COMM_WORLD,Req(40),ierr)


  emetteur = voisin( 0,+1,+1)
  call MPI_Irecv(ar46_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR35   ,MPI_COMM_WORLD,Req(41),ierr)

  emetteur = voisin( 0,+1,-1)
  call MPI_Irecv(ar45_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR36   ,MPI_COMM_WORLD,Req(42),ierr)

  emetteur = voisin( 0,-1,+1)
  call MPI_Irecv(ar36_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR45   ,MPI_COMM_WORLD,Req(43),ierr)

  emetteur = voisin( 0,-1,-1)
  call MPI_Irecv(ar35_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR46   ,MPI_COMM_WORLD,Req(44),ierr)


  ! Corners
  emetteur = voisin( +1,+1,+1)
  call MPI_Irecv(coin246_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C135   ,MPI_COMM_WORLD,Req(45),ierr)

  emetteur = voisin( +1,+1,-1)
  call MPI_Irecv(coin245_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C136   ,MPI_COMM_WORLD,Req(46),ierr)

  emetteur = voisin( +1,-1,+1)
  call MPI_Irecv(coin236_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C145   ,MPI_COMM_WORLD,Req(47),ierr)

  emetteur = voisin( +1,-1,-1)
  call MPI_Irecv(coin235_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C146   ,MPI_COMM_WORLD,Req(48),ierr)

  emetteur = voisin( -1,+1,+1)
  call MPI_Irecv(coin146_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C235   ,MPI_COMM_WORLD,Req(49),ierr)

  emetteur = voisin( -1,+1,-1)
  call MPI_Irecv(coin145_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C236   ,MPI_COMM_WORLD,Req(50),ierr)

  emetteur = voisin( -1,-1,+1)
  call MPI_Irecv(coin136_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C245   ,MPI_COMM_WORLD,Req(51),ierr)

  emetteur = voisin( -1,-1,-1)
  call MPI_Irecv(coin135_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C246   ,MPI_COMM_WORLD,Req(52),ierr)

   
  call MPI_Waitall(52,Req,Stat,ierr)

  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  ! 6 faces
  rho (1-Nbuf:0,1:nny,1:nnz        ) = face1_r(1:Nbuf,1:nny,1:nnz,1                ) 
  rhou(1-Nbuf:0,1:nny,1:nnz,1:N_vit) = face1_r(1:Nbuf,1:nny,1:nnz,2:N_vit+1        )  
  E   (1-Nbuf:0,1:nny,1:nnz        ) = face1_r(1:Nbuf,1:nny,1:nnz,  N_vit+2        ) 
#if MHD==1
  B   (1-Nbuf:0,1:nny,1:nnz,1:N_vit) = face1_r(1:Nbuf,1:nny,1:nnz,N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (1-Nbuf:0,1:nny,1:nnz,1:nfx) = face1_r(1:Nbuf,1:nny,1:nnz,nvar-nfx+1:nvar) 
#endif  

  rho (nnx+1:nnx+Nbuf,1:nny,1:nnz        ) = face2_r(1:Nbuf,1:nny,1:nnz,1                ) 
  rhou(nnx+1:nnx+Nbuf,1:nny,1:nnz,1:N_vit) = face2_r(1:Nbuf,1:nny,1:nnz,2:N_vit+1        ) 
  E   (nnx+1:nnx+Nbuf,1:nny,1:nnz        ) = face2_r(1:Nbuf,1:nny,1:nnz,  N_vit+2        ) 
#if MHD==1
  B   (nnx+1:nnx+Nbuf,1:nny,1:nnz,1:N_vit) = face2_r(1:Nbuf,1:nny,1:nnz,N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf,1:nny,1:nnz,1:nfx) = face2_r(1:Nbuf,1:nny,1:nnz,nvar-nfx+1:nvar) 
#endif  

  rho (1:nnx,1-Nbuf:0,1:nnz        ) = face3_r(1:nnx,1:Nbuf,1:nnz,1                )  
  rhou(1:nnx,1-Nbuf:0,1:nnz,1:N_vit) = face3_r(1:nnx,1:Nbuf,1:nnz,2:N_vit+1        )  
  E   (1:nnx,1-Nbuf:0,1:nnz        ) = face3_r(1:nnx,1:Nbuf,1:nnz,  N_vit+2        )  
#if MHD==1
  B   (1:nnx,1-Nbuf:0,1:nnz,1:N_vit) = face3_r(1:nnx,1:Nbuf,1:nnz,N_vit+3:2*N_vit+2)  
#endif  
#if SCALP==1
  fx  (1:nnx,1-Nbuf:0,1:nnz,1:nfx) = face3_r(1:nnx,1:Nbuf,1:nnz,nvar-nfx+1:nvar)  
#endif  

  rho (1:nnx,nny+1:nny+Nbuf,1:nnz        ) = face4_r(1:nnx,1:Nbuf,1:nnz,1                ) 
  rhou(1:nnx,nny+1:nny+Nbuf,1:nnz,1:N_vit) = face4_r(1:nnx,1:Nbuf,1:nnz,2:N_vit+1        ) 
  E   (1:nnx,nny+1:nny+Nbuf,1:nnz        ) = face4_r(1:nnx,1:Nbuf,1:nnz,  N_vit+2        ) 
#if MHD==1
  B   (1:nnx,nny+1:nny+Nbuf,1:nnz,1:N_vit) = face4_r(1:nnx,1:Nbuf,1:nnz,N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (1:nnx,nny+1:nny+Nbuf,1:nnz,1:nfx) = face4_r(1:nnx,1:Nbuf,1:nnz,nvar-nfx+1:nvar) 
#endif  
  
  rho (1:nnx,1:nny,1-Nbuf:0        ) = face5_r(1:nnx,1:nny,1:Nbuf,1                ) 
  rhou(1:nnx,1:nny,1-Nbuf:0,1:N_vit) = face5_r(1:nnx,1:nny,1:Nbuf,2:N_vit+1        ) 
  E   (1:nnx,1:nny,1-Nbuf:0        ) = face5_r(1:nnx,1:nny,1:Nbuf,  N_vit+2        ) 
#if MHD==1
  B   (1:nnx,1:nny,1-Nbuf:0,1:N_vit) = face5_r(1:nnx,1:nny,1:Nbuf,N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (1:nnx,1:nny,1-Nbuf:0,1:nfx) = face5_r(1:nnx,1:nny,1:Nbuf,nvar-nfx+1:nvar) 
#endif  

  rho (1:nnx,1:nny,nnz+1:nnz+Nbuf        ) = face6_r(1:nnx,1:nny,1:Nbuf,1                ) 
  rhou(1:nnx,1:nny,nnz+1:nnz+Nbuf,1:N_vit) = face6_r(1:nnx,1:nny,1:Nbuf,2:N_vit+1        ) 
  E   (1:nnx,1:nny,nnz+1:nnz+Nbuf        ) = face6_r(1:nnx,1:nny,1:Nbuf,  N_vit+2        ) 
#if MHD==1
  B   (1:nnx,1:nny,nnz+1:nnz+Nbuf,1:N_vit) = face6_r(1:nnx,1:nny,1:Nbuf,N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (1:nnx,1:nny,nnz+1:nnz+Nbuf,1:nfx) = face6_r(1:nnx,1:nny,1:Nbuf,nvar-nfx+1:nvar) 
#endif  
  
  ! 12 edges
  rho (1-Nbuf:0, 1-Nbuf:0, 1:nnz        ) = ar13_r(1:Nbuf, 1:Nbuf, 1:nnz, 1                )
  rhou(1-Nbuf:0, 1-Nbuf:0, 1:nnz,1:N_vit) = ar13_r(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         )
  E   (1-Nbuf:0, 1-Nbuf:0, 1:nnz        ) = ar13_r(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         )
#if MHD==1
  B   (1-Nbuf:0, 1-Nbuf:0, 1:nnz,1:N_vit) = ar13_r(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1-Nbuf:0, 1-Nbuf:0, 1:nnz,1:nfx ) = ar13_r(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar)
#endif  

  rho (1-Nbuf:0, nny+1:nny+Nbuf, 1:nnz        ) = ar14_r(1:Nbuf, 1:Nbuf, 1:nnz, 1                )
  rhou(1-Nbuf:0, nny+1:nny+Nbuf, 1:nnz,1:N_vit) = ar14_r(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         )
  E   (1-Nbuf:0, nny+1:nny+Nbuf, 1:nnz        ) = ar14_r(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         )  
#if MHD==1
  B   (1-Nbuf:0, nny+1:nny+Nbuf, 1:nnz,1:N_vit) = ar14_r(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1-Nbuf:0, nny+1:nny+Nbuf, 1:nnz,1:nfx ) = ar14_r(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar)
#endif  

  rho (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1:nnz        ) = ar23_r(1:Nbuf, 1:Nbuf, 1:nnz, 1                )
  rhou(nnx+1:nnx+Nbuf, 1-Nbuf:0, 1:nnz,1:N_vit) = ar23_r(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         )
  E   (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1:nnz        ) = ar23_r(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         )
#if MHD==1
  B   (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1:nnz,1:N_vit) = ar23_r(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1:nnz,1:nfx ) = ar23_r(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar)
#endif  

   rho (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1:nnz        ) = ar24_r(1:Nbuf, 1:Nbuf, 1:nnz, 1                )
   rhou(nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1:nnz,1:N_vit) = ar24_r(1:Nbuf, 1:Nbuf, 1:nnz,2:N_vit+1         )
   E   (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1:nnz        ) = ar24_r(1:Nbuf, 1:Nbuf, 1:nnz,  N_vit+2         )
#if MHD==1
   B   (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1:nnz,1:N_vit) = ar24_r(1:Nbuf, 1:Nbuf, 1:nnz, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
   fx  (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1:nnz,1:nfx ) = ar24_r(1:Nbuf, 1:Nbuf, 1:nnz, nvar-nfx+1:nvar)
#endif  

  rho (1-Nbuf:0, 1:nny, 1-Nbuf:0        ) = ar15_r(1:Nbuf, 1:nny, 1:Nbuf, 1                ) 
  rhou(1-Nbuf:0, 1:nny, 1-Nbuf:0,1:N_vit) = ar15_r(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         ) 
  E   (1-Nbuf:0, 1:nny, 1-Nbuf:0        ) = ar15_r(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         ) 
#if MHD==1
  B   (1-Nbuf:0, 1:nny, 1-Nbuf:0,1:N_vit) = ar15_r(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (1-Nbuf:0, 1:nny, 1-Nbuf:0,1:nfx ) = ar15_r(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar) 
#endif  

   rho (nnx+1:nnx+Nbuf, 1:nny, 1-Nbuf:0        ) = ar25_r(1:Nbuf, 1:nny, 1:Nbuf, 1                )
   rhou(nnx+1:nnx+Nbuf, 1:nny, 1-Nbuf:0,1:N_vit) = ar25_r(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         )
   E   (nnx+1:nnx+Nbuf, 1:nny, 1-Nbuf:0        ) = ar25_r(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         )
#if MHD==1
   B   (nnx+1:nnx+Nbuf, 1:nny, 1-Nbuf:0,1:N_vit) = ar25_r(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
   fx  (nnx+1:nnx+Nbuf, 1:nny, 1-Nbuf:0,1:nfx ) = ar25_r(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

   rho (1-Nbuf:0, 1:nny, nnz+1:nnz+Nbuf        ) = ar16_r(1:Nbuf, 1:nny, 1:Nbuf, 1                )
   rhou(1-Nbuf:0, 1:nny, nnz+1:nnz+Nbuf,1:N_vit) = ar16_r(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         )
   E   (1-Nbuf:0, 1:nny, nnz+1:nnz+Nbuf        ) = ar16_r(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         )
#if MHD==1
   B   (1-Nbuf:0, 1:nny, nnz+1:nnz+Nbuf,1:N_vit) = ar16_r(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
   fx  (1-Nbuf:0, 1:nny, nnz+1:nnz+Nbuf,1:nfx ) = ar16_r(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

  rho (nnx+1:nnx+Nbuf, 1:nny, nnz+1:nnz+Nbuf        ) = ar26_r(1:Nbuf, 1:nny, 1:Nbuf, 1                )
  rhou(nnx+1:nnx+Nbuf, 1:nny, nnz+1:nnz+Nbuf,1:N_vit) = ar26_r(1:Nbuf, 1:nny, 1:Nbuf,2:N_vit+1         )
  E   (nnx+1:nnx+Nbuf, 1:nny, nnz+1:nnz+Nbuf        ) = ar26_r(1:Nbuf, 1:nny, 1:Nbuf,  N_vit+2         )
#if MHD==1
  B   (nnx+1:nnx+Nbuf, 1:nny, nnz+1:nnz+Nbuf,1:N_vit) = ar26_r(1:Nbuf, 1:nny, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf, 1:nny, nnz+1:nnz+Nbuf,1:nfx ) = ar26_r(1:Nbuf, 1:nny, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

  rho (1:nnx, 1-Nbuf:0, 1-Nbuf:0        ) = ar35_r(1:nnx, 1:Nbuf, 1:Nbuf, 1                )
  rhou(1:nnx, 1-Nbuf:0, 1-Nbuf:0,1:N_vit) = ar35_r(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         )
  E   (1:nnx, 1-Nbuf:0, 1-Nbuf:0        ) = ar35_r(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         )
#if MHD==1
  B   (1:nnx, 1-Nbuf:0, 1-Nbuf:0,1:N_vit) = ar35_r(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1:nnx, 1-Nbuf:0, 1-Nbuf:0,1:nfx ) = ar35_r(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

   rho (1:nnx, nny+1:nny+Nbuf, 1-Nbuf:0        ) = ar45_r(1:nnx, 1:Nbuf, 1:Nbuf, 1                ) 
   rhou(1:nnx, nny+1:nny+Nbuf, 1-Nbuf:0,1:N_vit) = ar45_r(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) 
   E   (1:nnx, nny+1:nny+Nbuf, 1-Nbuf:0        ) = ar45_r(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         ) 
#if MHD==1
   B   (1:nnx, nny+1:nny+Nbuf, 1-Nbuf:0,1:N_vit) = ar45_r(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
   fx  (1:nnx, nny+1:nny+Nbuf, 1-Nbuf:0,1:nfx ) = ar45_r(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) 
#endif  

   rho (1:nnx, 1-Nbuf:0, nnz+1:nnz+Nbuf        ) = ar36_r(1:nnx, 1:Nbuf, 1:Nbuf, 1                 ) 
   rhou(1:nnx, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:N_vit) = ar36_r(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1          ) 
   E   (1:nnx, 1-Nbuf:0, nnz+1:nnz+Nbuf       ) = ar36_r(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2           ) 
#if MHD==1
   B   (1:nnx, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:N_vit ) = ar36_r(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
   fx  (1:nnx, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:nfx ) = ar36_r(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) 
#endif  

  rho (1:nnx, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf        ) = ar46_r(1:nnx, 1:Nbuf, 1:Nbuf, 1                )
  rhou(1:nnx, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:N_vit) = ar46_r(1:nnx, 1:Nbuf, 1:Nbuf,2:N_vit+1         )
  E   (1:nnx, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf        ) = ar46_r(1:nnx, 1:Nbuf, 1:Nbuf,  N_vit+2         )
#if MHD==1
  B   (1:nnx, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:N_vit) = ar46_r(1:nnx, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1:nnx, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:nfx ) = ar46_r(1:nnx, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

 
  ! 8 corners
   rho (1-Nbuf:0, 1-Nbuf:0, 1-Nbuf:0       ) = coin135_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                 )  
   rhou(1-Nbuf:0, 1-Nbuf:0, 1-Nbuf:0,1:N_vit) = coin135_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         )  
   E   (1-Nbuf:0, 1-Nbuf:0, 1-Nbuf:0       ) = coin135_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2          )  
#if MHD==1
   B   (1-Nbuf:0, 1-Nbuf:0, 1-Nbuf:0,1:N_vit) = coin135_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)  
#endif  
#if SCALP==1
   fx  (1-Nbuf:0, 1-Nbuf:0, 1-Nbuf:0,1:nfx) = coin135_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)  
#endif  

  rho (1-Nbuf:0, 1-Nbuf:0, nnz+1:nnz+Nbuf       ) = coin136_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                 )
  rhou(1-Nbuf:0, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:N_vit) = coin136_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         )
  E   (1-Nbuf:0, 1-Nbuf:0, nnz+1:nnz+Nbuf       ) = coin136_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2          )
#if MHD==1
  B   (1-Nbuf:0, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:N_vit) = coin136_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (1-Nbuf:0, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:nfx) = coin136_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

   rho (1-Nbuf:0, nny+1:nny+Nbuf, 1-Nbuf:0       ) = coin145_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                 )  
   rhou(1-Nbuf:0, nny+1:nny+Nbuf, 1-Nbuf:0,1:N_vit) = coin145_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         )  
   E   (1-Nbuf:0, nny+1:nny+Nbuf, 1-Nbuf:0       ) = coin145_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2          )  
#if MHD==1
   B   (1-Nbuf:0, nny+1:nny+Nbuf, 1-Nbuf:0,1:N_vit) = coin145_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)  
#endif  
#if SCALP==1
   fx  (1-Nbuf:0, nny+1:nny+Nbuf, 1-Nbuf:0,1:nfx) = coin145_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)  
#endif  
     
   rho (1-Nbuf:0, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf        ) = coin146_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) 
   rhou(1-Nbuf:0, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:N_vit) = coin146_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) 
   E   (1-Nbuf:0, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf        ) = coin146_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) 
#if MHD==1
   B   (1-Nbuf:0, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:N_vit) = coin146_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
   fx  (1-Nbuf:0, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:nfx) = coin146_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) 
#endif  

  rho (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1-Nbuf:0        ) = coin235_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) 
  rhou(nnx+1:nnx+Nbuf, 1-Nbuf:0, 1-Nbuf:0,1:N_vit) = coin235_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) 
  E   (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1-Nbuf:0        ) = coin235_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) 
#if MHD==1
  B   (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1-Nbuf:0,1:N_vit) = coin235_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf, 1-Nbuf:0, 1-Nbuf:0,1:nfx) = coin235_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) 
#endif  

   rho (nnx+1:nnx+Nbuf, 1-Nbuf:0, nnz+1:nnz+Nbuf        ) = coin236_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) 
   rhou(nnx+1:nnx+Nbuf, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:N_vit) = coin236_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) 
   E   (nnx+1:nnx+Nbuf, 1-Nbuf:0, nnz+1:nnz+Nbuf        ) = coin236_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         )  
#if MHD==1
   B   (nnx+1:nnx+Nbuf, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:N_vit) = coin236_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)  
#endif  
#if SCALP==1
   fx  (nnx+1:nnx+Nbuf, 1-Nbuf:0, nnz+1:nnz+Nbuf,1:nfx) = coin236_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)  
#endif  

  rho (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1-Nbuf:0        ) = coin245_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                ) 
  rhou(nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1-Nbuf:0,1:N_vit) = coin245_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         ) 
  E   (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1-Nbuf:0        ) = coin245_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         ) 
#if MHD==1
  B   (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1-Nbuf:0,1:N_vit) = coin245_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2) 
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, 1-Nbuf:0,1:nfx) = coin245_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar) 
#endif  

  rho (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf        ) = coin246_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1                )
  rhou(nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:N_vit) = coin246_r(1:Nbuf, 1:Nbuf, 1:Nbuf,2:N_vit+1         )
  E   (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf        ) = coin246_r(1:Nbuf, 1:Nbuf, 1:Nbuf,  N_vit+2         )
#if MHD==1
  B   (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:N_vit) = coin246_r(1:Nbuf, 1:Nbuf, 1:Nbuf, N_vit+3:2*N_vit+2)
#endif  
#if SCALP==1
  fx  (nnx+1:nnx+Nbuf, nny+1:nny+Nbuf, nnz+1:nnz+Nbuf,1:nfx) = coin246_r(1:Nbuf, 1:Nbuf, 1:Nbuf, nvar-nfx+1:nvar)
#endif  

  if(verbose) write(*,*) 'End of com_3d'

  return
  
end subroutine com_3D

#endif
