!> \file
!! Contains subroutine pasdt()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PASDT
!
!> Compute maximum allowed timestep from various
!! stability criteria (CFL, MHD, RHD, gravity, etc...)
!<
subroutine pasdt(calculate_dt)

  use parameters
  use divers     , only : verbose,dtold,dt,dt_glob,dt_mhd,dt_max,dt_cfl,dt_ff,CONDUC,RAY,HYDRO,DIFFU,GRAV,terminal_output,noutput
  use divers_ray , only : dt_ray
  use Conduction , only : dt_con,implicit_con,dt_con_exp,dt_con_imp
  use Diffusion  , only : dt_dif
  use unites     , only : an,seconde
  use para       , only : mype

  implicit none

#if PARA==1
  include 'mpif.h'
#endif

#if EOS==1
  real :: Sound_speed,dens
#endif

  ! Local variables
  real    :: dt_ini
  logical :: calculate_dt
  
  if (verbose) write(*,*) 'Start of pasdt'
  
  if(.not.calculate_dt)then
     if (verbose) write(*,*) 'End of pasdt - dt not computed'
     return
  endif

  dt_ini = 1.0E-20 * seconde

  dtold = dt

  if (HYDRO) then
     call pasdt_hydro
     if (dt_cfl == 0.) dt_cfl = dt_ini
     if (mype == 0)then
        noutput = noutput + 1
        write(terminal_output(noutput),003 ) dt_cfl/seconde
     endif
003 format('dt_cfl (s) =',1pe10.3)
  endif

#if RAYO==1
  if (RAY) then
     call pasdt_ray
     if (mype == 0)then
        noutput = noutput + 1
        write(terminal_output(noutput),004 ) dt_ray/seconde
     endif
004 format('dt_ray (s) =',1pe10.3)
  endif
#endif

#if CON==1
  if (CONDUC) then
     call pasdt_con
     if (dt_con == 0.) dt_con = dt_ini
     if (mype == 0)then
        noutput = noutput + 1
        write(terminal_output(noutput),005) dt_con_exp/seconde,dt_con_imp/seconde
     endif
005 format('dt_con (s) =',1pe10.3,2x,E10.3,'   (exp,imp)')
  endif
#endif

#if DIF==1
  if (DIFFU) then
     call pasdt_dif
     if (dt_dif == 0.) dt_dif = dt_ini
     if (mype == 0)then
        noutput = noutput + 1
        write(terminal_output(noutput),006) dt_dif/seconde
     endif
006 format('dt_dif (s) =',1pe10.3)
  endif
#endif

#if MHD==1
  call pasdt_mhd
  if (dt_mhd == 0.) dt_mhd = dt_ini
  if (mype == 0)then
     noutput = noutput + 1
     write(terminal_output(noutput),007) dt_mhd/seconde
  endif
007 format('dt_mhd (s) =',1pe10.3)
#endif

#if GRA==1
  if (GRAV) then
     call pasdt_grav
     if (dt_ff == 0.) dt_ff = dt_ini
     if (mype == 0)then
        noutput = noutput + 1
        write(terminal_output(noutput),008) dt_ff/seconde
     endif
008 format('dt_ff  (s) =',1pe10.3)
  endif
#endif


  dt = 1.0E50 * an
  
  if (HYDRO ) dt = min(dt,dt_cfl)
  if (CONDUC) dt = min(dt,dt_con)
  if (RAY   ) dt = min(dt,dt_ray)
  if (DIFFU ) dt = min(dt,dt_dif)
  if (GRAV  ) dt = min(dt,dt_ff )
#if MHD==1
              dt = min(dt,dt_mhd)
#endif
      
  dt = min(dt,dt_max)
  
  if (mype == 0)then
     noutput = noutput + 1
     write(terminal_output(noutput),009) dt/seconde
  endif
009 format('dt     (s) =',1pe10.3)

  if (verbose) write(*,*) 'End of pasdt'

  return
  
end subroutine pasdt
