!> \file
!! Contains subroutines read_params() and default_params()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine READ_PARAMS
!
!> Reads in all the simulation parameters from the
!! namelist.
!<
subroutine read_params(kmax,format_sortie)

  use parameters
  use divers
  use cl
  use cl_ray
  use cl_con
  use cl_dif
  use geom       , only : x,shift_gr,nshift_gr,x_glob, Cylindrique, Spherique, Cartesien, &
                        & geometrie,Lbox, geom_dir
  use para
  use mfilm      , only : FILM
  use cputime
  use gravity

  implicit none

  integer          :: kmax
  character(len=3) :: format_sortie

  !namelist for input parameters
  namelist/param/reprise,Nimp,Nout,verbose,tend,dt_max,slope_type,Solver, &
                 kmax,REF,DETO,nfx,geometrie,CONDUC,HYDRO,RAY,DIFFU,GRAV, &
                 HISTOGRAMME,FILM,FORMAT_SORTIE,STIR,Tcpu_max,i_restart,  &
                 ISOTHERME, riemann, riemann2d,slope_type
  namelist/param_grille/ndim,nx_glob,icl_glob,icl_ray_glob,icl_dif_glob,icl_con_glob,nsx,nsy,nsz,Nbuf
  namelist/job/name_of_job,comments
  namelist/repr/nrepr,trepr
  namelist/output/tout,ndata,datanames,dataunits,nsupp,suppnames,suppunits,suppdim

  if (mype == 0) then
     write(*,'(a)') '##########################################################################'
     write(*,'(a)') 'Initialising simulation parameters:'
  endif

  !Default value for input parameters
  call default_params(kmax,format_sortie)

  !Reading input parameters in file 'liste.nml'
  open(1,file='liste.nml')
  
  if(mype == 0) write(*,'(a)') 'reading namelist PARAMS'
  rewind(1) ; read(1,NML=param       )
  
  if(mype == 0) write(*,'(a)') 'reading namelist PARAM_GRILLE'
  rewind(1) ; read(1,NML=param_grille)
  
  if(mype == 0) write(*,'(a)') 'reading namelist JOB'
  rewind(1) ; read(1,NML=job         )
  
  if(mype == 0) write(*,'(a)') 'reading namelist REPR'
  rewind(1) ; read(1,NML=repr        )
  
  if(mype == 0) write(*,'(a)') 'reading namelist OUTPUT'
  rewind(1) ; read(1,NML=output      )
  
  close(1)
  
  if(mype == 0)then
     write(*,'(a)') 'Finished reading namelists'
     write(*,'(a)') '##########################################################################'
  endif

  Tcpu_max = Tcpu_max*60.

  temps_ref    = 0.
  temps_ray    = 0.
  temps_dif    = 0.
  temps_conduc = 0.
  temps_hydro  = 0.
  temps_comm   = 0.
  temps_io     = 0.
  temps_grav   = 0.

  return
  
end subroutine read_params

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine DEFAULT_PARAMS
!
!> Sets the simulation parameters to default values in
!! case they are not set in the namelist.
!<
subroutine default_params(kmax,format_sortie)

  use parameters
  use divers
  use divers_ray , only : ngrp_ray
  use diffusion  , only : ngrp_dif
  use cl
  use cl_ray
  use cl_con
  use cl_dif
  use geom       , only : x,shift_gr,nshift_gr,x_glob, Cylindrique, Spherique, Cartesien, &
                        & geometrie,Lbox, geom_dir
  use para
  use mfilm      , only : FILM
  use gravity

  implicit none

  integer          :: kmax
  character(len=3) :: format_sortie

  !local variables
  integer :: i

  !Default values for "cpu" namelist
  ncpu_x = 1
  ncpu_y = 1
  ncpu_z = 1

  !Default values for "param" namelist
  reprise       = .false.
  Nimp          =  10
  Nout          =   1
  verbose       = .false.
  tend          = 1.
  dt_max        =  1.e-5
  slope_type    = 1        ! 1=Minmod, 2=Moncen, 3=Positivity preserving, 4=Harmonic average, 5=VAVL, 6=Superbee
  Solver        = 'exact'  !'exact','acoustic','relax','eos'
  kmax          = 1000000
  REF           = .false.
  DETO          = .false.
  nfx           = 0        !Number of passive scalar
  geometrie     ='x'
  CONDUC        = .false.
  HYDRO         = .true.
  RAY           = .false.
  DIFFU         = .false.
  GRAV          = .false.
  HISTOGRAMME   = .false.
  FILM          = .false.
  format_sortie = 'bin'
  Tcpu_max      = 100       !Stop code execution after 100 mn
  i_restart     = 0
  ISOTHERME     = .false.
  STIR          = .false.
  ! Default values for the MHD solver 
  riemann       = 'hlld' ! Only possible option for the moment
  riemann2d     = 'hlld' ! hlld, hllf, hllq, roe,llf


  !Default values for "param_grille" namelist
  ndim         = 1
  nx_glob      = (/ 200,1,1 /)
  icl_glob     = (/ 2,2,2,2,2,2 /)
  icl_ray_glob = (/ 2,2,2,2,2,2 /)
  icl_dif_glob = (/ 2,2,2,2,2,2 /)
  icl_con_glob = (/ 2,2,2,2,2,2 /)
  nsx          = 1
  nsy          = 1
  nsz          = 1
  Nbuf         = 2

  !Default values for "param_grille" namelist
  name_of_job = "myjob"   !maximum 5 characters
  comments    = " job test " 

  !Default values for "repr" namelist
  nrepr      = 0
  trepr(1:2) = (/ 10.,20. /)

  !Default values for "output" namelist
  tout(1:3)    = (/ 5.,10.,15. /)
  ndata        = 1
  datanames(1) = "rho"
  dataunits    = (/ ("standard",i=1,20) /)
  nsupp        = 0
  suppnames    = (/ ("standard",i=1,5) /)
  suppunits    = (/ ("standard",i=1,5) /)
  suppdim      = 1

  !Default values for unknown namelist
  fourdim      = 0
  noutd        = 0  !no file output  _00 produced
  nreprd       = 0  !no file restart _00 produced

  !Default values for multigroup (overwritten by values in init_ray and init_diffusion)
  ngrp_ray = 0
  ngrp_dif = 0

  return

end subroutine default_params
