!> \file
!! Contains subroutine ctoprim()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CTOPRIM
!
!> Converts conservative variables to primitive in mhd.
!<
subroutine ctoprim(uin,q,bf,gravin,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  use parameters, only : smallr,smallc,n_vit
  use gammas
  use const
  use var_loc
  use geom
  
  implicit none

  integer :: i,j,k,idim
! integer :: n
  integer :: nvar,ndim
  integer :: iu1,iu2,ju1,ju2,ku1,ku2

  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar+3) :: uin
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:ndim  ) :: gravin
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar  ) :: q  
  real, dimension(iu1:iu2+1,ju1:ju2+1,ku1:ku2+1,1:3     ) :: bf
  real :: eint,smalle,smallp,eken,emag,dt
  
#ifdef ISO
  smalle = 1.
  smallp = smallr*ciso**2
#else
  smalle = smallc**2/gamma/(gamma-one)
  smallp = smallr*smallc**2/gamma
#endif

  ! Store face centered magnetic field
  do k = ku1, ku2
     do j = ju1, ju2
        do i = iu1, iu2+1
           if(i<=iu2)then
              bf(i,j,k,1) = uin(i,j,k,6)
           else
              bf(i,j,k,1) = uin(i-1,j,k,nvar+1)
           endif
        end do
     end do
  end do
  do k = ku1, ku2
     do j = ju1, ju2+1
        do i = iu1, iu2
           if(j<=ju2)then
              bf(i,j,k,2) = uin(i,j,k,7)
           else
              bf(i,j,k,2) = uin(i,j-1,k,nvar+2)
           endif
        end do
     end do
  end do
  do k = ku1, ku2+1
     do j = ju1, ju2
        do i = iu1, iu2
           if(k<=ku2)then
              bf(i,j,k,3) = uin(i,j,k,8)
           else
              bf(i,j,k,3) = uin(i,j,k-1,nvar+3)
           endif
        end do
     end do
  end do
  
  ! convert to primitive variable
  do k = ku1, ku2
     do j = ju1, ju2
        do i = iu1, iu2

           ! compute density
           q(i,j,k,1) = max(uin(i,j,k,1),smallr)

           ! compute velocities
           do idim = 1, n_vit
              q(i,j,k,idim+1) = uin(i,j,k,idim+1)/uin(i,j,k,1)
           end do

           ! compute specific kinetic energy
           eken = zero
           do idim = 1, n_vit
              eken = eken + half*q(i,j,k,idim+1)**2
           end do

           ! compute cell centered magnetic field
           q(i,j,k,n_vit+3) = half*(uin(i,j,k,n_vit+3)+uin(i,j,k,nvar+1))
           q(i,j,k,n_vit+4) = half*(uin(i,j,k,n_vit+4)+uin(i,j,k,nvar+2))
           q(i,j,k,n_vit+5) = half*(uin(i,j,k,n_vit+5)+uin(i,j,k,nvar+3))

           !compute magnetic energy
           emag = zero
           do idim = 1, n_vit
              emag = emag + half*q(i,j,k,idim+n_vit+2)**2
           end do

           ! Compute pressure
#ifdef ISO
           q(i,j,k,n_vit+2)=q(i,j,k,1)*ciso**2
#else
           eint = ( uin(i,j,k,n_vit+2) - emag ) /uin(i,j,k,1)-eken
           q(i,j,k,n_vit+2)=max((gamma-one)*q(i,j,k,1)*eint,smallp)
#endif

           ! gravity predictor step
           do idim = 1, ndim
              q(i,j,k,idim+1) = q(i,j,k,idim+1) + gravin(i,j,k,idim)*dt*half
           end do

        end do
     end do
  end do

!  ! Passive scalar
!  do n = 2*n_vit+3, nvar
!     do k = ku1, ku2
!        do j = ju1, ju2
!           do i = iu1, iu2
!              do l = 1, ngrid
!                 q(l,i,j,k,n) = uin(l,i,j,k,n)/uin(l,i,j,k,1)
!              end do
!           end do
!        end do
!     end do
!  end do

  return

end subroutine ctoprim
