!> \file
!! Contains subroutine exp_comobile_ray()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EXP_COMOBILE_RAY
!
!> This routine explicitly computes the comoving terms 
!! of radiation.
!<
subroutine exp_comobile_ray(rho,Ein,uS,imin,imax,jmin,jmax,kmin,kmax,nxmin,nxmax,nymin,nymax,nzmin,nzmax,radiation_model)

  use parameters, only : ndim
  use geom
  use var_loc   , only : xloc
  use divers    , only : verbose,dt
  use varray    , only : Eray,Fray,Erayold,Frayold
  use unites    , only : c
  use diffusion , only : Er_dif,fld_limiter
  
  implicit none

  integer                                                           , intent(in) :: imin,imax,jmin,jmax,kmin,kmax
  integer                                                           , intent(in) :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax              ), intent(in) :: rho,Ein
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim,1:ndim), intent(in) :: uS
  character(3)                                                      , intent(in) :: radiation_model

  integer, dimension(3,3) :: kron
  integer, dimension(3  ) :: indice
  integer                 :: i,j,k,idim,jdim,iface,ip1,jp1,kp1,im1,jm1,km1
  integer                 :: ixmin,ixmax,iymin,iymax,izmin,izmax
  integer                 :: nnx,nny,nnz
  integer                 :: nrad,irad,iloc

  real, dimension(:,:,:,:    ), allocatable :: radloc
  real, dimension(:,:,:,:,:  ), allocatable :: rads
  real, dimension(:,:,:,:,:,:), allocatable :: flux_rad
  real, dimension(:          ), allocatable :: flux_rad_tot
  real, dimension(1:ndim,1:ndim)            :: UX
  real, dimension(1:ndim,1:ndim)            :: dU        ! du(i,j) = dui/dxj
  real, dimension(6)                        :: signe
  real                                      :: Er,PrdivU
  real, dimension( 1:ndim      )            :: Fr,FrdivU
  real, dimension(  3, 3       )            :: Dedd
#if RAYO==1
  real, dimension(  3, 3       )            :: Dedd_dE
  real, dimension(  3, 3,1:ndim)            :: Dedd_dF
  logical                                   :: derive_chi = .false.
#endif
  real                                      :: coeff,xx,lambda_fld

  integer :: id
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax) :: Tgaz,kappa_abs,kappa_sca

  if(verbose) write(*,*) 'Start of exp_comobile_ray'

  ! for incoming or outgoing fluxes
  signe = 1. ; signe(2) = -1. ; signe(4) = -1. ; signe(6) = -1.

  kron = 0
  do idim=1,3
     kron(idim,idim) = 1
  enddo

  if(radiation_model.eq.'ray') then
     nrad = 1+ndim
  else if (radiation_model.eq.'dif') then
     nrad = 1
  endif

  allocate(radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nrad       ))
  allocate(radS  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nrad,1:ndim))

  ! Gather radiative sub-arrays
  if(verbose) write(*,*) 'Gather radiative sub-arrays'
  nnx = imax - imin + 1 ; nny =  jmax - jmin + 1 ; nnz = kmax - kmin + 1

  if(ndim.eq.1) then
     ixmin =  0 ; ixmax = nnx+1
     iymin =  1 ; iymax = 1
     izmin =  1 ; izmax = 1
  else if (ndim == 2) then
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  1 ; izmax = 1
  else if (ndim == 3) then
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  0 ; izmax = nnz+1
  endif

  if(radiation_model.eq.'ray') then
     if (ndim == 1) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1     ) = Erayold(imin-2:imax+2,     1:     1,     1:     1     )
        do idim=1,ndim
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1+idim) = Frayold(imin-2:imax+2,     1:     1,     1:     1,idim)
        enddo
     else if (ndim == 2) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1     ) = Erayold(imin-2:imax+2,jmin-2:jmax+2,     1:     1     )
        do idim=1,ndim
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1+idim) = Frayold(imin-2:imax+2,jmin-2:jmax+2,     1:     1,idim)
        enddo
     else if (ndim == 3) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1     ) = Erayold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2     )
        do idim=1,ndim
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1+idim) = Frayold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2,idim)
        enddo
     endif
  else if (radiation_model.eq.'dif') then
     if (ndim == 1) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1     ) = Er_dif (imin-2:imax+2,     1:     1,     1:     1     )
     else if (ndim == 2) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1     ) = Er_dif (imin-2:imax+2,jmin-2:jmax+2,     1:     1     )
     else if (ndim == 3) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1     ) = Er_dif (imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2     )
     endif

     call temperature(rho,Ein,Tgaz,nxmin,nxmax,nymin,nymax,nzmin,nzmax)
#if DIF==1
     call cal_kappa_dif(rho,Tgaz,kappa_abs,kappa_sca,nxmin,nxmax,nymin,nymax,nzmin,nzmax)
#endif
  endif

  allocate(flux_rad    (ixmin:ixmax,iymin:iymax,izmin:izmax,1:nrad ,2,ndim))
  allocate(flux_rad_tot(1:nrad                                            ))

  do idim = 1,ndim
     do k=1,nnz+kron(idim,3)
        do j=1,nny+kron(idim,2)
           do i=1,nnx+kron(idim,1)

              xx = 0.
              xx = max(0.,sign(1.,uS(i,j,k,idim,idim)))

              im1 = i - xx*kron(1,idim) ; jm1 = j - xx*kron(2,idim) ; km1 = k - xx*kron(3,idim)

              ! slopes could be added
              do irad=1,nrad
                 radS(i,j,k,irad,idim) = radloc(im1,jm1,km1,irad)
              enddo

           enddo
        enddo
     enddo
  enddo

  ! Fluxes
  do idim = 1,ndim
     do k=1,nnz
        do j=1,nny
           do i=1,nnx

              ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

              do irad=1,nrad
                 flux_rad(i,j,k,irad,1,idim) = radS(i  ,j  ,k  ,irad,idim) * uS(i  ,j  ,k  ,idim,idim) * ds(i  ,j  ,k  ,idim)
                 flux_rad(i,j,k,irad,2,idim) = radS(ip1,jp1,kp1,irad,idim) * uS(ip1,jp1,kp1,idim,idim) * ds(ip1,jp1,kp1,idim)
              enddo

           enddo
        enddo
     enddo
  enddo

  ! P.divU and F.D u
  do k=1,nnz
     do j=1,nny
        do i=1,nnx

           indice(1) = i
           indice(2) = j
           indice(3) = k

           do jdim=1,ndim
              ip1 = i + kron(1,jdim) ; jp1 = j + kron(2,jdim) ; kp1 = k + kron(3,jdim)

              iloc = indice(jdim)

              coeff = 1.
              if(geom_dir(jdim)=='phi_c' .or. geom_dir(jdim)=='the_s') coeff = xcloc(i,1)
              if(geom_dir(jdim)=='phi_s'                             ) coeff = xcloc(i,1)*sin(xcloc(j,2))

              do idim=1,ndim
                 dU(idim,jdim)=(uS(ip1,jp1,kp1,idim,jdim) - uS(i,j,k,idim,jdim) )/(xloc(iloc+1,jdim)-xloc(iloc,jdim))/coeff
              enddo
           enddo

           Er         = radloc(i,j,k,1       )
           if(radiation_model.eq.'ray') then
              Fr(1:ndim) = radloc(i,j,k,2:ndim+1)
#if RAYO==1
              call cal_Dedd(Er,Fr,Dedd,Dedd_dE,Dedd_dF,Derive_chi)
#endif
           else
              Dedd = 0.
#if DIF==1
              do idim = 1,ndim

                 ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)
                 im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)

                 coeff=1.
                 if(idim.eq.1) then
                    id  = i
                 else if(idim.eq.2) then
                    id  = j
                    if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') coeff = xcloc(i,1)
                 else if(idim.eq.3) then
                    id  = k
                    if (geom_dir(3) == 'phi_s') coeff = xcloc(i,1)*sin(xcloc(j,2))
                 endif

                 !FLD
                 xx = max(1.e-7,abs( (radloc(ip1,jp1,kp1,1) - radloc(im1,jm1,km1,1)) / &
                      (0.5*(radloc(ip1,jp1,kp1,1) + radloc(im1,jm1,km1,1))) ) )
                 xx = xx/((xcloc(id+1,idim)-xcloc(id-1,idim))*coeff) / (kappa_abs(i,j,k)*rho(i,j,k))

                 Dedd(idim,idim) = lambda_fld(xx)
              enddo
#endif

           endif


           PrDivu = 0. ; FrdivU = 0.
           do idim = 1,ndim
              do jdim = 1,ndim
                    PrdivU       = PrdivU       + Dedd(idim,jdim)*Er*dU(idim,jdim)
                 if(radiation_model.eq.'ray') &
                    FrdivU(idim) = FrDivU(idim) + Fr(jdim)          *du(idim,jdim)
              enddo
           enddo

           ! Geometry terms
           ! ~~~~~~~~~~~~~~
           do idim = 1,ndim
              ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)
              do jdim = 1,ndim
                 UX(jdim,idim) = 0.5*(uS(i,j,k,jdim,idim)+ uS(ip1,jp1,kp1,jdim,idim))
              enddo
           enddo

           if(geom_dir(2)=='the_s' .or. geom_dir(2)=='phi_c') then
              ! Term P:div U  : ( -2*P(r,theta)*u(theta)/r+P(theta,theta)*u(r)/r )
              !*******************************************************************
              PrdivU =  PrdivU + Er*( -Dedd(1,2)*(UX(2,1) + UX(2,2)) + Dedd(2,2)*UX(1,2) )/xcloc(i,1)
           endif

           if(geom_dir(3)=='phi_s') then
              ! Term P:div U  : ( -2*P(r    ,theta)*u(theta)/r+P(theta,theta)*u(r)/r   ! already treated because geom_dir(2)=='the_s'
              !                   -2*P(r    ,phi  )*u(phi  )/r
              !                   -2*P(theta,phi  )*u(phi  )*cot(theta)/r
              !                   +  P(phi  ,phi  )*(u(r)+u(theta)*cot(theta))/r      )
              !************************************************************************
              PrdivU =  PrdivU + Er*( -Dedd(1,3)*                     (UX(3,1) + UX(3,3))                                     &
                   &                  -Dedd(2,3)*                     (UX(3,2) + UX(3,3)) *cos(xcloc(j,2))/sin(xcloc(j,2))    &
                   &                  +Dedd(3,3)* (  UX(1,3) +          UX(2,3)           *cos(xcloc(j,2))/sin(xcloc(j,2)) )   )/ xcloc(i,1)
           endif

           if(radiation_model.eq.'ray') then
              if(geom_dir(2)=='the_s' .or. geom_dir(2)=='phi_c') then
                 ! Term (F.div) u over er     : -F(theta)*u(theta)/r
                 ! Term (F.div) u over etheta : +F(theta)*u(r    )/r
                 !**************************************************
                 FrdivU(1) =  FrdivU(1) - Fr(2)* UX(2,1)/xcloc(i,1)
                 FrdivU(2) =  FrdivU(2) + Fr(2)* UX(1,2)/xcloc(i,1)
              endif

              if(geom_dir(3)=='phi_s') then
                 ! Term (F.div) u over er     : -F(phi)*u(phi)/r 
                 ! Term (F.div) u over etheta : -F(phi)*u(phi)*cot(theta)/r
                 ! Term (F.div) u over ephi   : +(F(phi)*u(r)+cot(theta)*F(phi)*u(theta))/r
                 !*************************************************************************
                 FrdivU(1) =  FrdivU(1) - Fr(3) *  UX(3,1)                                              /xcloc(i,1)
                 FrdivU(2) =  FrdivU(2) - Fr(3) *            cos(xcloc(j,2))/sin(xcloc(j,2)) * UX(3,2)  /xcloc(i,1)
                 FrdivU(3) =  FrdivU(3) + Fr(3) * (UX(1,3) + cos(xcloc(j,2))/sin(xcloc(j,2)) * UX(2,3) )/xcloc(i,1)
              endif
           endif


           flux_rad_tot (1:nrad)  = 0.

           do idim  = 1,ndim
              do iface = 1,2
                 flux_rad_tot (1:nrad)  = flux_rad_tot (1:nrad)  + signe(iface)* dt*flux_rad(i,j,k,1:nrad ,iface,idim)
              enddo
           enddo

           radloc(i,j,k,1:nrad) = radloc(i,j,k,1:nrad) + flux_rad_tot (1:nrad) /dv(i,j,k)

              radloc(i,j,k,1        ) =  radloc(i,j,k,1        ) - PrdivU         * dt
           if(radiation_model.eq.'ray') &
              radloc(i,j,k,2:ndim+1 ) =  radloc(i,j,k,2:ndim+1 ) - FrdivU(1:ndim) * dt
        enddo
     enddo
  enddo

  ! Fill arrays with actualized sub-arrays
  if(radiation_model.eq.'ray') then
        Eray(imin:imax,jmin:jmax,kmin:kmax     ) = radloc(1:nnx,1:nny,1:nnz,1     )
     do idim=1,ndim
        Fray(imin:imax,jmin:jmax,kmin:kmax,idim) = radloc(1:nnx,1:nny,1:nnz,1+idim)
     enddo
  else if (radiation_model.eq.'dif') then
     Er_dif (imin:imax,jmin:jmax,kmin:kmax     ) = radloc(1:nnx,1:nny,1:nnz,1     )
  endif

  deallocate(flux_rad)
  deallocate(flux_rad_tot)

  deallocate(radloc)
  deallocate(radS)

  if(verbose) write(*,*) 'End of exp_comobile_ray'
  
  return

end subroutine exp_comobile_ray
