!> \file
!! Contains subroutine exp_source_ray()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EXP_SOURCE_RAY
!
!> This routine explicitly computes the radiative source
!! term rho*(kappa_abs+kappa_sca)F/c in moment and energy
!! hydro equations.
!<
subroutine exp_source_ray(rho,Ein,moment_ray,imin,imax,jmin,jmax,kmin,kmax,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use parameters, only : ndim
  use geom      , only : xcloc,geom_dir
  use var_loc   , only : xloc
  use divers    , only : verbose,RAY,DIFFU
  use varray    , only : Frayold
  use unites    , only : c
  use diffusion , only : Er_dif,fld_limiter
  
  implicit none

  integer,                                                     intent(in ) :: imin,imax,jmin,jmax,kmin,kmax
  integer,                                                     intent(in ) :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ), intent(in ) :: rho,Ein
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim), intent(out) :: moment_ray

  integer, dimension(3,3) :: kron
  integer                 :: i,j,k,idim,id,ip1,jp1,kp1,im1,jm1,km1
  integer                 :: nnx,nny,nnz
  integer                 :: ixmin,ixmax,iymin,iymax,izmin,izmax

  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) :: Tgaz,kappa_abs,kappa_sca
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) :: radloc
  real                                                        :: coeff,xx,lambda_fld

  if(verbose) write(*,*) 'Start of exp_source_ray'

  kron = 0
  do idim = 1,3
     kron(idim,idim) = 1
  enddo

  nnx = imax - imin + 1 ; nny =  jmax - jmin + 1 ; nnz = kmax - kmin + 1

  if(ndim.eq.1) then
     ixmin =  0 ; ixmax = nnx+1
     iymin =  1 ; iymax = 1
     izmin =  1 ; izmax = 1
  else if (ndim == 2) then
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  1 ; izmax = 1
  else if (ndim == 3) then
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  0 ; izmax = nnz+1
  endif

  moment_ray=0.
  call temperature(rho,Ein,Tgaz,nxmin,nxmax,nymin,nymax,nzmin,nzmax)
  
#if RAYO==1

  if(RAY) then
     call cal_kappa_ray(rho,Tgaz,kappa_abs,kappa_sca,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

     if (ndim == 1) then
        do idim=1,ndim
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim) = Frayold(imin-2:imax+2,     1:     1,     1:     1,idim)
        enddo
     else if (ndim == 2) then
        do idim=1,ndim
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim) = Frayold(imin-2:imax+2,jmin-2:jmax+2,     1:     1,idim)
        enddo
     else if (ndim == 3) then
        do idim=1,ndim
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim) = Frayold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2,idim)
        enddo
     endif

     do idim=1,ndim
        moment_ray(:,:,:,idim) = rho*(kappa_abs+kappa_sca)/c*radloc(:,:,:,idim)
     enddo
  endif
  
#endif

#if DIF==1

  if(DIFFU) then

     if (ndim == 1) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1   ) = Er_dif (imin-2:imax+2,     1:     1,     1:     1     )
     else if (ndim == 2) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1   ) = Er_dif (imin-2:imax+2,jmin-2:jmax+2,     1:     1     )
     else if (ndim == 3) then
           radloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1   ) = Er_dif (imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2     )
     endif

     call cal_kappa_dif(rho,Tgaz,kappa_abs,kappa_sca,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

     do k=1,nnz
     do j=1,nny
     do i=1,nnx

        do idim = 1,ndim
        
           im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)
           ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

           coeff=1.
           if(idim.eq.1) then
              id  = i
           else if(idim.eq.2) then
              id  = j
              if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') coeff = xcloc(i,1)
           else if(idim.eq.3) then
              id  = k
              if (geom_dir(3) == 'phi_s') coeff = xcloc(i,1)*sin(xcloc(j,2))
           endif

           !FLD
           xx = max(1.e-7,abs( (radloc(ip1,jp1,kp1,1) - radloc(im1,jm1,km1,1)) / &
                (0.5*(radloc(ip1,jp1,kp1,1) + radloc(im1,jm1,km1,1))) ) )
           xx = xx/((xcloc(id+1,idim)-xcloc(id-1,idim))*coeff) / (kappa_abs(i,j,k)*rho(i,j,k))

           moment_ray(i,j,k,idim) = moment_ray(i,j,k,idim)-lambda_fld(xx)*(radloc(ip1,jp1,kp1,1) - &
                                    radloc(im1,jm1,km1,1)) / ((xcloc(id+1,idim)-xcloc(id-1,idim))*coeff)
        enddo

     enddo
     enddo
     enddo

  endif
  
#endif

  return

end subroutine exp_source_ray
