!> \file
!! Contains subroutine diffusion_step()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine DIFFUSION_STEP
!
!> This routine performs the evolution over one timestep
!! of the radiative flux limited diffusion.
!<
subroutine diffusion_step(t_fld,t_com,monit)

  use parameters, only : nx,ndim
  use divers    , only : nsx,nsy,nsz,verbose
  use diffusion , only : dif_imp,Er_dif
  use cl_dif    , only : icl_dif,Er_dif_l

  implicit none
  
#if PARA==1
  include 'mpif.h'
#else
  real     :: MPI_Wtime
  external :: MPI_Wtime
#endif

  integer           :: i,j,k,ierr
  integer           :: ixm,ixp,iym,iyp,izm,izp
  real              :: t_fld,t_com,tr1,tr2
  logical           :: monit  
  character(len=50) :: Com_Moni

  if(verbose) write(*,*) 'Start of diffusion_step'

  Com_Moni = 'Before FLD step...'
  if (monit) call monitoring(Com_Moni)

  tr1 = MPI_Wtime(ierr)     
  if(Dif_imp) then
     call diffusion_imp
  else
     do i = 1,nsx
        do j = 1,nsy
           do k = 1,nsz
              ixm = 1 + (i-1)*nx(1)/nsx
              iym = 1 + (j-1)*nx(2)/nsy
              izm = 1 + (k-1)*nx(3)/nsz
              ixp = i*nx(1)/nsx
              iyp = j*nx(2)/nsy
              izp = k*nx(3)/nsz
              call diffusion_ex(ixm,ixp,iym,iyp,izm,izp)
           enddo
        enddo
     enddo
  endif
  tr2 = MPI_Wtime(ierr)     
  t_fld = t_fld + tr2-tr1

  call communications_dif(t_com)
  call limites_dif
  
  call communications(t_com)
  call limites

  Com_Moni = 'After FLD step...'
  if (monit) call monitoring(Com_Moni)

  if(verbose) write(*,*) 'End of diffusion_step'

  return

end subroutine diffusion_step
