!> \file
!! Contains subroutine pasdt_dif()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PASDT_DIF
!
!> This routine computes the FLD timestep.
!<
subroutine pasdt_dif

  use parameters
  use var    
  use unites     , only : c,an
  use geom       , only : x,geom_dir
  use para       , only : mype
  use diffusion
  use divers     , only : verbose,dt_max
  
  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  integer :: i,j,k,idim,ixmin,ixmax,iymin,iymax,izmin,izmax

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: T,kappa_abs,kappa_sca
  real, dimension(1:ndim)                                                :: dx
  real                                                                   :: dt,dt_glob,xx

  if (verbose) write(*,*) 'Start of pasdt_dif'

  if (Dif_imp) then
     dt_dif = min(dt_max,dt_dif_imp)
  else
     
     E = E - 0.5*sum(rhou**2,4)/rho
     if(allocated(B)) E = E - 0.5*sum(B**2,4)
     call temperature(rho,E,T,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
     E = E + 0.5*sum(rhou**2,4)/rho
     if(allocated(B)) E = E + 0.5*sum(B**2,4)
     call cal_kappa_dif(rho,T,kappa_abs,kappa_sca,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
 
     dt = 1.0E+50*an
  
     ixmin =  0 ; ixmax = nx(1)+1
     iymin =  1 ; iymax = 1
     izmin =  1 ; izmax = 1
     if(ndim.gt.1) then
        iymin =  0 ; iymax = nx(2)+1
     endif
     if(ndim.gt.2) then
        izmin =  0 ; izmax = nx(3)+1
     endif
     
     do k = izmin,izmax
     do j = iymin,iymax
     do i = ixmin,ixmax

                         dx(1) = x(i+1,1)-x(i,1)
        if (ndim .gt. 1) dx(2) = x(j+1,2)-x(j,2)
        if (ndim .gt. 2) dx(3) = x(k+1,3)-x(k,3)
        
        if (geom_dir(2) == 'phi_c') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
        if (geom_dir(2) == 'the_s') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
        if (geom_dir(3) == 'phi_s') dx(3) = 0.5*(x(i,1)+x(i+1,1))*sin(0.5*(x(j,2)+x(j+1,2))) * dx(3)
        
        xx = 0.
        do idim = 1,ndim
           xx = xx + 1./dx(idim)**2
        enddo
        
        dt = min(dt,0.8 * 1./(2.*xx) * kappa_abs(i,j,k)*rho(i,j,k)*3/c)

     enddo
     enddo
     enddo

     ! Timestep reduction
#if PARA==1
     info = 0
     call MPI_ALLREDUCE(dt, dt_glob,1,MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,info)
#else
     dt_glob = dt
#endif

     dt_dif = dt_glob
     
  endif

  dt_dif = min(dt_dif,dt_max)

  if (verbose) write(*,*) 'End of pasdt_dif'

  return
  
end subroutine pasdt_dif
