!> \file
!! Contains subroutines gs_mat(), cal_Dedd() and
!! essai_inv5x5 and functions det_3x3() and det_4x4()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GS_MAT
!
!> Performs implicit resolution of one timestep for
!! radiative transfer using the Gauss-Seidel
!! algorithm.
!<
subroutine gs_mat

!*********************************************************************
!            Different algorithms
!
!  Jacobi       :    D  xn+1 = -(L+U) xn + b
!
!  Gauss-Seidel : (L+D) xn+1 = -   U  xn + b
!*********************************************************************

  use parameters, only : ndim,nxmin,nxmax,nx,nvar_ray,pi
  use divers_ray, only : alpha,nitemax_ray,matrix_inv_solver,rad_trans_model,slope_type_ray, &
                         nitetot_ray,alpha_var,deja_converge,cal_valp,precond_ray,valp_min
  use divers    , only : mu,dt,temps,tout,iout,verbose,hydro
  use cputime   , only : temps_comm
  use unites    , only : c,c2,seconde
  use geom      , only : x,dx,ds,dv,xcloc, geometrie,cylindrique,spherique,geom_dir,dxc
  use var_loc   , only : xloc
  use varray    , only : Eray,Fray,Erayold,Frayold,kappa_abs_ray,kappa_emi_ray,kappa_sca_ray
  use var       , only : Tgaz,rho,fx
  Use varold    , only : Tgazold
  use cl_ray    , only : icl_ray
  use para      , only : mype,x_start,ncpu

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info,resume_glob
  real    :: res_norme2_glob, v_glob
#endif

  integer, dimension(2,3,2,3) :: Iv
  integer, dimension(3)       :: nnx,mx1,mx2
  integer                     :: nite,imax_res,jmax_res,kmax_res
  integer                     :: i,j,k,idim,jdim,kdim
  integer                     :: imin,imax,ipas,jmin,jmax,jpas,iloc
  integer                     :: iface,idp,idm
  integer                     :: ip,jp,kp,im,jm,km
  integer                     :: i1,Iconv,resume,nn
  
  real, dimension(:,:,:    ), allocatable :: deltaEray,deltaTgaz
  real, dimension(:,:,:,:  ), allocatable :: deltaFray
  real, dimension(:,:,:    ), allocatable :: Eray0,Tgaz0
  real, dimension(:,:,:,:  ), allocatable :: Fray0
  real, dimension(:,:,:,:  ), allocatable :: peray,pfray1,pdummy
  real, dimension(:,:,:,:,:), allocatable :: pfray
  real, dimension(1:5,1:5)                :: mat,inv
  real, dimension(1:5    )                :: vect,FFF
  real, dimension(     2,ndim)            :: ErayM,ErayP,dEray
  real, dimension(ndim,2,ndim)            :: FrayM,FrayP,dFray
  real, dimension(     2,ndim)            :: ffM,ffP
  real, dimension(1:3,1:3,2,ndim)         :: PrayM,PrayP,DeddP,DeddM
  real, dimension(        2,ndim)         :: eps,ap,am,lm,lp,lmp
  real                                    :: Eraytemp,Tgaztemp
  real, dimension(1:ndim)                 :: Fraytemp
  real, dimension(3,3       )             :: Dedd,Dedd_dE
  real, dimension(3,3,1:ndim)             :: Dedd_dF
  real, dimension(3,3,       2,1:ndim)    :: DeddM_dE, DeddP_dE
  real, dimension(3,3,1:ndim,2,1:ndim)    :: DeddM_dF, DeddP_dF
  real, dimension(3,3       ,2)           :: Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d
  real, dimension(3,3,1:ndim,2)           :: Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d
  real, dimension(2,1:ndim)               :: dSloc
  real, dimension(       2,1:ndim)        :: flux_F
  real, dimension(1:ndim,2,1:ndim)        :: flux_P
  real, dimension(     3,2,3     )        :: Rayon
  real                                    :: flux_F_tot
  real, dimension(1:ndim         )        :: flux_P_tot
  real, dimension(       2       )        :: signe
  real, dimension(nvar_ray,2*nvar_ray)    :: a1d,a2d,a3d
  real, parameter                         :: varres=1.e-4
  real                                    :: drp,drm,ff1,ff2
  real                                    :: alpha_var_glob,alpha_var_new
  real                                    :: resmax,resF,resmaxF,resmaxF_glob,res_norme2
  real                                    :: varE_max,varE,varT_max,varT,var_max,varF_max,varF
  real                                    :: conv,thetaM,thetaP,interpol_valp
  real                                    :: xx,xm,un_moins,un_plus,f,r
  real                                    :: source_ray,derive_source_ray,omega 

  character*2 :: var_pasconv     !< variable which did not converge (Tgaz, Eray, Fx or Fy)

  logical :: boucle,update,Chi_uniforme,Derive_Chi,Pentes_fixe,dX_total

!*********************************************************************************************************************************
!  Dedd,Dedd_dE (i,j)
!  Dedd_dF      (i,j,k)
!
!  DeddM_dE, DeddP_dE  (i,j  ,iface,idim)
!  DeddM_dF, DeddP_dF  (i,j,k,iface,idim)
!  
!  Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d  (i,j  ,iface)
!  Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d                          (i,j,k,iface)
!
!   iface : face index            (=1:   2) with 1 = back ; 2 = front
!   idim  : direction considered  (=1:ndim)
!
!   i,j   : tensorial coordinates       (=1:3   )
!   k     : flux direction considered   (=1:ndim)
!
!*********************************************************************************************************************************

  if (verbose) write(*,*) 'Start of gs'

  ! In order to have a uniform chi in the cells
  Chi_uniforme = .false.

  ! In order to include the derivatives in the matrix
  Derive_Chi   = .true.
  
  ! In order to use the slopes on the var_old
  Pentes_fixe = .true.

  ! dX_total = .true.  : compute the total variation
  ! dX_total = .false. : compute the increment with respect to the previous iteration
  dX_total = .true.

  ! incoming and outgoing fluxes
  signe(1) = -1. ; signe(2) = 1.

   do i = 1,3
      nnx(i) = 4 + nx(i)
      mx1(i) = max(-1,nxmin(i))
      mx2(i) = min(nnx(i)-2,nxmax(i)) 
   enddo
 
  ! geometry
  nn = max(nx(1),nx(2),nx(3))
  allocate(xloc(-1:nn+3,ndim))
  
                 xloc(-1:nx(1)+3,1) = x(-1:nx(1)+3,1)   
  if (ndim.gt.1) xloc(-1:nx(2)+3,2) = x(-1:nx(2)+3,2)   
  if (ndim.gt.2) xloc(-1:nx(3)+3,3) = x(-1:nx(3)+3,3)   

  call cal_geom(nx(1),nx(2),nx(3))

  ! neighbour offsets
  ! Iv(i,j,k,l) :: i = front or back face -- j = face direction -- k = front or back cell -- l = cell coordinate
  Iv = 0
  do idim=1,3
     Iv(1,idim,1,idim) = -1
     Iv(2,idim,2,idim) = +1
  enddo

  allocate(deltaTgaz(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3)       ) )
  allocate(deltaEray(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3)       ) )
  allocate(deltaFray(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim) ) 

  allocate(Tgaz0(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3)       ) )
  allocate(Eray0(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3)       ) )
  allocate(Fray0(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim) ) 
  
  allocate(peray (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  allocate(pfray1(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  allocate(pdummy(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  allocate(pfray (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim,1:ndim)) 

  omega = 1.

  deja_converge=.false.

  alpha_var = 1. 
  
  un_moins = 1. !- 1.0E-8
  un_plus  = 1. !+ 1.0E-8

  ! Boundary conditions
  call communications_ray(temps_comm)
  call limites_ray

  call cal_kappa_ray(rho,Tgaz,kappa_emi_ray,kappa_sca_ray,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  kappa_abs_ray = kappa_emi_ray

  do k=nxmin(3),nxmax(3)
     do j=nxmin(2),nxmax(2)
        do i=nxmin(1),nxmax(1)
           if (Eray(i,j,k) <= 0.) then
              write(*,*) '** cal_ray **'
              write(*,*) 'i,j,k,Eray(i,j,k)= ',i,j,k,Eray(i,j,k)
              read(*,*)
           endif
        enddo
     enddo
  enddo

  do k=nxmin(3),nxmax(3)
     do j=nxmin(2),nxmax(2)
        do i=nxmin(1),nxmax(1)
           if ( Tgaz(i,j,k) < 0.) then
              write(*,*) 'Debut de Ray'
              write(*,*) 'Tgaz =', Tgaz(i,j,k)
              write(*,*) 'i,j,k=',i,j,k
              read(*,*)
           endif
        enddo
     enddo
  enddo

  do k=nxmin(3),nxmax(3)
     do j=nxmin(2),nxmax(2)
        do i=nxmin(1),nxmax(1)
           f = 0.0
           do idim = 1,ndim
              f = f + (Fray(i,j,k,idim)/(c*Eray(i,j,k)))**2
           enddo
           f = sqrt(f)
           if(f > 1.0)then
              write(*,*) 'F/cE > 1 : i,j,k,F/cE,fx,fy,fz =',i,j,k,f,(Fray(i,j,k,idim)/(c*Eray(i,j,k)),idim=1,ndim)
              r = 1.0 / f
              do idim = 1,ndim
                 Fray(i,j,k,idim) = Fray(i,j,k,idim) * r
              enddo
!               read(*,*)
           endif
        enddo
     enddo
  enddo

  var_pasconv='00'


  nite   = 0
222  resmax = 1.
  resume = 0

  boucle=.true.
  
  do k=mx1(3),mx2(3)
  do j=mx1(2),mx2(2)
  do i=mx1(1),mx2(1)
     Tgaz0(i,j,k       ) = Tgazold(i,j,k       )
     Eray0(i,j,k       ) = Erayold(i,j,k       )
     Fray0(i,j,k,1:ndim) = Frayold(i,j,k,1:ndim)
     
     if (Erayold(i,j,k) <= 0.) then
        write(*,*) 'i,j,k,Eray(i,j,k)= ',i,j,k,Erayold(i,j,k)
        !read(*,*)
     endif
  enddo
  enddo
  enddo

  ! *********************
  ! Convergence loop
  ! *********************

  do while (boucle)     

     ! Boundary conditions
     call communications_ray(temps_comm)
     call limites_ray

     nite = nite + 1

     res_norme2 = 0.

     ! Fill delta variables
     do k=mx1(3),mx2(3)
     do j=mx1(2),mx2(2)
     do i=mx1(1),mx2(1)
     
        deltaTgaz(i,j,k       ) = Tgaz(i,j,k       ) - Tgaz0(i,j,k       )
        deltaEray(i,j,k       ) = Eray(i,j,k       ) - Eray0(i,j,k       )
        deltaFray(i,j,k,1:ndim) = Fray(i,j,k,1:ndim) - Fray0(i,j,k,1:ndim)
        
     enddo
     enddo 
     enddo

     ! Fill 0 variables
     if (dX_total) then
     
        do k=mx1(3),mx2(3)
        do j=mx1(2),mx2(2)
        do i=mx1(1),mx2(1)
        
           Tgaz0(i,j,k       ) = Tgazold(i,j,k       )
           Eray0(i,j,k       ) = Erayold(i,j,k       )
           Fray0(i,j,k,1:ndim) = Frayold(i,j,k,1:ndim)
           
        enddo
        enddo
        enddo
        
     else
     
        do k=mx1(3),mx2(3)
        do j=mx1(2),mx2(2)
        do i=mx1(1),mx2(1)
        
           Tgaz0(i,j,k       ) = Tgaz(i,j,k       )
           Eray0(i,j,k       ) = Eray(i,j,k       )
           Fray0(i,j,k,1:ndim) = Fray(i,j,k,1:ndim)
           
        enddo
        enddo
        enddo
        
     endif

     if (verbose) write(*,*) 'Nite = ',nite

     if (Pentes_fixe) then
        if (verbose) write(*,*) '     -- Eray slope --'
        Call slopes(Erayold,peray,pdummy,mx1(1),mx2(1),mx1(2),mx2(2),mx1(3),mx2(3),slope_type_ray)
        if (verbose) write(*,*) '     -- Fray slope --'
        do idim =1,ndim
           Call slopes(Frayold(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),idim),pfray1,pdummy, &
                       mx1(1),mx2(1),mx1(2),mx2(2),mx1(3),mx2(3),slope_type_ray)
           pfray(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),idim,1:ndim) = pfray1(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim)
        enddo
     else
        if (verbose) write(*,*) '     -- Eray slope --'
        Call slopes(Eray0,peray,pdummy,mx1(1),mx2(1),mx1(2),mx2(2),mx1(3),mx2(3),slope_type_ray)
        if (verbose) write(*,*) '     -- Fray slope --'
        do idim =1,ndim
           Call slopes(Fray0(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),idim),pfray1,pdummy, &
                       mx1(1),mx2(1),mx1(2),mx2(2),mx1(3),mx2(3),slope_type_ray)
           pfray(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),idim,1:ndim) = pfray1(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim)
        enddo
     endif

     resmaxF = 0.

     do k=1,nx(3)
     do j=1,nx(2)
     do i=1,nx(1)


        PrayM = 0. ; PrayP = 0.

        if (verbose) write(*,*) '     -- Interface values --'

        ! Copmute Er, Fr and theta at the cell interfaces 
        do idim  = 1,ndim
        do iface = 1,2

           im = i + Iv(iface,idim,1,1) ; jm = j + Iv(iface,idim,1,2); km = k + Iv(iface,idim,1,3)
           ip = i + Iv(iface,idim,2,1) ; jp = j + Iv(iface,idim,2,2); kp = k + Iv(iface,idim,2,3)

           if(idim.eq.1) then
              idm = im
              idp = ip
           else if(idim.eq.2) then
              idm = jm
              idp = jp
           else if(idim.eq.3) then
              idm = km
              idp = kp
           endif

           drp = 0.5*(x(idp  ,idim)-x(idm,idim))
           drm = 0.5*(x(idp+1,idim)-x(idp,idim)) 

           ErayM(iface,idim) = Eray0(im,jm,km) + drp*peray(im,jm,km,idim)
           ErayP(iface,idim) = Eray0(ip,jp,kp) - drm*peray(ip,jp,kp,idim)

           if(iface.eq.1) then
              dEray(iface,idim) = ErayP(iface,idim) - Eray0(i,j,k) 
           else
              dEray(iface,idim) = ErayM(iface,idim) - Eray0(i,j,k) 
           endif

           do jdim = 1,ndim
              FrayM(jdim,iface,idim) = Fray0(im,jm,km,jdim) + drp*pfray(im,jm,km,jdim,idim)
              FrayP(jdim,iface,idim) = Fray0(ip,jp,kp,jdim) - drm*pfray(ip,jp,kp,jdim,idim)
           enddo

           ! Ensure that f < 1
           f = 0.0
           do jdim = 1,ndim
              f = f + (FrayM(jdim,iface,idim)/(c*ErayM(iface,idim)))**2
           enddo
           f = sqrt(f)
           if(f > 1.0)then
              r = 1.0 / f
              do jdim = 1,ndim
                 FrayM(jdim,iface,idim) = FrayM(jdim,iface,idim) * r
              enddo
           endif
           f = 0.0
           do jdim = 1,ndim
              f = f + (FrayP(jdim,iface,idim)/(c*ErayP(iface,idim)))**2
           enddo
           f = sqrt(f)
           if(f > 1.0)then
              r = 1.0 / f
              do jdim = 1,ndim
                 FrayP(jdim,iface,idim) = FrayP(jdim,iface,idim) * r
              enddo
           endif

           ffM(iface,idim) = 0. ; ffP(iface,idim) = 0.
           do jdim = 1,ndim
              if(iface.eq.1) then
                 dFray(jdim,iface,idim) = FrayP(jdim,iface,idim)- Fray0(i,j,k,jdim)
              else
                 dFray(jdim,iface,idim) = FrayM(jdim,iface,idim)- Fray0(i,j,k,jdim)
              endif

              ffM(iface,idim) = ffM(iface,idim) + FrayM(jdim,iface,idim)**2
              ffP(iface,idim) = ffP(iface,idim) + FrayP(jdim,iface,idim)**2
           enddo
           ffM(iface,idim) = sqrt(ffM(iface,idim))/(c*ErayM(iface,idim))
           ffP(iface,idim) = sqrt(ffP(iface,idim))/(c*ErayP(iface,idim))

           ! Calculate theta incidence angle to interpolate eigenvalues
           ! theta needs to be calculated here before ffM is changed if it is greater than 1
           if(ffM(iface,idim).gt.1.e-5) then
              xx=FrayM(idim,iface,idim)/(c*ErayM(iface,idim))/ffM(iface,idim)
              if(abs(xx).gt.1.) then
                 thetaM=pi/2.*(1.-sign(1.,xx))
              else
                 thetaM=acos(xx)
              endif
           else
              thetaM=0.
           endif
           if(ffP(iface,idim).gt.1.e-5) then
              xx=FrayP(idim,iface,idim)/(c*ErayP(iface,idim))/ffP(iface,idim)
              if(abs(xx).gt.1.) then
                 thetaP=pi/2.*(1.-sign(1.,xx))
              else
                 thetaP=acos(xx)
              endif
           else
              thetaP=0.
           endif

           ! If ffM > 1 then set it to 1 - WARNING: NOT VERY SATISFACTORY!! Ed
           if (ffM(iface,idim).gt. 1.) ffM(iface,idim)= 1.
           if (ffP(iface,idim).gt. 1.) ffP(iface,idim)= 1.


           ! Compute Eddington tensor and its derivatives at the cell centre and interfaces
           if (Chi_uniforme) then
              call cal_Dedd(Eray0(ip,jp,kp),Fray0(ip,jp,kp,1:ndim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
              call cal_Dedd(Eray0(im,jm,km),Fray0(im,jm,km,1:ndim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
           else
              call cal_Dedd(ErayP(iface,idim),FrayP(1:ndim,iface,idim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
              call cal_Dedd(ErayM(iface,idim),FrayM(1:ndim,iface,idim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
           endif

           do jdim=1,ndim
           do kdim=1,ndim
              PrayM(jdim,kdim,iface,idim) = DeddM(jdim,kdim,iface,idim)*ErayM(iface,idim)
              PrayP(jdim,kdim,iface,idim) = DeddP(jdim,kdim,iface,idim)*ErayP(iface,idim)
           enddo
           enddo
           

           ! Compute epsilon for asymptotic preserving scheme
           if (verbose) write(*,*) '     -- compute epsilon --'
        
           ff1 = ffM(iface,idim)
           ff2 = ffP(iface,idim)

           if( ((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)).ne.0. .and. &
               ((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)).ne.0.       ) then
              eps(iface,idim) = max( 1./((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)*(2.*drm)) &
                   &                ,1./((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)*(2.*drp)) ) 
              eps(iface,idim) = max(eps(iface,idim),2.*abs(ff1),2.*abs(ff2)) 
              eps(iface,idim) = min(1.,eps(iface,idim)) 
           else
              eps(iface,idim) = 1.
           endif


           ! Compute eigenvalues at the cell interfaces
           if (verbose)  write(*,*) '     -- compute eigenvalues --'
           
           if (cal_valp) then

              select case(rad_trans_model)
      
              case('P1')
      
                 am(iface,idim)=min(-valp_min,-1./sqrt(3.))
                 ap(iface,idim)=max(+valp_min, 1./sqrt(3.))
         
              case('M1')
      
                 am(iface,idim)=min(-valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),1), &
                                              interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),1))
                 ap(iface,idim)=max(+valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),4), &
                                              interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),4))
      
              end select
      
           else
              am(iface,idim) = -1.
              ap(iface,idim) =  1.
           endif
           
           
           if(ap(iface,idim).ne.am(iface,idim)) then
              lm (iface,idim) = am(iface,idim)               /(ap(iface,idim)-am(iface,idim))
              lp (iface,idim) =                ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
              lmp(iface,idim) = am(iface,idim)*ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
           else ! then ap=am=0.
              lm (iface,idim) = -1./2
              lp (iface,idim) =  1./2
              
              lmp(iface,idim) =  0.
           endif

        enddo
        enddo



        call cal_Dedd(Eray0(i  ,j,k),Fray0(i  ,j,k,1:ndim),Dedd          ,Dedd_dE          ,Dedd_dF             , Derive_chi)
        call cal_Dedd(Eray0(i-1,j,k),Fray0(i-1,j,k,1:ndim),Dedd_1d(:,:,1),Dedd_dE_1d(:,:,1),Dedd_dF_1d(:,:,:,1) , Derive_chi)
        call cal_Dedd(Eray0(i+1,j,k),Fray0(i+1,j,k,1:ndim),Dedd_1d(:,:,2),Dedd_dE_1d(:,:,2),Dedd_dF_1d(:,:,:,2) , Derive_chi)
        if(ndim.gt.1) then
           call cal_Dedd(Eray0(i,j-1,k),Fray0(i,j-1,k,1:ndim),Dedd_2d(:,:,1),Dedd_dE_2d(:,:,1),Dedd_dF_2d(:,:,:,1) , Derive_chi)
           call cal_Dedd(Eray0(i,j+1,k),Fray0(i,j+1,k,1:ndim),Dedd_2d(:,:,2),Dedd_dE_2d(:,:,2),Dedd_dF_2d(:,:,:,2) , Derive_chi)
           if(ndim.gt.2) then
              call cal_Dedd(Eray0(i,j,k-1),Fray0(i,j,k-1,1:ndim),Dedd_3d(:,:,1),Dedd_dE_3d(:,:,1),Dedd_dF_3d(:,:,:,1) , Derive_chi)
              call cal_Dedd(Eray0(i,j,k+1),Fray0(i,j,k+1,1:ndim),Dedd_3d(:,:,2),Dedd_dE_3d(:,:,2),Dedd_dF_3d(:,:,:,2) , Derive_chi)
           endif
        endif


        ! Initialise the geometry
        do idim=1,ndim
           ip=i ; jp=j ; kp=k
           if(idim==1) ip=i+1
           if(idim==2) jp=j+1
           if(idim==3) kp=k+1
           dSloc(1,idim) = dS(i ,j ,k ,idim)
           dSloc(2,idim) = dS(ip,jp,kp,idim)
        enddo


        !******************************************************************
        !                     Compute Fluxes
        !******************************************************************
        rayon = 1.
        if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') then
           rayon(2,1:2,1:3) = xcloc(i  ,1)
           rayon(2,1  ,1  ) =  xloc(i  ,1)
           rayon(2,2  ,1  ) =  xloc(i+1,1)
        endif

        if (geom_dir(3) == 'phi_s' ) then
           rayon(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
           rayon(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
           rayon(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
           rayon(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
           rayon(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
           rayon(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
        endif

        flux_F = 0.
        flux_P = 0.
        do idim=1,ndim
           flux_F(1,idim) = (+  lp (1,idim)            *FrayM(idim,1,idim)                 &
                             -  lm (1,idim)            *dFray(idim,1,idim)                 &
                             -c*lmp(1,idim)*eps(1,idim)*ErayM(     1,idim)                 &
                             +c*lmp(1,idim)*eps(1,idim)*dEray(     1,idim) )*dSloc(1,idim)
           flux_F(2,idim) = (-  lm (2,idim)            *FrayP(idim,2,idim)                 &
                             +  lp (2,idim)            *dFray(idim,2,idim)                 &
                             +c*lmp(2,idim)*eps(2,idim)*ErayP(     2,idim)                 &
                             -c*lmp(2,idim)*eps(2,idim)*dEray(     2,idim) )*dSloc(2,idim)
        enddo

        do jdim = 1,ndim
           imin=1
           imax=ndim
           ipas=1
           if (jdim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
              imin=2
              idim=1
              flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                      -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                      +c *  lmp(1,idim)*eps(1,idim)                          * dFray(jdim     ,1,idim) &
                                      +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)*(PrayP(jdim,idim,1,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
              flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                      +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                      -c *  lmp(2,idim)*eps(2,idim)                          * dFray(jdim     ,2,idim) &
                                      +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)*(PrayM(jdim,idim,2,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
           endif
           if (jdim.eq.2 .and. geom_dir(2) == 'the_s') then
              ipas=2
              idim=2
              flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                      -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                      +c *  lmp(1,idim)*eps(1,idim)                          * FrayP(jdim     ,1,idim) &
                                      +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayP(jdim,idim,1,idim) ) *dV(i,j,k)/(xloc(j+1,2)-xloc(j,2)) !/xcloc(i,1)
              flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                      +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                      -c *  lmp(2,idim)*eps(2,idim)                          * FrayM(jdim     ,2,idim) &
                                      +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayM(jdim,idim,2,idim) ) *dV(i,j,k)/(xloc(j+1,2)-xloc(j,2)) !/xcloc(i,1)
           endif
           do idim=imin,imax,ipas
              flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                      -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                      +c *  lmp(1,idim)*eps(1,idim)                          * dFray(jdim     ,1,idim) &
                                      +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)*(PrayP(jdim,idim,1,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dSloc(1,idim) *rayon(jdim,1,idim)
              flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                      +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                      -c *  lmp(2,idim)*eps(2,idim)                          * dFray(jdim     ,2,idim) &
                                      +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)*(PrayM(jdim,idim,2,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dSloc(2,idim) *rayon(jdim,2,idim)
           enddo
        enddo


        flux_F_tot = 0.
        flux_P_tot = 0.
        do idim=1,ndim
           do iface=1,2
              flux_F_tot = flux_F_tot + signe(iface)*dt*flux_F(iface,idim)
              do jdim=1,ndim
                 flux_P_tot(jdim) = flux_P_tot(jdim) + signe(iface)*dt*flux_P(jdim,iface,idim) !*rayon(jdim,iface,idim)
              enddo
           enddo
        enddo

        flux_F_tot = flux_F_tot/dV(i,j,k)
        do idim=1,ndim
           flux_P_tot(idim) = flux_P_tot(idim)/dV(i,j,k)/rayon(idim,2,3)
        enddo


        ! Fill matrix and vector
        !******************************************************************
        vect = 0.
        mat  = 0.
        do idim=1,5
           mat(idim,idim) = 1.
        enddo
        
        ! Er equation without the source term
        mat(2,1) = 0.
        mat(2,2) = 1. 
        do idim = 1,ndim
           mat(2,2     ) = mat(2,2) - dt/dV(i,j,k)*c*(lmp(1,idim)*eps(1,idim)*dSloc(1,idim)+lmp(2,idim)*eps(2,idim)*dSloc(2,idim))
           mat(2,2+idim) =            dt/dV(i,j,k)  *(lm (1,idim)            *dSloc(1,idim)+lp (2,idim)            *dSloc(2,idim))
        enddo

        vect(2) = -Erayold(i,j,k) + mat(2,2)*Eray0(i,j,k) + flux_F_tot
        do idim = 1,ndim
           vect(2) = vect(2) + mat(2,2+idim)*Fray0(i,j,k,idim)
        enddo

        ! Tgaz equation (using the coefficients from the Er equation without the source term)
        call heat_capacity(rho(i,j,k),Tgaz0(i,j,k),conv,1,1,1,1,1,1)
        conv=1./conv

        mat(1,1  ) = 1.
        mat(1,2:5) = mat(2,2:5)* conv 
        
        vect(1)    = vect(2)*conv -Tgazold(i,j,k)+mat(1,1)*Tgaz0(i,j,k)
            
        ! Finally include the energy source term: rho c (kappa_em aT^4-kappa_abs E)
        
        !  T4 = Tgazold(i,j,k)**4 *(1. + 4.*(Tgaz(i,j,k)-tgazold(i,j,k))/Tgazold(i,j,k)) 
        !  T4 = -3.*Tgazold(i,j,k)**4 + 4.* Tgazold(i,j,k)**3*Tgaz(i,j,k)
        
        mat(2,1) =          - dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*derive_source_ray(Tgaz0(i,j,k))
        mat(2,2) = mat(2,2) + dt*kappa_abs_ray(i,j,k)*rho(i,j,k)*c
        vect (2) = vect (2) - dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*source_ray(Tgaz0(i,j,k))
        vect (2) = vect (2) + dt*kappa_abs_ray(i,j,k)*rho(i,j,k)*c*Eray0(i,j,k) 

        ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
        do idim = 1,ndim
           mat(2+idim,1     ) = 0.
           mat(2+idim,2     ) = 0.
           mat(2+idim,2+idim) = 1.
           jmin=1
           jmax=ndim
           jpas=1
           if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
              jmin=2
              jdim=1
              mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd(idim,jdim)
              mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/(xloc(i+1,1)-xloc(i,1))* &
                   (lmp(1,jdim)*eps(1,jdim) &
                   +lmp(2,jdim)*eps(2,jdim) )  
           endif
           if(idim.eq.2 .and. geom_dir(2) == 'the_s') then
              jpas=2
              jdim=1
              mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd(idim,jdim)
              mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                   (lmp(1,jdim)*eps(1,jdim) &
                   +lmp(2,jdim)*eps(2,jdim) )  
           endif
           do jdim = jmin,jmax,jpas
              mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/dV(i,j,k)* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd(idim,jdim)
              mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/dV(i,j,k)* &
                   (lmp(1,jdim)*eps(1,jdim)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) &
                   +lmp(2,jdim)*eps(2,jdim)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) )  
           enddo
        enddo

        ! Geometric sources
        if (cylindrique) then
           iloc=3
           if(geom_dir(2) == 'phi_c') iloc=2
           mat(3,2) = mat(3,2) + c2*dt/dV(i,j,k)* (Dedd(1,1) - Dedd(iloc,iloc)) * (dSloc(2,1)-dSloc(1,1))
        endif
        if (spherique) then
                                      mat(3,2) = mat(3,2) + c2*dt/dV(i,j,k)*                                 &
                (2.*Dedd(1,1) - Dedd(2,2) - Dedd(3,3)) * 0.5*(dSloc(2,1)-dSloc(1,1))
           if(geom_dir(2) == 'the_s') mat(4,2) = mat(4,2) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*&
                (               Dedd(2,2) - Dedd(3,3)) * 0.5*(dSloc(2,1)-dSloc(1,1))
        endif


        do idim=1,ndim
           vect(2+idim) = -Frayold(i,j,k,idim) + flux_P_tot(idim)

           ! Include the flux source term: -kappa_abs rho c F
           mat(2+idim,2+idim) = mat(2+idim,2+idim) + alpha*c*dt*(kappa_abs_ray(i,j,k)+kappa_sca_ray(i,j,k))*rho(i,j,k)
           vect(2+idim)  = vect(2+idim)  !+ dt*(kappa_abs_ray(i,j,k)+kappa_sca_ray(i,j,k))*rho(i,j,k)*c*(1.-alpha)*Frayold(i,j,k,idim)

           vect(2+idim) = vect(2+idim) + mat(2+idim,2)*Eray0(i,j,k)
           do jdim=1,ndim
              vect(2+idim) = vect(2+idim) + mat(2+idim,2+jdim)*Fray0(i,j,k,jdim)
           enddo
        enddo
        

        !*************************************************
        !     Derivative terms in the matrix
        !
        !      !!!!       CAUTION       !!!!
 	     ! To be done after mat(*,*)*Eray0 + mat(*,*)*Fray0
        !*************************************************
        ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
        do idim = 1,ndim
           jmin=1
           jmax=ndim
           jpas=1
           if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
              jmin=2
              jdim=1
              mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dE(idim,jdim     )  ! d/dEray
              do kdim=1,ndim
                 mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
              enddo
           endif
           if(idim.eq.2 .and. geom_dir(2) == 'the_s') then
              jpas=2
              jdim=2
              mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dE(idim,jdim     )  ! d/dEray
              do kdim=1,ndim
                 mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
              enddo
           endif
           do jdim = jmin,jmax,jpas
              ! The Dedd derivatives
              mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/dV(i,j,k)* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd_dE(idim,jdim     )  ! d/dEray
              do kdim=1,ndim
                 mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/dV(i,j,k)* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
              enddo
           enddo
        enddo
        
        ! Geometric sources
        if(cylindrique) then
           iloc=3
           if(geom_dir(2) == 'phi_c') iloc=2
              mat(3,2     ) = mat(3,2     ) + c2*dt/dV(i,j,k)* (Dedd_dE(1,1) - Dedd_dE(iloc,iloc)) * (dSloc(2,1)-dSloc(1,1))
           do idim=1,ndim
              mat(3,2+idim) = mat(3,2+idim) + c2*dt/dV(i,j,k)* (Dedd_dF(1,1,idim) - Dedd_dF(iloc,iloc,idim)) * (dSloc(2,1)-dSloc(1,1))
           enddo
        endif
        if (spherique) then
              mat(3,2     ) = mat(3,2     ) + c2*dt/dV(i,j,k)* (2.*Dedd_dE(1,1     ) - Dedd_dE(2,2     ) - Dedd_dE(3,3     )) * 0.5*(dSloc(2,1)-dSloc(1,1))
           do idim=1,ndim
              mat(3,2+idim) = mat(3,2+idim) + c2*dt/dV(i,j,k)* (2.*Dedd_dF(1,1,idim) - Dedd_dF(2,2,idim) - Dedd_dF(3,3,idim)) * 0.5*(dSloc(2,1)-dSloc(1,1))
           enddo
           if(geom_dir(2) == 'the_s') then
                 mat(4,2     ) = mat(4,2     ) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*(Dedd_dE(2,2     ) - Dedd_dE(3,3     )) * 0.5*(dSloc(2,1)-dSloc(1,1))
              do idim=1,ndim
                 mat(4,2+idim) = mat(4,2+idim) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*(Dedd_dF(2,2,idim) - Dedd_dF(3,3,idim)) * 0.5*(dSloc(2,1)-dSloc(1,1))
              enddo
           endif
        endif


        !******************************************************************
        !                     End of Flux computations
        !******************************************************************


        ! non-diagonal terms

        a1d = 0. ; a2d = 0. ; a3d = 0.

        ! Temperature equation
        a1d(1,2)  =  dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*conv*dSloc(1,1) !*ErayM(1)   
        a1d(1,3)  = -dt/dV(i,j,k)*   lp(1,1)         *conv*dSloc(1,1) !*FrayM(1,1) 

        ! Radiative energy equation
        a1d(2,2)  =  dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*dSloc(1,1)!*ErayM(1)                
        a1d(2,3)  = -dt/dV(i,j,k)*   lp(1,1)         *dSloc(1,1)!*FrayM(1,1)

        ! Radiative flux equation
        do idim=1,ndim
           if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
              a1d(2+idim,2) = -dt/(xloc(i+1,1)-xloc(i,1))*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) * (DeddM(idim,1,1,1)+Dedd_dE_1d(idim,1,1)) *Rayon(idim,1,1)/Rayon(idim,2,3)
              do jdim=1,ndim
                 a1d(2+idim,2+jdim) = -dt/(xloc(i+1,1)-xloc(i,1)) *   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) *  Dedd_dF_1d(idim,1,jdim,1)      *Rayon(idim,1,1)/Rayon(idim,2,3)
              enddo
              a1d(2+idim,2+idim) = a1d(2+idim,2+idim) + dt/(xloc(i+1,1)-xloc(i,1))*c*lmp(1,1)*eps(1,1)                                                *Rayon(idim,1,1)/Rayon(idim,2,3)
           else
              a1d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) * (DeddM(idim,1,1,1)+Dedd_dE_1d(idim,1,1))*dSloc(1,1) *Rayon(idim,1,1)/Rayon(idim,2,3)
              do jdim=1,ndim
                 a1d(2+idim,2+jdim) = -dt/dV(i,j,k)*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) *  Dedd_dF_1d(idim,1,jdim,1)*dSloc(1,1)       *Rayon(idim,1,1)/Rayon(idim,2,3)
              enddo
              a1d(2+idim,2+idim) = a1d(2+idim,2+idim) + dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*dSloc(1,1)                                                *Rayon(idim,1,1)/Rayon(idim,2,3)
           endif
        enddo

        ! Temperature equation
        a1d(1,nvar_ray+2)  =  dt/dV(i,j,k)*c*lmp(2,1)*eps(2,1)*conv*dSloc(2,1) !ErayP(2)   
        a1d(1,nvar_ray+3)  = -dt/dV(i,j,k)*   lm(2,1)         *conv*dSloc(2,1) !FrayP(1,2) 

        ! Radiative energy equation
        a1d(2,nvar_ray+2)  =  dt/dV(i,j,k)*c*lmp(2,1)*eps(2,1)*dSloc(2,1) !*ErayP(2)
        a1d(2,nvar_ray+3)  = -dt/dV(i,j,k)*   lm(2,1)         *dSloc(2,1) !*FrayP(1,2)

        ! Radiative flux equation
        do idim=1,ndim
           if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
              a1d(2+idim,nvar_ray+2) = dt/(xloc(i+1,1)-xloc(i,1))*   c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * (DeddP(idim,1,2,1)+Dedd_dE_1d(idim,1,2)) *Rayon(idim,2,1)/Rayon(idim,2,3)
              do jdim=1,ndim
                 a1d(2+idim,nvar_ray+2+jdim) = dt/(xloc(i+1,1)-xloc(i,1))* c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * Dedd_dF_1d(idim,1,jdim,2)          *Rayon(idim,2,1)/Rayon(idim,2,3)
              enddo
              a1d(2+idim,nvar_ray+2+idim) = a1d(2+idim,nvar_ray+2+idim) + dt/(xloc(i+1,1)-xloc(i,1))*c* lmp(2,1)*eps(2,1)                                     *Rayon(idim,2,1)/Rayon(idim,2,3)
           else
              a1d(2+idim,nvar_ray+2) = dt/dV(i,j,k)*   c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * (DeddP(idim,1,2,1)+Dedd_dE_1d(idim,1,2))*dSloc(2,1) *Rayon(idim,2,1)/Rayon(idim,2,3)
              do jdim=1,ndim
                 a1d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)* c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * Dedd_dF_1d(idim,1,jdim,2)*dSloc(2,1)          *Rayon(idim,2,1)/Rayon(idim,2,3)
              enddo
              a1d(2+idim,nvar_ray+2+idim) = a1d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c* lmp(2,1)*eps(2,1)*dSloc(2,1)                                     *Rayon(idim,2,1)/Rayon(idim,2,3)
           endif
        enddo

        if(ndim.gt.1) then
           ! coupling with j-1 terms
           a2d(1,1) = 0.
           a2d(1,2) =  dt/dV(i,j,k)*c*lmp(1,2)*eps(1,2)*conv*dSloc(1,2)
           a2d(1,3) = 0.
           a2d(1,4) = -dt/dV(i,j,k)*   lp(1,2)         *conv*dSloc(1,2)

           a2d(2,1) = 0.
           a2d(2,2) =  dt/dV(i,j,k)*c*lmp(1,2)*eps(1,2)*dSloc(1,2)
           a2d(2,3) = 0.
           a2d(2,4) = -dt/dV(i,j,k)*   lp(1,2)         *dSloc(1,2)

           do idim=1,ndim
              if(idim.eq.2 .and. geom_dir(2)=='the_s') then
                 a2d(2+idim,2) = -dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * (DeddM(idim,2,1,2)+Dedd_dE_2d(idim,2,1))  *Rayon(idim,1,2)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a2d(2+idim,2+jdim) = -dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,1)         *Rayon(idim,1,2)/Rayon(idim,2,3)
                 enddo
                 a2d(2+idim,2+idim) = a2d(2+idim,2+idim) + dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*c* lmp(1,2)*eps(1,2)                                                *Rayon(idim,1,2)/Rayon(idim,2,3)
              else
                 a2d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * (DeddM(idim,2,1,2)+Dedd_dE_2d(idim,2,1))*dSloc(1,2)      *Rayon(idim,1,2)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a2d(2+idim,2+jdim) = -dt/dV(i,j,k)*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,1)*dSloc(1,2)             *Rayon(idim,1,2)/Rayon(idim,2,3)
                 enddo
                 a2d(2+idim,2+idim) = a2d(2+idim,2+idim) + dt/dV(i,j,k)*c* lmp(1,2)*eps(1,2)*dSloc(1,2)                                                    *Rayon(idim,1,2)/Rayon(idim,2,3)
              endif
           enddo

           ! coupling with j+1 terms
           a2d(1,nvar_ray+1) = 0.
           a2d(1,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,2)*eps(2,2)*conv*dSloc(2,2)
           a2d(1,nvar_ray+3) = 0.
           a2d(1,nvar_ray+4) = -dt/dV(i,j,k)*   lm(2,2)         *conv*dSloc(2,2)

           a2d(2,nvar_ray+1) = 0.
           a2d(2,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,2)*eps(2,2)*dSloc(2,2)
           a2d(2,nvar_ray+3) = 0.
           a2d(2,nvar_ray+4) = -dt/dV(i,j,k)*   lm(2,2)         *dSloc(2,2)

           do idim=1,ndim
              if(idim.eq.2 .and. geom_dir(2)=='the_s') then
                 a2d(2+idim,nvar_ray+2) = dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * (DeddP(idim,2,2,2)+Dedd_dE_2d(idim,2,2))*Rayon(idim,2,2)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a2d(2+idim,nvar_ray+2+jdim) = dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,2)       *Rayon(idim,2,2)/Rayon(idim,2,3)
                 enddo
                 a2d(2+idim,nvar_ray+2+idim) = a2d(2+idim,nvar_ray+2+idim) + dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*c* lmp(2,2)*eps(2,2)                                    *Rayon(idim,2,2)/Rayon(idim,2,3)
              else
                 a2d(2+idim,nvar_ray+2) = dt/dV(i,j,k)*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * (DeddP(idim,2,2,2)+Dedd_dE_2d(idim,2,2))*dSloc(2,2)  *Rayon(idim,2,2)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a2d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,2)*dSloc(2,2)         *Rayon(idim,2,2)/Rayon(idim,2,3)
                 enddo
                 a2d(2+idim,nvar_ray+2+idim) = a2d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c* lmp(2,2)*eps(2,2)*dSloc(2,2)                                      *Rayon(idim,2,2)/Rayon(idim,2,3)
              endif
           enddo

           if(ndim.gt.2) then
              ! coupling with k-1 terms
              a3d(1,1) = 0.
              a3d(1,2) =  dt/dV(i,j,k)*c*lmp(1,3)*eps(1,3)*conv*dSloc(1,3)
              a3d(1,3) = 0.
              a3d(1,4) = 0.
              a3d(1,5) = -dt/dV(i,j,k)*   lp(1,3)         *conv*dSloc(1,3)

              a3d(2,1) = 0.
              a3d(2,2) =  dt/dV(i,j,k)*c*lmp(1,3)*eps(1,3)*dSloc(1,3)
              a3d(2,3) = 0.
              a3d(2,4) = 0.
              a3d(2,5) = -dt/dV(i,j,k)  * lp(1,3)         *dSloc(1,3)

              do idim=1,ndim
                 a3d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,3)*eps(1,3)**2+(1.-eps(1,3)**2)/2 ) * (DeddM(idim,3,1,3)+Dedd_dE_3d(idim,3,1))*dSloc(1,3)      *Rayon(idim,1,3)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a3d(2+idim,2+jdim) =  -dt/dV(i,j,k)*   c2*( lp(1,3)*eps(1,3)**2+(1.-eps(1,3)**2)/2 ) * Dedd_dF_3d(idim,3,jdim,1)*dSloc(1,3)            *Rayon(idim,1,3)/Rayon(idim,2,3)
                 enddo
                 a3d(2+idim,2+idim) = a3d(2+idim,2+idim) + dt/dV(i,j,k)*c * lmp(1,3)*eps(1,3)*dSloc(1,3)                                                   *Rayon(idim,1,3)/Rayon(idim,2,3)
              enddo

              ! coupling with k+1 terms
              a3d(1,nvar_ray+1) = 0.
              a3d(1,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,3)*eps(2,3)*conv*dSloc(2,3)
              a3d(1,nvar_ray+3) = 0.
              a3d(1,nvar_ray+4) = 0.
              a3d(1,nvar_ray+5) = -dt/dV(i,j,k)  * lm(2,3)         *conv*dSloc(2,3)

              a3d(2,nvar_ray+1) = 0.
              a3d(2,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,3)*eps(2,3)*dSloc(2,3)
              a3d(2,nvar_ray+3) = 0.
              a3d(2,nvar_ray+4) = 0.
              a3d(2,nvar_ray+5) = -dt/dV(i,j,k)  * lm(2,3)         *dSloc(2,3)

              do idim=1,ndim
                 a3d(2+idim,nvar_ray+2) =  dt/dV(i,j,k)*   c2*(-lm(2,3)*eps(2,3)**2+(1.-eps(2,3)**2)/2 )  *(DeddP(idim,3,2,3)+Dedd_dE_3d(idim,3,2))*dSloc(2,3) *Rayon(idim,2,3)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a3d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)*   c2*(-lm(2,3)*eps(2,3)**2+(1.-eps(2,3)**2)/2 )  *Dedd_dF_3d(idim,3,jdim,2)*dSloc(2,3)         *Rayon(idim,2,3)/Rayon(idim,2,3)
                 enddo
                 a3d(2+idim,nvar_ray+2+idim) = a3d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c * lmp(2,3)*eps(2,3)*dSloc(2,3)                                     *Rayon(idim,2,3)/Rayon(idim,2,3)
              enddo
           endif
        endif

        !***********************************************************
        !              Boundary conditions
        !***********************************************************
        if(i==1) then
           if(icl_ray(1)==1) then
              mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,     1:     nvar_ray)
           else if(icl_ray(1)==3) then
              mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,     1:     nvar_ray)
              mat(1:nvar_ray,3     ) = mat(1:nvar_ray,3     ) - 2.* a1d(1:nvar_ray,     3          )
           endif
           if(icl_ray(1).ne.2 .and. icl_ray(1).ne.0) a1d(1:nvar_ray,1:nvar_ray) = 0.
        else if(i==nx(1)) then
           if(icl_ray(2)==1) then
              mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
           else if(icl_ray(2)==3) then
              mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
              mat(1:nvar_ray,3     ) = mat(1:nvar_ray,3     ) - 2.* a1d(1:nvar_ray,nvar_ray+3          )
           endif
           if(icl_ray(2).ne.2 .and. icl_ray(2).ne.0) a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
        endif
        if(ndim.gt.1) then
           if(j==1) then
              if(icl_ray(3)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,     1:     nvar_ray)
              else if(icl_ray(3)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,     1:     nvar_ray)
                 mat(1:nvar_ray,4     ) = mat(1:nvar_ray,4     ) - 2.* a2d(1:nvar_ray,     4          )
              endif
              if(icl_ray(3).ne.2 .and. icl_ray(3).ne.0) a2d(1:nvar_ray,1:nvar_ray) = 0.
           else if(j==nx(2)) then
              if(icl_ray(4)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
              else if(icl_ray(4)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 mat(1:nvar_ray,4     ) = mat(1:nvar_ray,4     ) - 2.* a2d(1:nvar_ray,nvar_ray+4          )
              endif
              if(icl_ray(4).ne.2 .and. icl_ray(4).ne.0) a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
           endif
        endif
        if(ndim.gt.2) then
           if(k==1) then
              if(icl_ray(5)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,     1:     nvar_ray)
              else if(icl_ray(5)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,     1:     nvar_ray)
                 mat(1:nvar_ray,5     ) = mat(1:nvar_ray,5     ) - 2.* a3d(1:nvar_ray,     5          )
              endif
              if(icl_ray(5).ne.2 .and. icl_ray(5).ne.0) a3d(1:nvar_ray,1:nvar_ray) = 0.
           else if(k==nx(3)) then
              if(icl_ray(6)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
              else if(icl_ray(6)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 mat(1:nvar_ray,5     ) = mat(1:nvar_ray,5     ) - 2.* a3d(1:nvar_ray,nvar_ray+5          )
              endif
              if(icl_ray(6).ne.2 .and. icl_ray(6).ne.0) a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
           endif
        endif

        !**************  Normalised equations   ****************!
        !*******************************************************!
        !**        eqi / Xoldi et Aij=dFi/dXj *Xoldj          **!
        !**       to get     Aij (deltaX/Xold)j = eqi         **!
        !*******************************************************!
        !*******************************************************!
        vect(1  ) = vect(1  )/   Tgazold(i,j,k)
        vect(2  ) = vect(2  )/   Erayold(i,j,k)
        vect(3:5) = vect(3:5)/(c*Erayold(i,j,k))

        mat(1  ,1:5) = mat(1  ,1:5)/   Tgazold(i,j,k)
        mat(2  ,1:5) = mat(2  ,1:5)/   Erayold(i,j,k)
        mat(3:5,1:5) = mat(3:5,1:5)/(c*Erayold(i,j,k))

        mat(1:5,1  ) = mat(1:5,1  ) *   Tgazold(i,j,k)
        mat(1:5,2  ) = mat(1:5,2  ) *   Erayold(i,j,k)
        mat(1:5,3:5) = mat(1:5,3:5) * c*Erayold(i,j,k)

        a1d(1     ,1:2*nvar_ray)=a1d(1     ,1:2*nvar_ray)/   Tgazold(i,j,k)
        a1d(2     ,1:2*nvar_ray)=a1d(2     ,1:2*nvar_ray)/   Erayold(i,j,k)
        a1d(3:nvar_ray,1:2*nvar_ray)=a1d(3:nvar_ray,1:2*nvar_ray)/(c*Erayold(i,j,k))

        do i1=1,2
           a1d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) = a1d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) *   Tgazold(i+2*i1-3,j,k) 
           a1d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) = a1d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) *   Erayold(i+2*i1-3,j,k)
           a1d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) = a1d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) * c*Erayold(i+2*i1-3,j,k)
        enddo

        if(ndim.gt.1) then
           a2d(1     ,1:2*nvar_ray)=a2d(1     ,1:2*nvar_ray)/   Tgazold(i,j,k)
           a2d(2     ,1:2*nvar_ray)=a2d(2     ,1:2*nvar_ray)/   Erayold(i,j,k)
           a2d(3:nvar_ray,1:2*nvar_ray)=a2d(3:nvar_ray,1:2*nvar_ray)/(c*Erayold(i,j,k))

           do i1=1,2
              a2d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) = a2d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) *   Tgazold(i,j+2*i1-3,k) 
              a2d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) = a2d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) *   Erayold(i,j+2*i1-3,k)
              a2d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) = a2d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) * c*Erayold(i,j+2*i1-3,k)
           enddo
           
           if(ndim.gt.2) then
              a3d(1     ,1:2*nvar_ray)=a3d(1     ,1:2*nvar_ray)/   Tgazold(i,j,k)
              a3d(2     ,1:2*nvar_ray)=a3d(2     ,1:2*nvar_ray)/   Erayold(i,j,k)
              a3d(3:nvar_ray,1:2*nvar_ray)=a3d(3:nvar_ray,1:2*nvar_ray)/(c*Erayold(i,j,k))
              
              do i1=1,2
                 a3d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) = a3d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) *   Tgazold(i,j,k+2*i1-3) 
                 a3d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) = a3d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) *   Erayold(i,j,k+2*i1-3)
                 a3d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) = a3d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) * c*Erayold(i,j,k+2*i1-3)
              enddo
           endif
        endif

        vect(1:5) = - vect(1:5)

        call essai_inv5x5(mat,inv)
        vect(1:nvar_ray                 ) = matmul( inv(1:nvar_ray,1:nvar_ray),vect(1:nvar_ray                 ) )
        mat (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),mat (1:nvar_ray,     1:     nvar_ray) )
        a1d (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a1d (1:nvar_ray,     1:     nvar_ray) )
        a1d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a1d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) )
        if(ndim.gt.1) then
           a2d (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a2d (1:nvar_ray,     1:     nvar_ray) )
           a2d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a2d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) )
           if(ndim.gt.2) then
              a3d (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a3d (1:nvar_ray,     1:     nvar_ray) )
              a3d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a3d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) )
           endif
        endif
        
        
        FFF = 0.

        do i1=1,nvar_ray
           FFF(i1) =           mat(i1,     1) * deltaTgaz(i  ,j,k)/Tgazold(i  ,j,k) + mat(i1,     2) * deltaEray(i  ,j,k)/Erayold(i  ,j,k)
           FFF(i1) = FFF(i1) + a1d(i1,     1) * deltaTgaz(i-1,j,k)/Tgazold(i-1,j,k) + a1d(i1,     2) * deltaEray(i-1,j,k)/Erayold(i-1,j,k)
           FFF(i1) = FFF(i1) + a1d(i1,nvar_ray+1) * deltaTgaz(i+1,j,k)/Tgazold(i+1,j,k) + a1d(i1,nvar_ray+2) * deltaEray(i+1,j,k)/Erayold(i+1,j,k)
           FFF(i1) = FFF(i1) - vect(i1)
           do idim=1,ndim
              FFF(i1) = FFF(i1) + mat(i1,     2+idim) * deltaFray(i  ,j,k,idim)/c/Erayold(i  ,j,k)
              FFF(i1) = FFF(i1) + a1d(i1,     2+idim) * deltaFray(i-1,j,k,idim)/c/Erayold(i-1,j,k)
              FFF(i1) = FFF(i1) + a1d(i1,nvar_ray+2+idim) * deltaFray(i+1,j,k,idim)/c/Erayold(i+1,j,k) 
           enddo
        enddo

        if(ndim.gt.1) then
           do i1=1,nvar_ray
              FFF(i1) = FFF(i1) + a2d(i1,     1) * deltaTgaz(i,j-1,k)/Tgazold(i,j-1,k) + a2d(i1,     2) * deltaEray(i,j-1,k)/Erayold(i,j-1,k)
              FFF(i1) = FFF(i1) + a2d(i1,nvar_ray+1) * deltaTgaz(i,j+1,k)/Tgazold(i,j+1,k) + a2d(i1,nvar_ray+2) * deltaEray(i,j+1,k)/Erayold(i,j+1,k)
              do idim=1,ndim
                 FFF(i1) = FFF(i1) + a2d(i1,     2+idim) * deltaFray(i,j-1,k,idim)/c/Erayold(i,j-1,k)
                 FFF(i1) = FFF(i1) + a2d(i1,nvar_ray+2+idim) * deltaFray(i,j+1,k,idim)/c/Erayold(i,j+1,k) 
              enddo
           enddo
           if(ndim.gt.2) then
              do i1=1,nvar_ray
                 FFF(i1) = FFF(i1) + a3d(i1,     1) * deltaTgaz(i,j,k-1)/Tgazold(i,j,k-1) + a3d(i1,     2) * deltaEray(i,j,k-1)/Erayold(i,j,k-1)
                 FFF(i1) = FFF(i1) + a3d(i1,nvar_ray+1) * deltaTgaz(i,j,k+1)/Tgazold(i,j,k+1) + a3d(i1,nvar_ray+2) * deltaEray(i,j,k+1)/Erayold(i,j,k+1)
                 do idim=1,ndim
                    FFF(i1) = FFF(i1) + a3d(i1,     2+idim) * deltaFray(i,j,k-1,idim)/c/Erayold(i,j,k-1)
                    FFF(i1) = FFF(i1) + a3d(i1,nvar_ray+2+idim) * deltaFray(i,j,k+1,idim)/c/Erayold(i,j,k+1) 
                 enddo
              enddo
           endif
        endif

        resF = max(abs(FFF(1)),abs(FFF(2)))
        do idim=1,ndim
           resF = max(resF,abs(FFF(2+idim)))
        enddo
        if (resmaxF.lt.resF) then 
           resmaxF=max(resF,resmaxF)
           imax_res=i
           jmax_res=j
           kmax_res=k
           if(resmaxF.eq.abs(FFF(1))) then
              var_pasconv='Tg'
           else if(resmaxF.eq.abs(FFF(2))) then
              var_pasconv='Er'
           else if(resmaxF.eq.abs(FFF(3))) then
              var_pasconv='Fx'
           else if(ndim.gt.1) then
              if(resmax.eq.abs(FFF(4))) var_pasconv='Fy'
           else if(ndim.gt.2) then
              if(resmax.eq.abs(FFF(5))) var_pasconv='Fz'
           endif
        endif
        resmaxF = max(resF,resmaxF)

        res_norme2 = res_norme2 + FFF(1)**2 + FFF(2)**2
        do idim=1,ndim
           res_norme2 = res_norme2 + FFF(2+idim)**2
        enddo
        
        do i1=1,nvar_ray
           vect(i1) = -FFF(i1)+mat(i1,1)*deltaTgaz(i,j,k)/Tgazold(i,j,k)+mat(i1,2)*deltaEray(i,j,k)/Erayold(i,j,k)
           do idim=1,ndim
              vect(i1) = vect(i1)+mat(i1,2+idim)*deltaFray(i,j,k,idim)/c/Erayold(i,j,k)
           enddo
        enddo

        vect(1) = vect(1) - mat(1,2) * deltaEray(i,j,k)/Erayold(i,j,k)
        do idim=1,ndim
           vect(1) = vect(1) - mat(1,2+idim) * deltaFray(i,j,k,idim)/c/Erayold(i,j,k)
        enddo
        vect(1) = vect(1) / mat(1,1) 
        deltaTgaz(i,j,k  ) = ( deltaTgaz(i,j,k  )/  Tgazold(i,j,k) + omega * (vect(1)-deltaTgaz(i,j,k  )/  Tgazold(i,j,k)) ) *  Tgazold(i,j,k) 
        

        vect(2) = vect(2) - mat(2,1) * deltaTgaz(i,j,k)/Tgazold(i,j,k)
        do idim=1,ndim
           vect(2) = vect(2) - mat(2,2+idim) * deltaFray(i,j,k,idim)/c/Erayold(i,j,k)
        enddo
        vect(2) = vect(2) / mat(2,2)
        deltaEray(i,j,k  ) = ( deltaEray(i,j,k  )/  Erayold(i,j,k) + omega * (vect(2)-deltaEray(i,j,k  )/  Erayold(i,j,k)) ) *  Erayold(i,j,k) 

        do idim=1,ndim
           vect(2+idim) = vect(2+idim) - mat(2+idim,1) * deltaTgaz(i,j,k)/Tgazold(i,j,k) - mat(2+idim,2) * deltaEray(i,j,k)/Erayold(i,j,k)
           do jdim=1,ndim
              if(jdim.ne.idim) vect(2+idim) = vect(2+idim) - mat(2+idim,2+jdim) * deltaFray(i,j,k,jdim)/c/Erayold(i,j,k)
           enddo
           vect(2+idim) = vect(2+idim) / mat(2+idim,2+idim) 
           deltaFray(i,j,k,idim) = ( deltaFray(i,j,k,idim)/c/Erayold(i,j,k) + omega * (vect(2+idim)-deltaFray(i,j,k,idim)/c/Erayold(i,j,k)) ) *  c*Erayold(i,j,k) 
        enddo
            
     enddo   ! End of loop over all i,j,k cells
     enddo
     enddo

     update = .true. ; alpha_var_new = alpha_var
     do while (update)

        update = .false. ; varE_max = 0. ; varT_max = 0. ; varF_max = 0.
        do k = 1,nx(3)
        do j = 1,nx(2)
        do i = 1,nx(1)

           Tgaztemp         = Tgaz0(i,j,k       ) + alpha_var*deltaTgaz(i,j,k       )
           Eraytemp         = Eray0(i,j,k       ) + alpha_var*deltaEray(i,j,k       )
           Fraytemp(1:ndim) = Fray0(i,j,k,1:ndim) + alpha_var*deltaFray(i,j,k,1:ndim)
           
           
           if (Eraytemp < 0. ) then
              write(*,*) 'Negative Eray...',i,j,k
              write(*,*)  Eraytemp,  Eray0(i,j,k)
              alpha_var_new = alpha_var/2.
              write(*,*) 'alpha_var =',alpha_var
              !update = .true.
              read(*,*)
           endif
           
           xx = 0.
           do idim = 1,ndim
              xx = xx +  Fraytemp(idim)**2.
           enddo
           xx = sqrt(xx)
           xm = un_moins*c*Eraytemp
           
           if (xx > xm) then
              Iconv = 0.
              !resume = 1
              !write(*,*) '...Resume...'
              write(*,*) '   i,j,k=',mype,i,j,k
              write(*,*) '   Eray(i,j,k       )        ',mype,Eraytemp
              write(*,*) '   Fray(i,j,k,1:ndim),normeF=',mype,Fraytemp(1:ndim),xx
              write(*,*) '   f   (i,j,k,1:ndim),normef=',mype,Fraytemp(1:ndim)/c/Eraytemp,xx/c/Eraytemp
              write(*,*) '   normef-1.=',mype,xx/c/Eraytemp-1.
              alpha_var_new = alpha_var/2.
              write(*,*) 'alpha_var =',alpha_var_new
              !update = .true.
              !read(*,*)
           endif
           
           
        enddo
        enddo
        enddo

#if PARA==1
        call MPI_ALLREDUCE(alpha_var_new,alpha_var_glob,1,MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,info)
#else
        alpha_var_glob = alpha_var_new
#endif
        if (alpha_var .ne. alpha_var_glob) update = .true. 
        alpha_var = alpha_var_glob
        alpha_var_new = alpha_var

     enddo  ! end of do while (update)

     do k = 1,nx(3)
     do j = 1,nx(2)
     do i = 1,nx(1)


        varE = abs(Eray(i,j,k  ) - (Eray0(i,j,k  ) + alpha_var*deltaEray(i,j,k  ))) / Eray(i,j,k)
        varT = abs(Tgaz(i,j,k  ) - (Tgaz0(i,j,k  ) + alpha_var*deltaTgaz(i,j,k  ))) / Tgaz(i,j,k)
        varE_max = max(varE, varE_max)
        varT_max = max(varT, varT_max)
        do idim = 1,ndim
           varF = abs(Fray(i,j,k,idim) - (Fray0(i,j,k,idim) + alpha_var*deltaFray(i,j,k,idim))) / Eray(i,j,k)/c
           varF_max = max(varF, varF_max)
        enddo


        Tgaz(i,j,k)         = Tgaz0(i,j,k       ) + alpha_var*deltaTgaz(i,j,k       )
        Eray(i,j,k)         = Eray0(i,j,k       ) + alpha_var*deltaEray(i,j,k       )
        Fray(i,j,k,1:ndim)  = Fray0(i,j,k,1:ndim) + alpha_var*deltaFray(i,j,k,1:ndim)

        
     enddo
     enddo
     enddo

#if PARA==1
     ! Resume reduction to see if f > 1 somewhere in the grid and if so then reduce dt
     info = 0
     call MPI_ALLREDUCE(resume,resume_glob,1,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,info)
     resume = resume_glob
#endif

     if (resume.gt.0) then
        Eray = Eray0
        Fray = Fray0
        Tgaz = Tgaz0

        dt = dt*0.8
        write(*,*) 'dt = ',dt

        read(*,*)
        goto 222
     endif

#if PARA==1
     ! Resmax reduction
     info = 0
     call MPI_ALLREDUCE(res_norme2,res_norme2_glob,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
     res_norme2 = res_norme2_glob

     ! Absolute residual reduction
     info = 0
     call MPI_ALLREDUCE(resmaxF,resmaxF_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)

     info = 0
     call MPI_ALLREDUCE(varE_max,v_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
     varE_max = v_glob

     info = 0
     call MPI_ALLREDUCE(varT_max,v_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
     varT_max = v_glob

     info = 0
     call MPI_ALLREDUCE(varF_max,v_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
     varF_max = v_glob

#else
     resmaxF_glob = resmaxF
#endif

     var_max = max(varE_max,varT_max)
     var_max = max(var_max ,varF_max)

     res_norme2 = sqrt(res_norme2)

     ! Convergence history
!!$     if (temps+dt >= tout(iout)) then
!!$        open(10,file='GS_history.dat',form='formatted',access='append')
!!$        write(10,111) nite,res_norme2
!!$111 format('nite, residual=',i5,ES10.3)
!!$        close(10)
!!$     endif

     if (res_norme2  < varres .or. nite>=nitemax_ray .or. var_max < varres) boucle = .false.


  enddo
  ! *********************
  ! End of convergence loop
  ! *********************
        
  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)
           
     xx = 0.
     do idim = 1,ndim
        xx = xx +  Fray(i,j,k,idim)**2.
     enddo
     xx = sqrt(xx)
     xm = un_moins*c*Eray(i,j,k)
           
     if (xx > xm) then
        if (xx/c/Eray(i,j,k)-1. > 1.0E-3 ) then
           Iconv = 0.
           write(*,*) '   i,j,k=',i,j,k
           write(*,*) '   Fray(i,j,k,1:ndim),normeF=',Fray(i,j,k,1:ndim),xx
           write(*,*) '   f   (i,j,k,1:ndim),normef=',Fray(i,j,k,1:ndim)/c/Eray(i,j,k),xx/c/Eray(i,j,k)
           write(*,*) '   normef-1.=',xx/c/Eray(i,j,k)-1.
           resume=1 
           read(*,*)
        endif
        Eray(i,j,k) = xx/c*(1.+1.0E-7)
     endif
     
  enddo
  enddo
  enddo
  
#if PARA==1
  ! Resume reduction to see if f > 1 somewhere in the grid and if so then reduce dt
  info = 0
  call MPI_ALLREDUCE(resume,resume_glob,1,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,info)
  resume = resume_glob
#endif
  
  if(resume .gt. 0) then
     write(*,*) 'Gauss-Seidel : Resume=',resume

     dt = dt * 0.8
     write(*,*) 'dt =',dt/seconde,'secondes'
     goto 222
  endif

  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)
           
     xx = 0.
     do idim = 1,ndim
        xx = xx +  Fray(i,j,k,idim)**2.
     enddo
     xx = sqrt(xx)
     xm = un_moins*c*Eray(i,j,k)
           
     if (xx > xm) then
        Iconv = 0.
        write(*,*) '*****!!!!!******   i,j,k=',i,j,k
        write(*,*) '   Fray(i,j,k,1:ndim),normeF=',Fray(i,j,k,1:ndim),xx
        write(*,*) '   f   (i,j,k,1:ndim),normef=',Fray(i,j,k,1:ndim)/c/Eray(i,j,k),xx/c/Eray(i,j,k)
        write(*,*) '   normef-1.=',xx/c/Eray(i,j,k)-1.
        Eray(i,j,k) = xx/c*0.9999999999999
        
        read(*,*)
     endif
     
  enddo
  enddo
  enddo

  ! Boundary conditions
  call communications_ray(temps_comm)
  call limites_ray

  if (nite >= nitemax_ray .and. matrix_inv_solver.ne.'GS-GMRES' .and. matrix_inv_solver.ne.'yoyo') then
     write(*,*) 'nite = ',nite
     read(*,*)
  endif

  nitetot_ray = nitetot_ray + nite

  ! Convergence history
  if(var_pasconv.ne.'00' .and. resmaxF==resmaxF_glob) then
     write(*,*) 'Variable that has converged worst: ',var_pasconv
     write(*,001) nite,resmaxF,imax_res+x_start(1)-1,jmax_res+x_start(2)-1,kmax_res+x_start(3)-1, nitetot_ray
  endif
001 format('nite=',I6,'  resmax=',1pe12.5,'  imax,jmax,kmax=',I4,I4,I4,'      nitetot =',I8)

  if (verbose) write(*,*) 'Fin de ray'

  if(nite < nitemax_ray) deja_converge=.true.


 ! Array deallocation
  deallocate(deltaTgaz,deltaEray,deltaFray)
  deallocate(Tgaz0,Eray0,Fray0)
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)
  deallocate(peray,pfray1,pfray,pdummy)
  
  return

end subroutine gs_mat

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_DEDD
!
!> Computes the Eddington tensor Dedd and its derivatives
!! with respect to Er and Fr, Dedd_dE and Dedd_dF,
!! respectively.
!<
subroutine cal_Dedd(E,F,Dedd,Dedd_dE,Dedd_dF,Derive_chi)

  use parameters, only : ndim
  use unites    , only : c
  use divers_ray, only : rad_trans_model

  implicit none
  
  integer :: i

  real                       :: E
  real,dimension(1:ndim    ) :: F
  real,dimension(3,3       ) :: Dedd,Dedd_dE
  real,dimension(3,3,1:ndim) :: Dedd_dF
  real                       :: normeF,chi,chiprime,gprime,hprime
  real                       :: fx,fy,fz,g,h,nx,ny,nz
  
  logical :: Derive_chi

  Dedd = 0. ; Dedd_dE = 0. ; Dedd_dF = 0.

  select case(rad_trans_model)
  
  case('P1')
  
     do i = 1,3
        Dedd(i,i) = 1./3.
     enddo
     return
     
  case('M1')

     fx = 0. ; fy = 0. ; fz = 0.

                   fx = F(1)/(c*E) 
     if(ndim.gt.1) fy = F(2)/(c*E)
     if(ndim.gt.2) fz = F(3)/(c*E)

     normef = 0.
     do i=1,ndim
        normef = normef + (F(i)/c/E)**2
     enddo
     normef = sqrt(normef)

     if(normef.gt.1) normef = 1.

     chi      = (3.+4.*normef**2)/(5.+2.*sqrt(4.-3.*normef**2))
     chiprime = 2.*normef/sqrt(4.-3.*normef**2)

     g      = (1.-chi)/2.
     h      = (3.*chi-1.)/2.
     gprime = -chiprime/2
     hprime = 3.*chiprime/2

     nx = 0. ; ny = 0. ; nz = 0.

     ! n is a unit vector
     normef = 0.
     do i=1,ndim
        normef = normef + (F(i)/c/E)**2
     enddo
     normef = sqrt(normef)

     if(normef.gt.1.e-5) then
        nx = fx/normef
        ny = fy/normef
        nz = fz/normef
     endif

     ! Eddington tensor
     Dedd(1,1) = g + h*nx**2 ; Dedd(1,2) =     h*nx*ny ; Dedd(1,3) =     h*nx*nz
     Dedd(2,1) =     h*ny*nx ; Dedd(2,2) = g + h*ny**2 ; Dedd(2,3) =     h*ny*nz
     Dedd(3,1) =     h*nz*nx ; Dedd(3,2) =     h*nz*ny ; Dedd(3,3) = g + h*nz**2

     if (Derive_chi) then

        ! Eddington tensor derivative with respect to Ei * Ei
        Dedd_dE(1,1) = - normef * (gprime + hprime*nx**2) ; Dedd_dE(1,2) = - normef * (         hprime*nx*ny) ; Dedd_dE(1,3) = - normef * (         hprime*nx*nz)
        Dedd_dE(2,1) = - normef * (         hprime*ny*nx) ; Dedd_dE(2,2) = - normef * (gprime + hprime*ny**2) ; Dedd_dE(2,3) = - normef * (         hprime*ny*nz)
        Dedd_dE(3,1) = - normef * (         hprime*nz*nx) ; Dedd_dE(3,2) = - normef * (         hprime*nz*ny) ; Dedd_dE(3,3) = - normef * (gprime + hprime*nz**2)

        if(normef.gt.1.e-5) then
           ! Eddington tensor derivative with respect to Fx * Ei
           Dedd_dF(1,1,1) = ( gprime*nx + hprime*nx*nx**2 + h*(-2.*nx*nx**2 /normef+2.*nx/normef) )/c
           Dedd_dF(1,2,1) = (             hprime*nx*nx*ny + h*(-2.*nx*nx*ny /normef+   ny/normef) )/c
           Dedd_dF(1,3,1) = (             hprime*nx*nx*nz + h*(-2.*nx*nx*nz /normef+   nz/normef) )/c 

           Dedd_dF(2,1,1) = (             hprime*nx*ny*nx + h*(-2.*nx*ny*nx /normef+   ny/normef) )/c
           Dedd_dF(2,2,1) = ( gprime*nx + hprime*nx*ny**2 + h*(-2.*nx*ny**2 /normef             ) )/c
           Dedd_dF(2,3,1) = (             hprime*nx*ny*nz + h*(-2.*nx*ny*nz /normef             ) )/c

           Dedd_dF(3,1,1) = (             hprime*nx*nz*nx + h*(-2.*nx*nz*nx /normef+   nz/normef) )/c
           Dedd_dF(3,2,1) = (             hprime*nx*nz*ny + h*(-2.*nx*nz*ny /normef             ) )/c
           Dedd_dF(3,3,1) = ( gprime*nx + hprime*nx*nz**2 + h*(-2.*nx*nz**2 /normef             ) )/c
   
           if(ndim.gt.1) then
              ! Eddington tensor derivative with respect to Fy * Ei
              Dedd_dF(1,1,2) = ( gprime*ny + hprime*ny*nx**2 + h*(-2.*ny*nx**2 /normef             ) )/c
              Dedd_dF(1,2,2) = (             hprime*ny*nx*ny + h*(-2.*ny*nx*ny /normef+   nx/normef) )/c
              Dedd_dF(1,3,2) = (             hprime*ny*nx*nz + h*(-2.*ny*nx*nz /normef             ) )/c 
                                                                                             
              Dedd_dF(2,1,2) = (             hprime*ny*ny*nx + h*(-2.*ny*ny*nx /normef+   nx/normef) )/c
              Dedd_dF(2,2,2) = ( gprime*ny + hprime*ny*ny**2 + h*(-2.*ny*ny**2 /normef+2.*ny/normef) )/c
              Dedd_dF(2,3,2) = (             hprime*ny*ny*nz + h*(-2.*ny*ny*nz /normef+   nz/normef) )/c
                                                                                             
              Dedd_dF(3,1,2) = (             hprime*ny*nz*nx + h*(-2.*ny*nz*nx /normef             ) )/c
              Dedd_dF(3,2,2) = (             hprime*ny*nz*ny + h*(-2.*ny*nz*ny /normef+   nz/normef) )/c
              Dedd_dF(3,3,2) = ( gprime*ny + hprime*ny*nz**2 + h*(-2.*ny*nz**2 /normef             ) )/c

              if(ndim.gt.2) then
                 ! Eddington tensor derivative with respect to Fz * Ei
                 Dedd_dF(1,1,3) = ( gprime*nz + hprime*nz*nx**2 + h*(-2.*nz*nx**2 /normef             ) )/c
                 Dedd_dF(1,2,3) = (             hprime*nz*nx*ny + h*(-2.*nz*nx*ny /normef             ) )/c
                 Dedd_dF(1,3,3) = (             hprime*nz*nx*nz + h*(-2.*nz*nx*nz /normef+   nx/normef) )/c 
              
                 Dedd_dF(2,1,3) = (             hprime*nz*ny*nx + h*(-2.*nz*ny*nx /normef             ) )/c
                 Dedd_dF(2,2,3) = ( gprime*nz + hprime*nz*ny**2 + h*(-2.*nz*ny**2 /normef             ) )/c
                 Dedd_dF(2,3,3) = (             hprime*nz*ny*nz + h*(-2.*nz*ny*nz /normef+   ny/normef) )/c
                                                                                                
                 Dedd_dF(3,1,3) = (             hprime*nz*nz*nx + h*(-2.*nz*nz*nx /normef+   nx/normef) )/c
                 Dedd_dF(3,2,3) = (             hprime*nz*nz*ny + h*(-2.*nz*nz*ny /normef+   ny/normef) )/c
                 Dedd_dF(3,3,3) = ( gprime*nz + hprime*nz*nz**2 + h*(-2.*nz*nz**2 /normef+2.*nz/normef) )/c

              endif
           endif
        endif

     endif
     
  end select
  
  return

end subroutine cal_Dedd

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine ESSAI_INV5X5
!
!> Inverts a 5x5 matrix.
!<
subroutine essai_inv5x5(mat,inv)

  implicit none
  integer :: idim
  real,dimension(1:5,1:5) :: mat,inv
  real :: det

  real :: a,b,c,d,e,f,g,h,i,j,k,l,m,n,p,q,alpha,beta,gamma,delta,eps,zeta,det_4x4

  a = mat(1,1) ; b = mat(1,2) ;     c = mat(1,3) ;     d = mat(1,4) ;     e = mat(1,5)
  f = mat(2,1) ; g = mat(2,2) ;     h = mat(2,3) ;     i = mat(2,4) ;     j = mat(2,5)
                 k = mat(3,2) ;     l = mat(3,3) ; alpha = mat(3,4) ;  beta = mat(3,5)
                 m = mat(4,2) ; gamma = mat(4,3) ;     n = mat(4,4) ; delta = mat(4,5)
                 p = mat(5,2) ;   eps = mat(5,3) ;  zeta = mat(5,4) ;     q = mat(5,5)

  inv = 0.
  do idim = 1,5
     inv(idim,idim) = 1.
  enddo

  det = a*det_4x4(g,h,i,j,k,l,alpha,beta,m,gamma,n,delta,p,eps,zeta,q) - f*det_4x4(b,c,d,e,k,l,alpha,beta,m,gamma,n,delta,p,eps,zeta,q)

  inv(1,1) =   det_4x4(g,h,i,j,k ,l,alpha,beta,m ,gamma,n,delta,p ,eps,zeta,  q)
  inv(2,1) = - det_4x4(f,h,i,j,0.,l,alpha,beta,0.,gamma,n,delta,0.,eps,zeta,  q)
  inv(3,1) =   det_4x4(f,g,i,j,0.,k,alpha,beta,0.,m    ,n,delta,0.,  p,zeta,  q)
  inv(4,1) = - det_4x4(f,g,h,j,0.,k,l    ,beta,0.,m,gamma,delta,0.,  p,eps ,  q)
  inv(5,1) =   det_4x4(f,g,h,i,0.,k,l   ,alpha,0.,m,gamma,    n,0.,  p,eps,zeta)

  inv(1,2) = - det_4x4(b,c,d,e,k ,l,alpha,beta,m ,gamma,n,delta,p ,eps,zeta,  q)
  inv(2,2) =   det_4x4(a,c,d,e,0.,l,alpha,beta,0.,gamma,n,delta,0.,eps,zeta,  q)
  inv(3,2) = - det_4x4(a,b,d,e,0.,k,alpha,beta,0.,m    ,n,delta,0.,  p,zeta,  q)
  inv(4,2) =   det_4x4(a,b,c,e,0.,k,l    ,beta,0.,m,gamma,delta,0.,  p,eps ,  q)
  inv(5,2) = - det_4x4(a,b,c,d,0.,k,l   ,alpha,0.,m,gamma,    n,0.,  p,eps,zeta)

  inv(1,3) =   det_4x4(b,c,d,e,g,h,i,j,m ,gamma,n,delta,p ,eps,zeta,  q)
  inv(2,3) = - det_4x4(a,c,d,e,f,h,i,j,0.,gamma,n,delta,0.,eps,zeta,  q)
  inv(3,3) =   det_4x4(a,b,d,e,f,g,i,j,0.,m    ,n,delta,0.,  p,zeta,  q)
  inv(4,3) = - det_4x4(a,b,c,e,f,g,h,j,0.,m,gamma,delta,0.,  p,eps ,  q)
  inv(5,3) =   det_4x4(a,b,c,d,f,g,h,i,0.,m,gamma,    n,0.,  p,eps,zeta)

  inv(1,4) = - det_4x4(b,c,d,e,g,h,i,j,k ,l,alpha,beta,p ,eps,zeta,  q)
  inv(2,4) =   det_4x4(a,c,d,e,f,h,i,j,0.,l,alpha,beta,0.,eps,zeta,  q)
  inv(3,4) = - det_4x4(a,b,d,e,f,g,i,j,0.,k,alpha,beta,0.,  p,zeta,  q)
  inv(4,4) =   det_4x4(a,b,c,e,f,g,h,j,0.,k,l    ,beta,0.,  p,eps ,  q)
  inv(5,4) = - det_4x4(a,b,c,d,f,g,h,i,0.,k,l,  alpha ,0.,  p,eps,zeta)

  inv(1,5) =   det_4x4(b,c,d,e,g,h,i,j,k ,l,alpha,beta,m ,gamma,n,delta)
  inv(2,5) = - det_4x4(a,c,d,e,f,h,i,j,0.,l,alpha,beta,0.,gamma,n,delta)
  inv(3,5) =   det_4x4(a,b,d,e,f,g,i,j,0.,k,alpha,beta,0.,m    ,n,delta)
  inv(4,5) = - det_4x4(a,b,c,e,f,g,h,j,0.,k,l    ,beta,0.,m,gamma,delta)
  inv(5,5) =   det_4x4(a,b,c,d,f,g,h,i,0.,k,l,  alpha ,0.,m,gamma,    n)
 
  inv = inv/det
  
  return

end subroutine essai_inv5x5

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function det_3x3
!
!> Computes the determinant of a 3x3 matrix
!<
function det_3x3(a,b,c,d,e,f,g,h,i)

  implicit none

  real :: a,b,c,d,e,f,g,h,i,det_3x3

  det_3x3 = a*e*i + c*d*h + b*f*g - c*e*g - b*d*i - a*f*h

end function det_3x3

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function det_4x4
!
!> Computes the determinant of a 4x4 matrix
!<
function det_4x4(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p)

  implicit none

  real :: a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,det_4x4,det_3x3

  det_4x4 =   a* det_3x3(f,g,h,j,k,l,n,o,p) &
            - e* det_3x3(b,c,d,j,k,l,n,o,p) &
            + i* det_3x3(b,c,d,f,g,h,n,o,p) &
            - m* det_3x3(b,c,d,f,g,h,j,k,l)

end function det_4x4
