!> \file
!! Contains subroutine init_ray()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_RAY
!
!> Initialises all the M1 radiative transfer parameters and
!! variables.
!<
subroutine init_ray

  use parameters
  use divers_ray
  use cl_ray
  use para      , only : mype

  implicit none

  integer :: i
  character(len=70) :: string

  namelist/param_ray/matrix_inv_solver,precond_ray,alpha,nitemax_ray,varrel_ray,cal_valp, &
                     alpha_explicite,valp_min,varrel_gaz,slope_type_ray,ngrp_ray,rad_trans_model

  matrix_inv_solver = 'GS'
  slope_type_ray    = 1
  precond_ray       = .false.
  alpha             = 1.
  nitemax_ray       = 100
  varrel_ray        = 0.05
  varrel_gaz        = 0.2
  cal_valp          = .true.
  valp_min          = 0.
  alpha_explicite   = 1.
  rad_trans_model   = 'M1' ! 'M1' or 'P1'
  ngrp_ray          = 1

  open  (2,file='liste.nml')
  rewind(2                 )
  read  (2,NML=param_ray   )
  close (2                 )

  if(ngrp_ray.ne.1) then
     write(string,'(a,i3)' ) ' M1 Multigroup not yet implemented, ngrp_ray=: ',ngrp_ray
     call close_program(mype,string)
  endif

  call allocate_array_ray

  if(cal_valp) call read_valp

  if(       matrix_inv_solver.ne.'GS'       .and. matrix_inv_solver.ne.'GMRES' .and. matrix_inv_solver.ne.'GS-GMRES' &
      .and. matrix_inv_solver.ne.'GMRES-GS' .and. matrix_inv_solver.ne.'yoyo' ) then
     call close_program(mype,' Unknown radiative matrix inversion solver: '//matrix_inv_solver)
  endif

  if(matrix_inv_solver.ne.'GS') then
     call allocate_gmres
     call initVarGmres()
  endif

  do i=1,6
     ordre_des_faces_pour_le_ray(i) = i
  enddo

  ordre_des_faces_pour_le_ray(1) = 1
  ordre_des_faces_pour_le_ray(2) = 2
  if(ndim.gt.2) then
     ordre_des_faces_pour_le_ray(3) = 5
     ordre_des_faces_pour_le_ray(4) = 6
     ordre_des_faces_pour_le_ray(5) = 4
     ordre_des_faces_pour_le_ray(6) = 3
  else
     ordre_des_faces_pour_le_ray(3) = 4
     ordre_des_faces_pour_le_ray(4) = 3
  endif

  return
  
end subroutine init_ray
