!> \file
!! Contains subroutine limites_ray()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine LIMITES_RAY
!
!> This routine fills the ghost cells at the domain
!! boundaries with the appropriate radiative transfer values
!! according to the type of boudary condition.
!!
!! Type of boundaries:
!!   - 1 : free-flow
!!   - 2 : periodic
!!   - 3 : reflexive
!!   - 4 : imposed values
!!   - 5 : analytical
!<
subroutine limites_ray
  
  use varray    , only : Eray,Fray
  use var       , only : Tgaz
  use parameters, only : ndim,nx,nxmin,nxmax,Nbuf
  use cl_ray
  use cl        , only : T_l
  use divers    , only : verbose
  use divers_ray, only : ordre_des_faces_pour_le_ray

  implicit none

  integer  :: i,iface,idim
  integer, dimension(1:6) :: face

  ! ICL = 1 : FREE-FLOW
  ! ICL = 2 : PERIODIC       (already done in the communications)
  ! ICL = 3 : REFLEXIVE
  ! ICL = 4 : IMPOSED VALUES (inside the ghost cells)
  ! ICL = 5 : ANALYTICAL

  if(verbose) write(*,*) 'Start of limites_ray'

  ! Order in which the faces are treated
  do i = 1,6
     face(i) = ordre_des_faces_pour_le_ray(i)
  enddo

  do iface = 1,2*ndim

    ! 1) ------------------

    if (face(iface) == 1) then
    
    if (ICL_RAY(1) == 1) then ! gradient nul
       do i=1-Nbuf,0
          Eray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Eray(1,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Tgaz(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Tgaz(1,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Fray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim) = Fray(1,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
       enddo
#if PARA==0
    else if (ICL_RAY(1) == 2) then ! periodique 
       do i=1-Nbuf,0
          Eray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Eray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Tgaz(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Tgaz(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Fray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim) = Fray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
       enddo
#endif
    else if (ICL_RAY(1) == 3) then ! refexif
       do i=1-Nbuf,0
          Eray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =  Eray(1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Tgaz(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =  Tgaz(1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Fray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim) =  Fray(1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
          Fray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1     ) = -Fray(1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1     )
       enddo

    else if (ICL_RAY(1) == 4) then ! valeurs imposees
       do i=1-Nbuf,0
          Eray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =   Er_l(1)
          Tgaz(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =    T_l(1)
          do idim =1,ndim
             Fray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),idim) =   Fr_l(idim,1)
          enddo
       enddo
    else if (ICL_RAY(1) == 5) then ! cl analytiques
       call cl_ana_ray(1)
    endif

    endif

    !2)----------------------------------------

    if (face(iface) == 2) then

    if (ICL_RAY(2) == 1) then ! gradient nul
       do i=1,Nbuf
          Eray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Eray(nx(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Tgaz(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Tgaz(nx(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Fray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim) = Fray(nx(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
       enddo
#if PARA==0
    else if (ICL_RAY(2) == 2) then ! periodique 
       do i=1,Nbuf
          Eray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Eray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Tgaz(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) = Tgaz(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Fray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim) = Fray(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
       enddo
#endif
    else if (ICL_RAY(2) == 3) then ! reflexif
       do i=1,Nbuf
          Eray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =  Eray(nx(1)+1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Tgaz(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =  Tgaz(nx(1)+1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       )
          Fray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim) =  Fray(nx(1)+1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
          Fray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1     ) = -Fray(nx(1)+1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1     )
       enddo

    else if (ICL_RAY(2) == 4) then ! valeurs imposees
       do i=1,Nbuf
          Eray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =   Er_l(       2)
          Tgaz(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ) =    T_l(       2)
          do idim =1,ndim
             Fray(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),idim) =   Fr_l(idim,2)
          enddo
       enddo
    else if (ICL_RAY(2) == 5) then ! cl analytiques
       call cl_ana_ray(2)
    endif

    endif

    !3)----------------------------------------

    if (face(iface) == 3) then

    if (ICL_RAY(3) == 1) then ! gradient nul
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) = Eray(nxmin(1):nxmax(1),1,nxmin(3):nxmax(3)       )
          Tgaz(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) = Tgaz(nxmin(1):nxmax(1),1,nxmin(3):nxmax(3)       )
          Fray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3),1:ndim) = Fray(nxmin(1):nxmax(1),1,nxmin(3):nxmax(3),1:ndim)
       enddo
#if PARA==0
    else if (ICL_RAY(3) == 2) then ! periodique 
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) = Eray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       )
          Tgaz(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) = Tgaz(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       )
          Fray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3),1:ndim) = Fray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3),1:ndim)
       enddo
#endif
    else if (ICL_RAY(3) == 3) then ! refexif
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) =  Eray(nxmin(1):nxmax(1),1-i,nxmin(3):nxmax(3)       )
          Tgaz(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) =  Tgaz(nxmin(1):nxmax(1),1-i,nxmin(3):nxmax(3)       )
          Fray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3),1:ndim) =  Fray(nxmin(1):nxmax(1),1-i,nxmin(3):nxmax(3),1:ndim)
          Fray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3),2     ) = -Fray(nxmin(1):nxmax(1),1-i,nxmin(3):nxmax(3),2     )
       enddo

    else if (ICL_RAY(3) == 4) then ! valeurs imposees
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) =   Er_l(       3)
          Tgaz(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       ) =    T_l(       3)
          do idim =1,ndim
             Fray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3),idim) =   Fr_l(idim,3)
          enddo
       enddo
    else if (ICL_RAY(3) == 5) then ! cl analytiques
       call cl_ana_ray(3)
    endif
    
    endif

    !4)----------------------------------------

    if (face(iface) == 4) then

    if (ICL_RAY(4) == 1) then ! gradient nul
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) = Eray(nxmin(1):nxmax(1),nx(2),nxmin(3):nxmax(3)       )
          Tgaz(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) = Tgaz(nxmin(1):nxmax(1),nx(2),nxmin(3):nxmax(3)       )
          Fray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3),1:ndim) = Fray(nxmin(1):nxmax(1),nx(2),nxmin(3):nxmax(3),1:ndim)
       enddo
#if PARA==0
    else if (ICL_RAY(4) == 2) then ! periodique 
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) = Eray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       )
          Tgaz(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) = Tgaz(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)       )
          Fray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3),1:ndim) = Fray(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3),1:ndim)
       enddo
#endif
    else if (ICL_RAY(4) == 3) then ! refexif
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) =  Eray(nxmin(1):nxmax(1),nx(2)+1-i,nxmin(3):nxmax(3)       )
          Tgaz(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) =  Tgaz(nxmin(1):nxmax(1),nx(2)+1-i,nxmin(3):nxmax(3)       )
          Fray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3),1:ndim) =  Fray(nxmin(1):nxmax(1),nx(2)+1-i,nxmin(3):nxmax(3),1:ndim)
          Fray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3),2     ) = -Fray(nxmin(1):nxmax(1),nx(2)+1-i,nxmin(3):nxmax(3),2     )
       enddo

    else if (ICL_RAY(4) == 4) then ! valeurs imposees
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) =   Er_l(4)
          Tgaz(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)       ) =    T_l(4)
          do idim =1,ndim
             Fray(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3),idim) =   Fr_l(idim,4)
          enddo
       enddo
    else if (ICL_RAY(4) == 5) then ! cl analytiques
       call cl_ana_ray(4)
    endif

    endif

    !5)----------------------------------------

    if (face(iface) == 5) then

    if (ICL_RAY(5) == 1) then ! gradient nul
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) = Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1       )
          Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) = Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1       )
          Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i,1:ndim) = Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1,1:ndim)
       enddo
#if PARA==0
    else if (ICL_RAY(5) == 2) then ! periodique 
     do i=1-Nbuf,0
        Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) = Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       )
        Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) = Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       )
        Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i,1:ndim) = Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i,1:ndim)
     enddo
#endif
    else if (ICL_RAY(5) == 3) then ! refexif
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) =  Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1-i       )
          Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) =  Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1-i       )
          Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i,1:ndim) =  Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1-i,1:ndim)
          Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i,3     ) = -Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1-i,3     )
       enddo
    else if (ICL_RAY(5) == 4) then
       do i=1-Nbuf,0
          Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) =  Er_l(5)
          Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       ) =   T_l(5)
          do idim =1,ndim
             Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i,idim) =  Fr_l(idim,5)
          enddo
       enddo
    else if (ICL_RAY(5) == 5) then ! cl analytiques
       call cl_ana_ray(5)
    endif

   endif


    !6)----------------------------------------

    if (face(iface) == 6) then

    if (ICL_RAY(6) == 1) then ! gradient nul
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) = Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)       )
          Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) = Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)       )
          Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i,1:ndim) = Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3),1:ndim)
       enddo
#if PARA==0
    else if (ICL_RAY(6) == 2) then ! periodique 
     do i=1,Nbuf
        Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) = Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       )
        Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) = Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i       )
        Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i,1:ndim) = Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i,1:ndim)
     enddo
#endif
    else if (ICL_RAY(6) == 3) then ! refexif
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) =  Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+1-i       )
          Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) =  Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+1-i       )
          Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i,1:ndim) =  Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+1-i,1:ndim)
          Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i,3     ) = -Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+1-i,3     )
       enddo
    else if (ICL_RAY(6) == 4) then
       do i=1,Nbuf
          Eray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) =   Er_l(6)
          Tgaz(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i       ) =    T_l(6)
          do idim =1,ndim
             Fray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i,idim) =   Fr_l(idim,6)
          enddo
       enddo
    else if (ICL_RAY(6) == 5) then ! cl analytiques
       call cl_ana_ray(6)
    endif

    endif

  enddo

  if(verbose) write(*,*) 'End of limites_ray'
  
  return

end subroutine limites_ray
