!     ---------------
subroutine INIT
!     ---------------
  use parameters 
  use unites
  use divers
  use var
  use geom        , only : x, x_glob, lbox
  use para        , only : mype, x_start
  use gravity
  use cl
  use gammas
  
  implicit none

#if PARA==1
  include 'mpif.h'
#endif

  integer :: i, j, k, idim, nx_dim, i1, i2
  Real, dimension(3) :: dx
  real :: rho1,rho2,p0,p,xc,yc

  namelist/condinit/lbox,rho1,rho2,p0
  !----------------------

  if (verbose)  write(*,*) 'debut de init'

  temps =0.

  lbox(1) = 0.5
  lbox(2) = 1.5
  lbox(3) = 1.
  rho1    = 1.
  rho2    = 2.
  p0      = 2.5

  ! Reading the namelist (previous values are erased)
  if(mype == 0) write(*,'(a)') 'reading namelist CONDINIT'
  open  (1,file='liste.nml')
  rewind(1                 )
  read  (1,NML=condinit    )
  close (1                 )

  tend   = tend   * seconde
  dt_max = dt_max * seconde
  
 
  !  maillage global
  !  ~~~~~~~~~~~~~~
  
  do idim = 1,3
     nx_dim = nx_glob(idim)
     
     x_glob(       1,idim) = 0
     x_glob(nx_dim+1,idim) = Lbox(idim)
     
     dx(idim) = (x_glob(nx_dim+1,idim) - x_glob(1,idim))/float(nx_dim)
     do i=2,nx_dim
        x_glob(i,idim) = x_glob(i-1,idim) + dx(idim)
     enddo
     
  enddo
  
  !  maillage local
  !  ~~~~~~~~~~~~~~
  
  do idim = 1,3
     
     i1 = 1
     if (icl(2*idim-1) == 0) i1 = 1-nbuf
     i2 = nx(idim)+1
     if (icl(2*idim  ) == 0) i2 = nx(idim)+1+nbuf
     
     do i=i1,i2
        x(i,idim) = x_glob(x_start(idim)+i-1,idim)
     enddo
     
  enddo


  ! Init
  ! ~~~~
  
  do k=1,nx(3) 
     do j=1,nx(2) 
        do i=1,nx(1) 
           
           xc = 0.5*(x(i,1)+x(i+1,1))
           yc = 0.5*(x(j,2)+x(j+1,2)) 
           
           if (yc>0.5*lbox(2)) then

              rho (i,j,k) = rho2
              p           = p0 - 0.1*rho2*(yc-0.5*lbox(2))
 
           else
              
              rho(i,j,k) = rho1
              p          = p0 + 0.1*rho1*(0.5*lbox(2)-yc)
              
           endif

           rhou(i,j,k,:) = 0.
           
           rhou(i,j,k,2) = rho(i,j,k)*0.01*(1+cos(4*PI*(xc-0.5*lbox(1))))*(1+cos(3*PI*(yc-0.5*lbox(2))))/4

           E   (i,j,k  ) = p/(gamma-1.) + 0.5*sum(rhou(i,j,k,:)**2)/rho(i,j,k)
           
              

        enddo
     enddo
  enddo
   
  if (verbose)  write(*,*) 'fin de init'

  return
  
end subroutine INIT

!*******************************************************************************************************************

!     ---------------------
subroutine DIMENSIONS
!     ---------------------
  use unites
  use gammas

  implicit none

  ! code units
  unitM = 1.0E+00 *gramme
  unitL = 1.0E+00 *centimetre
  unitT = 1.0E+00 *seconde
  unitD = unitM/unitL**3

  ! Specific heats ratio Cp/Cv
  gamma = 1.4
  
  return
  
end subroutine DIMENSIONS
