 module states
   real :: DL,UL,PL,CL
   real :: DR,UR,PR,CR
 end module states
 module gammas
   real gamma,g1,g2,g3,g4,g5,g6,g7,g8
 end module gammas
!
! *****************************************
!
 program exact

   use states
   use gammas

   implicit none

   integer :: i,Ncell
   real    :: diaph, domlen, ds, dx, pm, mpa, ps, s, timeout, um, us, xpos

   domlen  = 1.
   diaph   = 0.5
   Ncell   = 1000
   gamma   = 1.4
   timeout = 0.1

   !Sod
   !~~~
   dl      =  1.0 
   ul      =  0.0
   pl      =  1.0

   dr      =  0.125
   ur      =  0.0
   pr      =  0.01

   mpa     = 1.

   g1 = (gamma - 1.)/(2.*gamma)
   g2 = (gamma + 1.)/(2.*gamma)
   g3 = 2.0*gamma/(gamma - 1.0)
   g4 = 2.0      /(gamma - 1.0)
   g5 = 2.0      /(gamma + 1.0)
   g6 = (gamma - 1.0)/(gamma + 1.0)
   g7 = (gamma - 1.0)/2.0
   g8 = (gamma - 1.0)

   cl = sqrt(gamma*pl/dl)
   cr = sqrt(gamma*pr/dr)
   
   if (g4*(cl+cr) < (ur-ul) ) then

      write(*,*) 'You must include vacuum......'
      stop

   endif

   call starpu(pm,um,mpa)

   dx = domlen/float(Ncell)

   write(*,*) 'x     rho     u      P       E'
   do i=1,Ncell
     
      xpos = (float(i) -0.5)*dx
      s    = (xpos - diaph)/timeout

      call sample(pm,um,s,ds,us,ps)

      write(*,900) xpos, ds, us, ps/mpa, ps/ds/g8/mpa

   enddo

900 format(6(E12.4,2x))

 end program exact
!
!**********************************************************
!

!
!**********************************************************
!

!
!**********************************************************
!
 subroutine STARPU(p, u, mpa)
!
! Compute the pressure and velocity in the star region
!
   use states

   implicit none

   real :: p, u, mpa
   
   integer :: i,nriter
   real    :: change, fl, fld, fr, frd, pold, pstart, tolpre, udiff

   tolpre = 1.0E-06
   nriter = 20
   
   call GUESSP(pstart)

   pold  = pstart
   udiff = ur - ul

   write(*,*) '----------------------------'
   write(*,*) '  I        change           '
   
   do i=1,nriter 
      call prefun(fl,fld,pold,dl,pl,cl)
      call prefun(fr,frd,pold,dr,pr,cr)

      p = pold - (fl + fr + udiff)/(fld+frd)
      change = 2.0*(p-pold)/(p+pold)
      write(*,*) i,'     ',change
      if(change < tolpre) goto 20
      if (p < tolpre) p=tolpre      
      pold = p
   enddo

   write(*,*) 'No convergence of NR in STARPU'

20 continue
   
   u = 0.5*(ul + ur + fr - fl)

   write(*,*) 'P = ',p/mpa
   write(*,*) 'U = ',u

   return

 end subroutine STARPU
!
!**********************************************************
!

 subroutine GUESSP(pm)
!
! Choose a good pressure as 'first-guess'
!
   use states
   use gammas

   implicit none

   real, intent(out):: pm

   real :: cup, gel, ger, pmax, pmin, ppv, pq, ptl, ptr, qmax, quser, um

   quser = 2.
   
!  Pressure given by the PVRS solver

   cup = 0.25*(dl+dr)*(cl+cr)
   ppv = 0.5 *(pl+pr)+0.5*(ul-ur)*cup
   ppv  = max(0.0,ppv)
   pmin = min(pl,pr)
   pmax = max(pl,pr)
   qmax = pmax/pmin

   if (qmax < quser .and. (pmin <= ppv .and. ppv <= pmax)) then
      ! choose the pressure given by PVRS
      pm = ppv

   else
      
      if (ppv < pmin) then
         ! choose the solver with two rarefaction waves
         pq  = (pl/pr)**g1
         um  = (pq*ul/cl + ur/cr + g4*(pq-1.0))/(pq/cl + 1.0/cr)
         ptl = 1.0 + g7*(ul-um)/cl
         ptr = 1.0 + g7*(um-ur)/cr
         pm  = 0.5*(pl*ptl**g3 + pr*ptr**g3)

      else
         ! choose the solver with two shocks
         gel = sqrt((g5/dl)/(g6*pl+ppv))
         ger = sqrt((g5/dr)/(g6*pr+ppv))
         pm  = (gel*pl + ger*pr - (ur-ul))/(gel+ger)
         
      endif


   endif
   
   return

 end subroutine GUESSP
!
!**********************************************************
!

 subroutine PREFUN(f, fd, p, dk, pk, ck)
!
! functions of pression FL and FR
!
   use gammas
   
   implicit none

   real f, fd, p, dk, pk, ck
   real :: ak, bk, prat, qrt

   if (p < pk) then   !* rarefaction wave
      prat = p/pk
      f    = g4*ck*(prat**g1 - 1.0)
      fd   = (1.0/(dk*ck))*prat**(-g2)

   else               !* shock
      ak  = g5/dk
      bk  = g6*pk
      qrt = sqrt(ak/(bk + p))
      f   = (p-pk)*qrt
      fd  = (1.0-0.5*(p-pk)/(bk+p))*qrt
   endif

   return
 end subroutine PREFUN

!
!**********************************************************
!

 subroutine SAMPLE(PM, UM, S, D, U, P)
!
!  Search the solution between all the different waves
!
   use states
   use gammas

   implicit none

   real, intent(in ) :: pm, um, s 
   real, intent(out) :: d, p, u

   real :: c,cml,cmr,pml, pmr, shl, shr, sl, sr, stl, str

   
   if(s < um) then      !** We are at the LEFT of the contact discontinuity

      if (pm < pl) then      !* Rarefaction wave at left
         
         shl = ul - cl
         
         if  (s < shl) then           ! We are in the left state
            d = dl; u = ul; p=pl
         
         else
            
            cml = cl*(pm/pl)**g1
            stl = um - cml
            
            if ( s > stl ) then  ! we are in the star-left state
               d = dl*(pm/pl)**(1.0/gamma); u = um; p = pm
               
            else                 ! we are in the left fan
               c = g5*(cl + g7*(ul-s))  
               d = dl*(c/cl)**g4; u = g5*(cl + g7*ul + s) ; p = pl*(c/cl)**g3
               
            endif
         endif
       
      else                   !* Shock on the left

         pml = pm/pl
         sl  = ul - cl*sqrt(g2*pml + g1)

         if ( s < sl ) then      ! We are in the left state
            d = dl; u = ul; p=pl
            
         else                    ! we are in the star-left state
            d = dl*(pml + g6)/(pml*g6 + 1.0); u = um; p = pm
         endif
      endif

   else                 !** We are at the RIGHT of the contact discontinuity
         
      if (pm > pr) then     !* shock on the right
         pmr = pm/pr
         sr  = ur + cr*sqrt(g2*pmr + g1)
         
         if ( s > sr) then       ! we are in the right state
            d = dr; u = ur; p=pr
         else                    ! we are in the star-right state
            d = dr*(pmr + g6)/(pmr*g6 + 1.0);  u = um; p = pm
         endif
            
      else                 !* rarefaction wave at right
         shr = ur + cr

         if (s > shr) then       ! we are in the right state
            d = dr; u = ur; p=pr
         else
            cmr = cr*(pm/pr)**g1
            str = um + cmr
            
            if ( s < str) then   ! we are in the star-right state
               d = dr*(pm/pr)**(1.0/gamma) ; u = um; p = pm

            else                 ! we are in the right fan
               c = g5*(cr - g7*(ur-s))  
               d = dr*(c/cr)**g4; u = g5*(-cr + g7*ur + s) ; p = pr*(c/cr)**g3

            endif                  
         endif
         
      endif

   endif

   return

 end subroutine SAMPLE
